/* $Id: argset.c,v 1.3 2003/06/03 14:04:57 mquinson Exp $ */

/* argset.c - defines a set of arguments of tests.                           */

/* Copyright (C) 2001,2002,2003 the OURAGAN project.                         */
/* Code GPL'ed (see www.gnu.org).                                            */
/* Author(s): Martin Quinson                                                 */

#include "fast_internals.h"
#include "argset_elm.h" /* internal api */

#include "argset.h" /* external api */

#include <stdio.h>
#include <string.h> /* strtok */

/* internal representation of a set of args */
struct fast_argset_ {
  int elem_count;
  fast_argset_elm_t **elem;
  char *binary;

  /* to cache some caracteristics about the set */
  int nb_tests;    /* there will be #nb_tests# experiments */
  /* how many elems are of this type */
  int typed_count[fast_argset_type_count]; 
  /* how many possiblity there is for this type */
  int poss_count[fast_argset_type_count];
  char **string_id_list; /* list of all string id possibilities */
};

/*====[ Prototype of private functions ]=====================================*/

/*----[ fast_argset_commit ]------------------------------------( private)--*/
/* Convert and copy #argset# to #strings#.                                   */
/* #argset# a struct describing the args to do, #strings# is what should be  */
/*      passed to execv.                                                     */
/*                                                                           */
/* The functions suppose that #strings# is big enought                       */
/*  (it is the case when allocated by fast_argset_first)                     */
/*---------------------------------------------------------------------------*/
static int fast_argset_commit(char **strings,fast_argset_t *args);
/*----[ fast_argset_concat_str ]---------------------------------( private )-*/
/* Concat str2 at the end of str1,                                           */
/*   quoting (ie, doubling) any occurence of FAST_SEP in str2,               */
/*   puts a FAST_SEP between them if none is null,                           */
/*   frees the old str1 if needed,                                           */
/*   and return the adress of the new str1                                   */
/*   (or NULL if problems)                                                   */
/*                                                                           */
/* FIXME: realloc(str1) in place of using new_string                         */
/*---------------------------------------------------------------------------*/
static char *fast_argset_concat_str(char *str1,const char *str2);
/*----[ dump_string_list ]---------------------------------------( private )-*/
/* Debugging output                                                          */
/*---------------------------------------------------------------------------*/
static void dump_string_list(const char *name,int size,char **list);
/*----[ free_string_list ]---------------------------------------( private )-*/
/* Frees a #size# long string #list#.                                        */
/*---------------------------------------------------------------------------*/
static void free_string_list(int size,char ***list);
/*----[ fast_argset_alloc_string_id_list ]-----------------------------------*/
/* Return a list of all strings tests id.                                    */
/* Returns if anything went ok                                               */
/*---------------------------------------------------------------------------*/
static int fast_argset_alloc_string_id_list(fast_argset_t *argset);

/*===========================================================================*/
/*====[ Private functions ]==================================================*/

/*----[ fast_argset_concat_str ]---------------------------------( private )-*/
/* Concat str2 at the end of str1,                                           */
/*   quoting (ie, doubling) any occurence of FAST_SEP in str2,               */
/*   puts a FAST_SEP between them if none is null,                           */
/*   frees the old str1 if needed,                                           */
/*   and return the adress of the new str1                                   */
/*   (or NULL if problems)                                                   */
/*                                                                           */
/* FIXME: realloc(str1) in place of using new_string                         */
/*---------------------------------------------------------------------------*/
static char *fast_argset_concat_str(char *str1,const char *str2) 
{    
  //  int nb_sep; /* to count the number of separator in the string */
  char *new_string;//,*p,*q;

  if (!str2) 
    return str1;
  
  if (!(new_string=(char *)malloc(sizeof(char)*((str1 ? strlen(str1) : 1)
						+ strlen(str2) + 2)))) { 
    if (str1)
      free(str1);
    return NULL;
  }
  if (str1) {
    sprintf(new_string,"%s%c%s",str1,FAST_SEP,str2);
    free(str1);
  } else {
    sprintf(new_string,"%s",str2);
  }
  return new_string;
}

/*----[ dump_string_list ]---------------------------------------( private )-*/
/* Debugging output                                                          */
/*---------------------------------------------------------------------------*/
static void dump_string_list(const char *name,int size,char **list) 
{
  int i;
  printf("dump string_list %s at %p of size %d\n",
	 (name?name:"(noname)"),list,size);
  if (name) 
    printf("dump %s\n",name);
  fflush(stdout);
  for (i=0 ; i<size ; i++) {
    printf(" %s[%d] (@%p)",(name?name:"(noname)"),i,list[i]);fflush(stdout);
    printf("=%s\n",list[i]);
    fflush(stdout);
  }
  if (!name)
    printf("\n");
  fflush(stdout);
}

/*----[ free_string_list ]---------------------------------------( private )-*/
/* Frees a #size# long string #list#.                                        */
/*---------------------------------------------------------------------------*/
static void free_string_list(int size,char ***list) 
{
  int i;
  //  printf("free %d elem in %p\n",size,*list);fflush(stdout);
  if (*list && size>0) {
    for (i=0 ; i<size ; i++) 
      if ((*list)[i]) {
       	//printf("   free %s\n",(*list)[i]);fflush(stdout);
	free((*list)[i]);
      }
    free(*list);
  }
  *list=NULL;
}
  
/*===========================================================================*/
/*====[ Creator, destructor ]================================================*/

/*----[ fast_argset_free ]---------------------------------------------------*/
void fast_argset_free(fast_argset_t *argset) 
{
  int i;
  
  if (argset) {
    if (argset->string_id_list) {
      if (argset->poss_count[fast_argset_string]) {
	free_string_list(argset->poss_count[fast_argset_string],
			 &(argset->string_id_list));
      } else {
	/* in case of (id_count == 0 || id_size == 0), special non 
	   null string_id_list where allocated. Free it */
	free(argset->string_id_list[0]);
	free(argset->string_id_list);
      }
    }
    for (i=0 ; i<argset->elem_count ; i++) 
      fast_argset_elm_free(argset->elem[i]);
    if (argset->elem)   free(argset->elem);
    if (argset->binary) free(argset->binary);
    free(argset);
  }
}

/*----[ fast_argset_alloc ]--------------------------------------------------*/
/* Alloc the struct and do some initialization                               */
/*---------------------------------------------------------------------------*/
int fast_argset_alloc(const char *binary,
		      /*OUT*/fast_argset_t **where)
{
  int i;

  if (!(*where=(fast_argset_t *)malloc(sizeof(fast_argset_t)))) {
    FAIL("Out of memory\n");
  }
  (*where)->elem_count      = 0;
  (*where)->elem            = NULL;

  (*where)->binary          = strdup(binary);

  (*where)->nb_tests        = -1;
  for (i=0 ; i<fast_argset_type_count ; i++) {
    (*where)->typed_count[i]  = -1;
    (*where)->poss_count[i]   = -1;
  }

  (*where)->string_id_list = NULL;
  
  if (0) dump_string_list(NULL,0,NULL); /* turnaround warning unused */
  return 1; 
}

/*----[ fast_argset_alloc_string_id_list ]-----------------------------------*/
/* Return a list of all strings tests id.                                    */
/* Returns if anything went ok                                               */
/*---------------------------------------------------------------------------*/
static int fast_argset_alloc_string_id_list(fast_argset_t *argset)
{
  /* Handle copy of elem is a bit tricky. We can't aggregate what we find 
   * imediatly. We have to do with table of list of string, and then aggregate
   * the second level
   */
  int cur; /* number for this arg */
  int i,j,k,len;

  fast_argset_ctn_string_t *sctn;   
  int step=0; /* how big are the string lists */
  char ***list; 
  int list_size=0; /* used size */
  int insert_point; /* where to add entries */
  int id_count,id_size;

  if (!argset)
    return 0;

  if (!fast_argset_possibilities(fast_argset_string,argset,&id_count) ||
      !fast_argset_elem_count(fast_argset_string,argset,&id_size))
    return 0;
  if (id_count == 0 || id_size == 0) {
    /* Return a list containing one item, the empty string, so that fast works
       even if there is no string arg */
    if (!(argset->string_id_list=malloc(sizeof(char*))))
      FAIL("Out of memory");
    argset->string_id_list[0]=strdup("");
    return 1;
  }

  /* Alloc the working space */
  if (!(list=malloc(sizeof(char**)*id_count)))
    FAIL("Out of memory");
  memset(list,0,sizeof(char**)*id_count);

  for (i=0 ; i<id_count ; i++)
    if (!(list[i]=malloc(sizeof(char*)*id_size))) {
      while (--i >= 0)
	free(list[i]);
      free(list);
      FAIL("Out of memory\n");
    }

  /* Alloc the return space */
  if (!(argset->string_id_list=malloc(sizeof(char*)*id_count))) {
    for (i=0 ; i<id_count ; i++) free(list[i]);
    free(list);
    FAIL("Out of memory\n");
  }

  /* do the job */
  for (cur=0 ; cur<argset->elem_count ; cur++) {
    if (argset->elem[cur]->ctn->type == fast_argset_string &&
	argset->elem[cur]->to_id) {
      step++;
      insert_point=list_size;
      for (i=0 ; i==0 || i<list_size ; i++) {
	if (argset->elem[cur]->copyof != -1) {
	  list[i][step-1] = list[i][ argset->elem[cur]->copyof ];
	} else {
	  sctn=(fast_argset_ctn_string_t*)(argset->elem[cur]->ctn);
	  /* reuse the pos i for the first possible value of this arg */
	  list[i][step-1]=sctn->values[0];
	  if (!insert_point)
	    insert_point++;
	  /* put the others in free space */
	  for (j=1 ; j<sctn->size ; j++) {
	    for (k=0 ; k<step ; k++)
	      list[insert_point][k]=list[i][k];
	    
	    list[insert_point++][step-1]=sctn->values[j];
	  }
	}
      }
      list_size=insert_point;
    }
  }
  
  /* build the strings */
  for (i=0 ; i<id_count ; i++) {
    /* count how big it will be */
    len=step; /* separators and '\0' */
    for (j=0 ; j<id_size ; j++)
      len += strlen(list[i][j]);
    /* malloc */
    if (!(argset->string_id_list[i]=malloc(sizeof(char)*len))) {
      while (--i>=0) free((argset->string_id_list)[i]);
      for (i=0 ; i<id_count ; i++) free(list[i]);
      free(list);
      FAIL("Out of memory\n");
    }
    /* fill it */
    for (j=0 ; j<id_size ; j++) {
      if (j) {
	strcat(argset->string_id_list[i],"!");
	strcat(argset->string_id_list[i],list[i][j]);
      } else {
	strcpy(argset->string_id_list[i],list[i][j]);
      }
    }
  }
      

  /* free the working space */
  for (i=0 ; i<id_count ; i++)
    free(list[i]);
  free(list);
  
  return 1;
}  

/*----[ fast_argset_from_string ]--------------------------------------------*/
/* build the argset to bench #binary# described in #string#, frees #string#  */
/*---------------------------------------------------------------------------*/
int fast_argset_from_string(const char *binary,char *string,
			    /*OUT*/fast_argset_t **where)
{
  fast_argset_elm_t *arg; /* the one being filled */
  int i;

  char *tok, /* the token parsed (example: '%s=n|n') */
    *p; /* to travel in the token */

  printf("Parse the string `%.255s'\n",string);fflush(stdout);

  if (!fast_argset_alloc(binary,where))
    return 0;

  for (tok=strtok (string, ";") ; tok!=NULL ; tok=strtok (NULL, ";")) {
    DDEBUG1("Parse token `%s'\n",tok);

    (*where)->elem_count++;
    if (!((*where)->elem=realloc((*where)->elem,
				 sizeof(fast_argset_elm_t*)*
				 (*where)->elem_count))) {
      free(string);
      FAIL("Out of memory\n");
    }

    if (!((*where)->elem[(*where)->elem_count - 1]=
	  (fast_argset_elm_t*)malloc(sizeof(fast_argset_elm_t)))) {
      free(string);
      FAIL("Out of memory\n");
    }
    arg = (*where)->elem[ (*where)->elem_count - 1 ];    
    arg->copyof=-1;

    p=tok;
    if (*(p++) != '%') {
      ERROR2("Parse error while reading the %dth arg of tests (`%.255s'). \n"
	     "It should begin with '%%'\n",
	     (*where)->elem_count,
	     tok);
      fast_argset_free((*where));
      free(string);
      return 0;
    }
    
    if (*p == '!') {
      arg->to_id=0;
      p++;
    } else {
      arg->to_id=1;
    }

    if (*p != 'a' && *(p+1) != '=') {
      ERROR4("Parse error while reading the %dth arg of tests (`%.255s'). \n"
	     "%dth char should be '=' and not '%c'\n",
	     (*where)->elem_count,tok,p-string,*p);
      fast_argset_free((*where));
      free(string);
      return 0;
    }

    if (*p == 'c') { /* it's a copy */
      arg->copyof=atol(p+2);
      arg->ctn=(*where)->elem[ arg->copyof ]->ctn;

    } else if (*p == 'a') { /* it's a BLAS alpha type */
      char *buff;
      if (!(buff=strdup("0|1|2"))) {
	fast_argset_free((*where));
	free(string);
	FAIL("Out of memory\n.");
      }
      if (!fast_argset_ctn_from_string(fast_argset_string,buff,&(arg->ctn) )) {
	ERROR1("Parse error while parsing token `%.255s'\n."
	       "Abording\n",tok);
	fast_argset_free((*where));
	free(string);
	return 0;
      }
      free(buff);
 	
    } else if (*p=='s'||*p=='f'||*p=='d') { /*it's a string, float or integer*/
      if (!fast_argset_ctn_from_string(*p=='s'?
				       fast_argset_string :
				       fast_argset_num,
				       p+2,
				       &(arg->ctn) )) {
	ERROR1("Parse error while parsing token `%.255s'\n."
	       "Abording\n",tok);
	fast_argset_free((*where));
	free(string);
	return 0;
      }
    } else { /* unknown */
      fast_argset_free((*where));
      free(string);
      ERROR2("Parse error: unknown data type %c in test definition (%s)\n",
	      *p,tok);
      return 0;
    }
  }

  /* Sanity checks: 
   *  - we don't copy a non existant element
   * Still to do:
   *  - there is no loop in copy cycles.
   *  - we don't copy copies, but the original
   *  - we only copy things located before us in the elm table
   */
  for (i=0 ; i<(*where)->elem_count ; i++)
    if ((*where)->elem[i]->copyof >= (*where)->elem_count) {
      ERROR2("Element num %d is a copy of the non existant %d element.\n",
	     i,(*where)->elem[i]->copyof);
      fast_argset_free((*where));
      free(string);
      return 0;
    }
  free(string);

  /*
   * Allocate the string ids, so that it's always done, so that the freeing
   * code is easier.
   */

  return fast_argset_alloc_string_id_list(*where);
}

/*----[ fast_argset_dump ]---------------------------------------------------*/
/* debugging output, always return 1.                                        */
/*---------------------------------------------------------------------------*/
int fast_argset_dump(fast_argset_t *argset)
{
  fast_argset_ctn_num_t    *nctn;
  fast_argset_ctn_string_t *sctn;
  int i,cur;

  printf("----[ argset dump ]--------------------------\n");
  printf("  binary=`%s'\n",argset->binary);
  printf("  any param count=%d  ;  nb_tests=%d\n"
	 "  num params count=%d  ;  string params count=%d\n"
	 "  possible num=%d  ;  possible string=%d\n"
	 "  string_id_list=%p\n\n",
	 argset->elem_count,argset->nb_tests,    
	 argset->typed_count[fast_argset_num],
	 argset->typed_count[fast_argset_string],
	 argset->poss_count[fast_argset_num],
	 argset->poss_count[fast_argset_string],
	 argset->string_id_list);
  for (cur=0;cur<argset->elem_count;cur++) {
    nctn=(fast_argset_ctn_num_t   *)(argset->elem[cur]->ctn);
    sctn=(fast_argset_ctn_string_t*)(argset->elem[cur]->ctn);
    printf("  ....[ arg %d dump ].........................\n",cur);
    printf("    toid=%d  ;  copyof=%d\n",
	   argset->elem[cur]->to_id,argset->elem[cur]->copyof);
    printf("    type=%s(=%d)\n",
	   (nctn->type==fast_argset_num?"num":
	    (nctn->type==fast_argset_string?"string":"??")),
	   nctn->type);
    switch (nctn->type) {
    case fast_argset_num:
      printf("      val=%f\n",nctn->val);
      printf("      slot_count=%d  ;  slot_cur=%d\n",
	     nctn->slot_count,nctn->slot_cur);
      for (i=0;i<nctn->slot_count;i++) 
	printf("      slot %d=%f:%f:%f\n",
	       i,nctn->min[i],nctn->max[i],nctn->incr[i]);
      break;
    case fast_argset_string:
      printf("      size=%d  ;  pos=%d\n",sctn->size,sctn->pos);
      for (i=0;i<sctn->size;i++)
	printf("      slot %d=`%s'\n",i,sctn->values[i]);
      break;
    default: /*fast_argset_all is not a valid type. It's for searching*/
      break;
    }
  }
  return 1;
}
/*===========================================================================*/
/*====[ to get all possibility ]=============================================*/

/*----[ fast_argset_commit ]-------------------------------------( private)--*/
/* Convert and copy #argset# to #strings#.                                   */
/* #argset# a struct describing the args to do, #strings# is what should be  */
/*      passed to execv.                                                     */
/*                                                                           */
/* The functions suppose that #strings# is big enought                       */
/*  (it is the case when allocated by fast_argset_first)                     */
/*---------------------------------------------------------------------------*/
static int fast_argset_commit(char **strings,fast_argset_t *argset) 
{
  fast_argset_ctn_num_t    *nctn;
  fast_argset_ctn_string_t *sctn;

  int i;
  char **p;
  char buf[512];
  
  for (p=strings+1,i=0 ; 
       i<argset->elem_count ; 
       p++,i++) {
    if (*p)
      free(*p);
    nctn=(fast_argset_ctn_num_t   *)(argset->elem[i]->ctn);
    sctn=(fast_argset_ctn_string_t*)(argset->elem[i]->ctn);
    switch (sctn->type) {
    case fast_argset_string:
      *p=strdup(sctn->values[sctn->pos]);
      break;
    case fast_argset_num:
      snprintf(buf,512,"%f",nctn->val);
      *p=strdup(buf);
      break;
    default: /*fast_argset_all is not a valid type. It's for searching*/
      break;
    }
  }
  //dump_string_list(NULL,argset->poss_count[fast_argset_string],strings);
  return 1;
}

/*----[ fast_argset_first ]--------------------------------------------------*/
/* allocate the char** to be passed to execv, reset #argset#, and put the    */
/*  result in the output list.                                               */
/* Returns if all went ok                                                    */
/*---------------------------------------------------------------------------*/
int fast_argset_first(fast_argset_t *argset,
		      /*OUT*/char ***where) 
{
  int i;

  if (!(*where=malloc(sizeof(char *)*(argset->elem_count + 2))))
    FAIL("Out of memory\n");

  memset(*where,0,sizeof(char *)*(argset->elem_count + 2));
  (*where)[0]=strdup(argset->binary);
  for (i=0 ; i<argset->elem_count ; i++)
    if (!fast_argset_elm_reset(argset->elem[i]))
      return 0;
  if (!fast_argset_commit(*where,argset))
    return 0;
  return 1;
}

/*----[ fast_argset_next ]---------------------------------------------------*/
/* Increments the list of args (builded by test_args_first) to the next.     */
/* returns true if it was ok, and false if there is no new test to do        */
/*                                                                           */
/* If no increment was possible, frees #args_execv# and reset the args.      */
/*---------------------------------------------------------------------------*/
int fast_argset_next(fast_argset_t *argset,char ***where) 
{
  int i;
  int carry=1;

  for (i=0 ; i<argset->elem_count && carry ; i++)
    carry=fast_argset_elm_inc(argset->elem[i]);
  if (carry && *where) { /* frees #args_execv# */
    for (i=0; (*where)[i] ; i++)
      free((*where)[i]);
    free(*where);
    *where=NULL;
    return 0;
  }
  if (!fast_argset_commit(*where,argset))
    return 0;
  
  return !carry;
}

/*===========================================================================*/
/*====[ get some stats about the argset ]====================================*/

/*++++[ stats about the argset (cached) ]++++++++++++++++++++++++++++++++++++*/

/*----[ fast_argset_elem_count ]---------------------------------------------*/
/* Counts how many parameters of type #type# there is                        */
/*  use #fast_argset_all# to count all types                                 */
/* Returns if anything went ok                                               */
/*---------------------------------------------------------------------------*/
int fast_argset_elem_count(fast_argset_type_t type,fast_argset_t *argset,
			   /*OUT*/int *val)
{
  int i;
  
  if (!argset)
    return 0;

  if (type==fast_argset_all) {
    *val=argset->elem_count;
    return 1;
  }

  /* cache the result */
  if (argset->typed_count[type] != -1) {
    *val=argset->typed_count[type];
    return 1;
  }

  *val=0;
  for (i=0; i<argset->elem_count ;i++)
    if (argset->elem[i]->ctn->type==type && argset->elem[i]->to_id)
      (*val) ++;

  argset->typed_count[type]=*val;
  return 1;
}

/*----[ fast_argset_possibilities ]------------------------------------------*/
/* Count how many test there will be while variating args of type #type#.    */
/*  use #fast_argset_all# to search all types.                               */
/*  (usefull to allocate a table for each one)                               */
/*---------------------------------------------------------------------------*/
int fast_argset_possibilities(fast_argset_type_t type,fast_argset_t *argset,
			      /*OUT*/int *val)
{
  int this; /* number for this arg */
  int i;
  int cur;
  fast_argset_ctn_num_t    *nctn;
  fast_argset_ctn_string_t *sctn;   

  if (!argset) return 0;

  /* cache */
  if (argset->poss_count[type] != -1) {
    *val=argset->poss_count[type];
    return 1;
  }

  /* search */
  *val=0;
  for (cur=0 ; cur<argset->elem_count ; cur++) {
    this=0;
    if (argset->elem[cur]->to_id && argset->elem[cur]->copyof == -1) {
      nctn=(fast_argset_ctn_num_t*)   (argset->elem[cur]->ctn);
      sctn=(fast_argset_ctn_string_t*)(argset->elem[cur]->ctn);
      this=0;

      if (nctn->type == fast_argset_num &&
	  (type == fast_argset_num || type == fast_argset_all))
	for (i=0;i<nctn->slot_count;i++) {
	  this+=(int) ((nctn->max[i] - nctn->min[i]) / nctn->incr[i]) + 1;
	  if (((int) (nctn->max[i] - nctn->min[i]) ) % ((int)nctn->incr[i]))
	    this++;
	}

      if (nctn->type == fast_argset_string &&
	  (type == fast_argset_string || type == fast_argset_all))
	this=sctn->size;
    }

    if (this) 
      *val = (*val ? *val * this : this);
  }
  
  argset->poss_count[type]=*val;
  return 1;
  
}

/*----[ fast_argset_string_id_list ]-----------------------------------------*/
/* Give me the list of possible ids                                          */
/*---------------------------------------------------------------------------*/
int fast_argset_string_id_list(fast_argset_t *argset, /*OUT*/char ***where) 
{
  *where=argset->string_id_list;
  return 1;
}

/*++++[ Stats about this test ]++++++++++++++++++++++++++++++++++++++++++++++*/

/*----[ fast_argset_id ]-----------------------------------------------------*/
/* build the id for tests of type #type#, or all if #fast_argset_all# is used*/
/*   (in order to put it in a DB, for example)                               */
/* If searching for all types, append #prefix#                               */
/* The result of this function should be freed after use                     */
/* Returns if anything went ok                                               */
/*---------------------------------------------------------------------------*/
int fast_argset_id(fast_argset_type_t type,const char *prefix,
		   fast_argset_t *argset,
		   /*OUT*/char **where)
{
  fast_argset_ctn_num_t    *nctn;
  fast_argset_ctn_string_t *sctn;   
  char buffer[512];
  int cur;

  *where=NULL;
  if (!argset)
    return 1;

  if (type == fast_argset_all)
    *where=fast_argset_concat_str(NULL,prefix);

  for (cur=0 ; cur<argset->elem_count ; cur++) {
    if (argset->elem[cur]->to_id) {
      nctn=(fast_argset_ctn_num_t*)   (argset->elem[cur]->ctn);
      sctn=(fast_argset_ctn_string_t*)(argset->elem[cur]->ctn);
      if (nctn->type == fast_argset_string &&
	  (type == fast_argset_all || type == fast_argset_string)) {
	*where=fast_argset_concat_str(*where,sctn->values[sctn->pos]);
      }
      if (nctn->type == fast_argset_num &&
	  (type == fast_argset_all || type == fast_argset_num)){
	sprintf(buffer,"%f",nctn->val);
	*where=fast_argset_concat_str(*where,buffer);
      }
    }
  }
  if (!*where) /* no arg where found. Don't answer NULL */
    *where=strdup("");

  return 1;
}

/*----[ fast_argset_num_list ]-----------------------------------------------*/
/* build the id for numbers of this test (example, nb=3, list={128,128,128}) */
/* Returns if anything went ok                                               */
/*---------------------------------------------------------------------------*/
int fast_argset_num_list(fast_argset_t *argset,
			 /*OUT*/double **where)
{
  int i=0,cur;
  int nb_num;

  if (!argset)
    return 0;

  if (!fast_argset_possibilities(fast_argset_num,argset,&nb_num) ||
      nb_num==0)
    return 1;

  if (!(*where=(double *)malloc(sizeof(double) * nb_num )))
    FAIL("Out of memory\n");
  
  for (cur=0 ; cur<argset->elem_count ; cur++) {
    if (argset->elem[cur]->to_id) {
      switch (argset->elem[cur]->ctn->type) {
      case fast_argset_num:
	(*where)[i++]=((fast_argset_ctn_num_t*)argset->elem[cur]->ctn)->val;
	break;
      case fast_argset_string:
	/* no number to find in strings */
      case fast_argset_all:
	/*not a valid type. It's for searching*/
	break;
      }
    }
  }
  return 1;
}

