/* $Id: argset_ctn.c,v 1.3 2003/06/03 14:04:57 mquinson Exp $ */

/* arg_ctn.c - defines one argument of tests (this is the content of 
   a linked list defined in argset.[ch]). */

/* Copyright (C) 2001,2002,2003 the OURAGAN project.                         */
/* Code GPL'ed (see www.gnu.org).                                            */
/* Author(s): Martin Quinson                                                 */

#include "argset_ctn.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h> /* strdup */
#include "fast_internals.h"

/*===========================================================================*/
/*====[ memory management, initialisation ]==================================*/

/*----[ fast_arg_ctn_free ]--------------------------------------------------*/
/* well, frees the struct                                                    */
/*---------------------------------------------------------------------------*/
void fast_argset_ctn_free(fast_argset_ctn_t *arg)
{
  fast_argset_ctn_string_t *sarg=(fast_argset_ctn_string_t*)arg;
  fast_argset_ctn_num_t    *narg=(fast_argset_ctn_num_t   *)arg;

  int i;

  switch (arg->type) {
  case fast_argset_string:
    for (i=0;i<sarg->size;i++) {
      free(sarg->values[i]);
    }
    free(sarg->values);
    break;
  case fast_argset_num:
    free(narg->min);
    free(narg->max);
    free(narg->incr);
    break;
  default: /*not a valid type. It's for searching*/
    break;
  }
  free(arg);

}

/*----[ fast_argset_ctn_alloc ]----------------------------------------------*/
/* alloc an empty content structure of #type#, and do so initialization      */
/*---------------------------------------------------------------------------*/
static int fast_argset_ctn_alloc (fast_argset_type_t type,
				  /*OUT*/fast_argset_ctn_t **where)
{
  fast_argset_ctn_num_t    *nres; 
  fast_argset_ctn_string_t *sres;

  switch (type) {
  case fast_argset_num:
    if (!(nres=malloc(sizeof(fast_argset_ctn_num_t))))
      FAIL("Out of memory\n");
    
    *where=(fast_argset_ctn_t *)nres;
    nres->type=fast_argset_num;
    nres->slot_count=0;
    nres->slot_cur=0;
    nres->min = nres->max = nres->incr = NULL;
    break;

  case fast_argset_string:
    if (!(sres=malloc(sizeof(fast_argset_ctn_string_t)))) 
      FAIL("Out of memory\n");

    sres->type=fast_argset_string;
    sres->size=0; /* used as iterator to see where to add a string */
    sres->pos=0;
    sres->values=NULL;
    *where=(fast_argset_ctn_t *)sres;
    break;
  default: /*fast_argset_all is not a valid type. It's for searching*/
    break;
  }
  return 1;
}

/*----[ fast_argset_ctn_add_string ]-----------------------------------------*/
/* add the tests described by string #p# to the existing ctn struct          */
/*---------------------------------------------------------------------------*/
int fast_argset_ctn_add_string(fast_argset_type_t type,
			       char *p_input,
			       /*OUT*/fast_argset_ctn_t **where)
{
  fast_argset_ctn_num_t    *nres=(fast_argset_ctn_num_t    *)(*where); 
  fast_argset_ctn_string_t *sres=(fast_argset_ctn_string_t *)(*where); 

  int quoted; /* true if last char was an unquoted quote */
  int sep_found; /* at the end of a part, true if we found a separator */

  char *new,*p,*p_input_cpy;
  new = p = p_input_cpy = strdup(p_input);

  switch (type) {
  case fast_argset_num:
    nres->slot_count ++;
    if (!(nres->min  = realloc(nres->min,sizeof(double)*nres->slot_count))  ||
	!(nres->max  = realloc(nres->max,sizeof(double)*nres->slot_count))  ||
	!(nres->incr = realloc(nres->incr,sizeof(double)*nres->slot_count)))
      FAIL("Out of memory\n");
    
    nres->min[nres->slot_cur]=strtod(new,&p);
    if (p==new) {
      fast_argset_ctn_free(*where);
      FAIL("min not valid\n");
    }

    new=p;
    nres->max[nres->slot_cur]=strtod(new+1,&p);
    if (new+1==p) {
      fast_argset_ctn_free(*where);
      FAIL("Max not valid\n");
    }

    new=p;
    nres->incr[nres->slot_cur]=strtod(new+1,&p);
    if (new+1==p || nres->incr==0) {
      nres->incr[nres->slot_cur]=1; /* default value */
    }

    nres->val=nres->min[0];
    break;

  case fast_argset_string:
    /* do the following on the string freshly added :
       change the separator to a \0, 
       and strdup the end to the next string to add */

    while (*p != '\0') {
      /* FIXME: each member of the values is too big; 
	 it contains all futhers parts after a '\0'
	 For example, we will have:
	 "first string'\0'second string|third string"
	 "second string'\0'third string"
	 "third string"
	 but I don't see it as a critical issue */
      if (!(sres->values = realloc(sres->values,
				   sizeof(char*)*(++(sres->size) ))))
	FAIL("Out of memory\n");

      quoted=0;
      sep_found=0;
      for(p = sres->values[sres->size - 1] 
	    = strdup(p);
	  *p != '\0' && !sep_found;
	  (p)++ ) 
	{
	  if (*p == '|' && !quoted) {
	    sep_found=1;
	    *p = '\0'; /* stop this part */
	  }
	  if (*p == '\\') {
	    if (quoted) {
	      /* memmove the string to remplace \\ with \ */
	      memmove(p-1,p,strlen(p));
	    }
	    quoted = !quoted;
	  } else {
	    quoted = 0;
	  }
	}
    }
    break;
  default: /*fast_argset_all is not a valid type. It's for searching*/
    break;
  }
  free(p_input_cpy);
  return 1;
}

/*----[ fast_argset_ctn_from_string ]----------------------------------------*/
/* parse the string and allocate a new struct                                */
/*---------------------------------------------------------------------------*/
int fast_argset_ctn_from_string(fast_argset_type_t type,
				char *string,
				/*OUT*/fast_argset_ctn_t **where) 
{

  return fast_argset_ctn_alloc(type,where) &&
    fast_argset_ctn_add_string(type,string,where);
}

/*----[ fast_argset_ctn_add_string ]-----------------------------------------*/
/* parse the string and add the content to the existing struct               */
/*---------------------------------------------------------------------------*/


/*===========================================================================*/
/*====[ functions to travel between values ]=================================*/

/*----[ fast_argset_ctn_reset ]----------------------------------------------*/
/* put the value to zero                                                     */
/*---------------------------------------------------------------------------*/
int fast_argset_ctn_reset(fast_argset_ctn_t *arg) 
{

  switch(arg->type) {
  case fast_argset_string:
    ((fast_argset_ctn_string_t*)arg)->pos=0;
    break;
        
  case fast_argset_num:
    ((fast_argset_ctn_num_t*)arg)->val=
      ((fast_argset_ctn_num_t*)arg)->min[0];
    ((fast_argset_ctn_num_t*)arg)->slot_cur = 0;
    break;
  default: /*fast_argset_all is not a valid type. It's for searching*/
    break;
  }
  return 1;
}

/*----[ fast_argset_ctn_inc ]------------------------------------------------*/
/* increments this arg, and return if there is a carry                       */
/*---------------------------------------------------------------------------*/
int fast_argset_ctn_inc(fast_argset_ctn_t *arg) 
{
  fast_argset_ctn_string_t *sarg=(fast_argset_ctn_string_t*)arg;
  fast_argset_ctn_num_t    *narg=(fast_argset_ctn_num_t*)arg;

  switch(arg->type) {
  case fast_argset_string:
    if (++(sarg->pos) >= sarg->size) {/* was last string */
      fast_argset_ctn_reset(arg);
      return 1;
    }
    break;

  case fast_argset_num:
    narg->val += narg->incr[narg->slot_cur];
    if (narg->val > narg->max[narg->slot_cur]) {
      /* over maximum */
      narg->slot_cur++;
      if (!(narg->slot_cur < narg->slot_count)){
	fast_argset_ctn_reset(arg);
	return 1;
      }
      narg->val = narg->min[narg->slot_cur];
    }
    break;
  default: /*fast_argset_all is not a valid type. It's for searching*/
    break;
  }

  return 0;
}


