/* $Id: bench_all.c,v 1.7 2005/03/05 17:51:01 mquinson Exp $ */

/* bench_all - run all the needed experiments */

#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* sysconf */
#include <math.h>

#include "autobench.h"
#include "ArgSet/argset.h"
#include "Utils/diagnostic.h"
#include "Estimator/estimator.h"
#include "bench_all.h"

#include "bench_one.h" /* to do each bench */
#include "testsset.h"/* tests of same string id */

#undef DEBUG_bench_buffered
#undef DEBUG_bench_buffered_verbose

#ifdef DEBUG_bench_buffered_verbose
#define DEBUG_Buffer(word,val1,val2) do {\
  printf("%s (%s;",word,string_id);\
  for (i=0;i<nb_params;i++) \
    printf("%f;",params[i]);\
  printf(") = %f ; %f\n",val1,val2); }while (0);
#else
#define DEBUG_Buffer(mot,val1,val2) do {}while(0);
#endif

/*****************************************************************************/
/***                         Type definitions                              ***/
/*****************************************************************************/
/* struct describing which bench are to do (given by input plugin) */
struct fast_bench_todo_ {
  char *category; /* the path to the function */
  char *binary; /* name of the binary to test (example 'bench_blas') */
  char *args; /* the args to test 
		 (example "%s=dgemm;%s=n|t;%s=n|t;%d=128:1024:128;") */

  int options;

  fast_buff_t *buf;

  struct fast_bench_todo_ *next;
};  
typedef struct fast_bench_todo_ fast_bench_todo_t;


struct fast_bench_workspace_ {
  int count;
  fast_bench_testsset_t **testssets;
  fast_bench_todo_t *todo;          /* list of things to do */
  fast_benchs_to_store_t *to_store; /* result of the interpolation */
};
/* typedef in .h */
  

/*----[ f_b_w_new ]----------------------------------------------------------*/
/* Create the struct                                                         */
/*---------------------------------------------------------------------------*/
int fast_bench_workspace_new(fast_bench_workspace_t **ws) 
{
  if (!(*ws = malloc(sizeof(fast_bench_workspace_t)))) {
    FAIL("Malloc error\n");
  }
  (*ws)->count=0;
  (*ws)->testssets = NULL;
  (*ws)->todo = NULL;
  (*ws)->to_store = NULL;

  return 1;
}

/*----[ f_b_w_add ]----------------------------------------------------------*/
/* Add things to do in the workspace                                         */
/*---------------------------------------------------------------------------*/
int fast_bench_workspace_add(fast_bench_workspace_t *ws,
			     char *category,
			     char *binary,
			     char *args,
			     int options)
{
  fast_bench_todo_t *todo,*curr;

  if (!ws) {
    FAIL("Can't add something to do to a null workspace\n");
  }
  if (!(todo=(fast_bench_todo_t *)malloc(sizeof(fast_bench_todo_t)))) {
    FAIL("Malloc error\n");
  }

  /* passed as arguments to the function */
  todo->category = strdup(category);
  todo->binary = strdup(binary);
  todo->args= strdup(args);
  todo->options = options;
  /* hardcoded */
  fast_buff_open(&(todo->buf),todo->category,NULL,NULL);
  todo->next=NULL;
  if (ws->todo) {
    curr=ws->todo;
    while (curr->next) 
      curr=curr->next;
    curr->next=todo;
  } else {
    ws->todo = todo;
  }
  return 1;
}

/*****************************************************************************/
/***                    Prototype of private functions                     ***/
/*****************************************************************************/
/* does the tests for theses test_args, 
   and returns the corresponding testssets */
static int
fast_bench_workspace_do_benchs(fast_bench_workspace_t *ws);


/*----[ fast_got_bench ]-----------------------------------------------------*/
/* Makes what should be done after we've got a result:                       */
/*   put this in the buffer, and sync the buffer if needed                   */
/*---------------------------------------------------------------------------*/
int fast_got_bench(fast_bench_workspace_t *ws,
		   int cur_set, int cur_test,
		   fast_buff_t *buf,
		   int options);

/*----[ dump_options ]-------------------------------------------(internal)--*/
/* Debuging purpose.                                                         */
/*---------------------------------------------------------------------------*/
static void dump_options(int options) {
  LOG2("%s%s\n",
       (options & FAST_MEM_STD    ? "MEM_STD   \t"  : ""),
       (options & FAST_MEM_HIGH   ? "MEM_HIGH  \t"  : ""));

  LOG1("%s\n",
       (options & FAST_TIME       ? "TIME      \t" : ""));

  LOG4("%s%s%s%s\n",
       (options & FAST_BUF_IN     ? "BUF_IN    \t" : ""),
       (options & FAST_BUF_DO     ? "BUF_DO    \t" : ""),
       (options & FAST_BUF_OUT    ? "BUF_OUT   \t" : ""),
       (options & FAST_BUF_CREATE ? "BUF_CREATE\t" : ""));

  LOG3("%s%s%s\n",
       (options & FAST_FIT_LOW    ? "FIT_LOW   \t" : ""),
       (options & FAST_FIT_MED    ? "FIT_MED   \t" : ""),
       (options & FAST_FIT_HIGH   ? "FIT_HIGH  \t" : ""));
}

/* Take a unique testsset, and interpole it to build a to_store */
static fast_benchs_to_store_t *
testsset2to_store(fast_bench_testsset_t *testsset,
		   fast_benchs_to_store_t *to_store,
		   int options,char *category);
/*****************************************************************************/
/***                       Code of private functions                       ***/
/*****************************************************************************/

/* does the tests for theses test_args, 
   and returns the corresponding testssets */
static int
fast_bench_workspace_do_benchs(fast_bench_workspace_t *ws)
{
  int j;
  int wcur; /* current testsset */

  char **args_for_execv; /* for execv */
  char **list_string_id; /* to count how many char args there will be */
  /* do not free list_string_id since it's an internal data of the argset */
  int nb_tests_num; /* size of each testsset, ie nb of benchs to do */
  int nb_params; /* nb of numerical params to each bench */

  fast_argset_t *argset;
  char* string_id; /* identificator of test set */
  double *num_list; /* list of values of numerical params */

  /* parse the args */
  if (!fast_argset_from_string(ws->todo->binary,ws->todo->args,&argset)) {
    ERROR("Error while parsing the args. Skipping\n");
    ws->count=0;
    ws->testssets=NULL;
    return 0;
  }

  /* init some variables */
  if (!fast_argset_possibilities(fast_argset_string,argset,&(ws->count)) ||
      !fast_argset_elem_count(fast_argset_num,argset,&nb_params) ||
      !fast_argset_possibilities(fast_argset_num,argset,&nb_tests_num) ||
      !fast_argset_string_id_list(argset,&list_string_id)) {
    /* do not free list_string_id since it's an internal data of the argset */
    ws->count=-1;
    ws->testssets=NULL;
    fast_argset_free(argset);
    return 0;
  }
  if (ws->count == 0) {
    ws->count = 1;
  }

  /* allocate the needed testssets */
  if (!(ws->testssets=(fast_bench_testsset_t **)
	malloc(sizeof(fast_bench_testsset_t*) * ws->count))) {
    ERROR("Malloc error");
    ws->count=0;
    fast_argset_free(argset);
    return 0;
  }

  /* initialize the testssets */
  for (wcur=0 ; wcur<ws->count ; wcur++) {
    if (!((ws->testssets)[wcur] =
	  fast_bench_testsset_new(strdup(list_string_id[wcur]),
				  nb_tests_num,
				  nb_params))) {
      while (--wcur >= 0) {
	fast_bench_testsset_free((ws->testssets)[wcur]);
      }
      free (ws->testssets);
      ws->testssets=NULL;
      ws->count=0;
      fast_argset_free(argset);
      return 0;
    }
  }

  /*** Main loop on all tests ***/
  if (!fast_argset_first(argset,&args_for_execv)) {
    ws->testssets=NULL;
    ws->count=0;
    fast_argset_free(argset);
    return 0;
  }

  do {
    /* Get some info about this test */
    if (!fast_argset_id(fast_argset_string,NULL,argset,&string_id) ||
	!fast_argset_num_list(argset,&num_list)){
      ws->testssets=NULL;
      ws->count=0;
      fast_argset_free(argset);
      return 0;
    }

    /* search the good testsset */
    for (wcur=0 ; 
	 wcur<ws->count && strcmp((ws->testssets)[wcur]->id,string_id) ;
	 wcur++);
    if (wcur==ws->count) {
      ERROR1("Can't find a testsset of id '%s',"
	     "abording the bench of this binary\n",
	     string_id);
      for (wcur=0;wcur<ws->count;wcur++) {
	DDEBUG1("I have a testsset `%s'\n",(ws->testssets)[wcur]->id);
	fast_bench_testsset_free((ws->testssets)[wcur]);
      }
      free(ws->testssets);
      fast_argset_free(argset);
      ws->count=0;
      return 0;
    }

    /*   extract the numerical args and store them in place */
    for (j=0 ;  j<(ws->testssets)[wcur]->nb_params ; j++) {
      (ws->testssets)[wcur]->t_params[nb_params 
				     * (ws->testssets)[wcur]->t_size
				     + j ]
	= (ws->testssets)[wcur]->s_params[nb_params 
					 * (ws->testssets)[wcur]->t_size
					 + j] 
	= num_list[j];
      /* t_size = s_size. True until after verification of results */
    }

#ifdef TEST_FFT     
    {int N=1;
    for (j=0 ;  j<(ws->testssets)[wcur]->nb_params ; j++) {
      N*=(int)pow((double)j+2, num_list[j]+2);
      if (N>10000) {
	N=0; 
	break;
      }
    }
    if (!N) continue;
    printf("%d ", N);
    }
#endif
     
    /* Init the values about this test to -1 */
    (ws->testssets)[wcur]->t_val[(ws->testssets)[wcur]->t_size] = -1;
    (ws->testssets)[wcur]->s_val[(ws->testssets)[wcur]->t_size] = -1;

    /* Retrieve the buffered info */
    (ws->testssets)[wcur]->t_buf[(ws->testssets)[wcur]->t_size] = -1;
    (ws->testssets)[wcur]->s_buf[(ws->testssets)[wcur]->t_size] = -1;
    if (ws->todo->buf && ws->todo->options & FAST_BUF_IN) {
      fast_buff_retrieve(ws->todo->buf,
			 string_id,nb_params,num_list,
			 &((ws->testssets)[wcur]->t_buf[(ws->testssets)[wcur]
							->t_size]),
			 &((ws->testssets)[wcur]->s_buf[(ws->testssets)[wcur]
							->t_size]));
#ifdef DEBUG_bench_buffered_verbose
      printf("RETRIVE (%s;",string_id);
      for (j=0 ;  j<(ws->testssets)[wcur]->nb_params ; j++)
	printf("%f;",num_list[j]);
      printf(")= %f ; %f\n",
	     (ws->testssets)[wcur]->t_buf[(ws->testssets)[wcur]->t_size],
	     (ws->testssets)[wcur]->s_buf[(ws->testssets)[wcur]->t_size]);
#endif
    }

    /* Do the test */
    /*
    LOG3("Test %d of %d in testsset %s\n",
	 (ws->testssets)[wcur]->t_size, nb_tests_num,
	 (ws->testssets)[wcur]->id);
    */
    if (ws->todo->options & FAST_BUF_DO
	|| (ws->testssets)[wcur]->t_buf[(ws->testssets)[wcur]->t_size] == -1
	|| (ws->testssets)[wcur]->s_buf[(ws->testssets)[wcur]->t_size] == -1
	|| (ws->testssets)[wcur]->s_buf[(ws->testssets)[wcur]->t_size] == 0 ) {
      fast_do_bench(ws->todo->options,args_for_execv,
		    /* where to put the time val */
		    &((ws->testssets)[wcur]->t_val[(ws->testssets)[wcur]->t_size]),
		    /* where to put the space val */
		    &((ws->testssets)[wcur]->s_val[(ws->testssets)[wcur]->t_size]));
    }
    /* in the standalone version, the result is there right now */
    fast_got_bench(ws, wcur, (ws->testssets)[wcur]->t_size,
		   ws->todo->buf, ws->todo->options);
    /* the testsset have a bench more in it */
    (ws->testssets)[wcur]->t_size++; 
    (ws->testssets)[wcur]->s_size++; 
    
    /* free the infos about this test */
    free(string_id);
    free(num_list);

    } while (fast_argset_next(argset,&args_for_execv)); 

  fast_argset_free(argset);
  return 1;
}

/*----[ fast_got_bench ]-----------------------------------------------------*/
/* Makes what should be done after we've got a result:                       */
/*   put this in the buffer, and sync the buffer if needed                   */
/*---------------------------------------------------------------------------*/
int fast_got_bench(fast_bench_workspace_t *ws,
		   int cur_set, int cur_test,
		   fast_buff_t *buf,
		   int options)
{
  int do_store=0; /* if we should save this value */
#ifdef DEBUG_bench_buffered_verbose
  int i;
#endif

  fast_bench_testsset_t *testsset = ws->testssets[cur_set];

  /* Get the good TIME from the benched and the buffered one */
  double benched = testsset->t_val[cur_test];
  double buffered = testsset->t_buf[cur_test];

  testsset->t_val[cur_test] = 
    (benched != -1 ? 
     (buffered != -1 ? 
      (benched > buffered ? buffered : benched) :
      benched) :
     buffered);
  do_store |= (benched != -1 && benched != buffered);

  /* Get the good SPACE from the benched and the buffered one */
  benched = testsset->s_val[cur_test];
  buffered = testsset->s_buf[cur_test];

  testsset->s_val[cur_test] = 
    (benched != -1 ? 
     (benched < buffered ? buffered : benched) :
     buffered);
  do_store |= (benched != -1 && benched != buffered);

  /* Store the value if needed */
  if (buf && do_store && (options & FAST_BUF_OUT)) {
    fast_buff_store(buf,
		    testsset->id,testsset->nb_params,
		    /* params for a test are one shunk, so use it as table */
		    &(testsset->t_params[testsset->nb_params * 
					 testsset->t_size]),
		    testsset->t_val[cur_test],testsset->s_val[cur_test]);

#ifdef DEBUG_bench_buffered_verbose
    printf("STORE (%s;",testsset->id);
    for (i=0;i<testsset->nb_params;i++) 
      printf("%f;",testsset->t_params[testsset->nb_params * 
				      testsset->t_size + i]);
    printf(") = %f ; %f\n",
	   testsset->t_val[cur_test],testsset->s_val[cur_test]);
#endif
  }
#ifdef DEBUG_bench_buffered_verbose
  else {
    printf ("no STORE\n");
  }
#endif
#ifdef DEBUG_bench_buffered
  printf("buffered_bench(%s,",testsset->id);
  for (i=0; i<testsset->nb_params ; i++)
    printf ("%s%4d",(i?"":""),
	    (int)testsset->t_params[testsset->nb_params*cur_test + i]);
  printf(")={time=%f ; space=%f}\n",
	 testsset->t_val[cur_test],testsset->s_val[cur_test]);
  fflush(stdout);
#endif
#ifdef DEBUG_bench_buffered_verbose
  printf ("\n");
#endif     
    
  return 1;
}

static fast_benchs_to_store_t *
testsset2to_store(fast_bench_testsset_t *testsset,
		   fast_benchs_to_store_t *to_store,
		   int options,char *category) 
{
  fast_benchs_to_store_t *res;
  /* to do the estimations */
  fast_estimator_t *estim;
  int order = 1;

  if (!(res=(fast_benchs_to_store_t *)
	malloc(sizeof(fast_benchs_to_store_t)))) {
    ERROR("Malloc error\n");
    return to_store;
  }
  res->next = to_store;
  res->id = strdup(testsset->id);
  res->category=strdup(category);
  if (options & FAST_FIT_LOW) 
    order *= 2;
  if (options & FAST_FIT_MED)
    order *= testsset->nb_params;

  printf ("\n%s\n%s (test: %s ; string args: %s)\n%s\n",
	  "------------------",
	  "- TIME ESTIMATOR -",category, testsset->id, 
	  "------------------");
  printf("nb_params=%d ; t_size=%d ; order=%d\n",
	 testsset->nb_params, testsset->t_size,order); 
  fflush(stdout);
  estim=fast_estimator_alloc_array(testsset->nb_params,
				   testsset->t_size,
				   testsset->t_params,
				   testsset->t_val,
				   order);             
  if (estim) {
    res->time_interp = fast_estimator_to_string(estim);
    fast_estimator_test_array(estim,
			      testsset->nb_params,
			      testsset->t_size,
			      testsset->t_params,
			      testsset->t_val);
    //     fast_estimator_dump(stdout,estim);
    fast_estimator_free(estim);
  } else {
    free(res);
    return to_store;
  }
  
  printf ("\n%s\n%s (test: %s ; string args: %s)\n%s\n",
	  "-------------------",
	  "- SPACE ESTIMATOR -",category, testsset->id, 
	  "-------------------");
  fflush(stdout);
  estim=fast_estimator_alloc_array(testsset->nb_params,
				   testsset->s_size,
				   testsset->s_params,
				   testsset->s_val,
				   order);
  if (estim) {
    fast_estimator_test_array(estim,
			      testsset->nb_params,
			      testsset->s_size,
			      testsset->s_params,
			      testsset->s_val);
    res->space_interp = fast_estimator_to_string(estim);
    fast_estimator_free(estim);
  } else {
    free(res->time_interp);
    free(res);
    return to_store;
  }

  return res;
}


/*****************************************************************************/
/***                        Code of public functions                       ***/
/*****************************************************************************/

/*----[ fast_bench_all ]-----------------------------------------------------*/
/* The magic function doing all the kernel of fast/bench :                   */
/*   - parse the args to see which tests are needed                          */
/*   - realize them                                                          */
/*   - interpolate the result                                                */
/*   - build an answer to feed the output plugin                             */
/* (the todo in input are in a linked list, and the answer also)             */
/*---------------------------------------------------------------------------*/
int 
fast_bench_all(fast_bench_workspace_t *ws)
{
  int i;
  fast_bench_todo_t *next;

  while (ws->todo) {
    LOG ("============\n");
    LOG ("= NEW TEST =\n");
    LOG ("============\n");
    dump_options(ws->todo->options);

    /* do all the benchs */
    fast_bench_workspace_do_benchs(ws);

    /* interpole the testssets */
    for (i=0 ; i<ws->count ; i++)
      fast_bench_testsset_clean(ws->testssets[i]);/* clean anything before */
    for (i=0 ; i<ws->count ; i++) {
      ws->to_store=testsset2to_store(ws->testssets[i],ws->to_store,
			    ws->todo->options,ws->todo->category);
      fast_bench_testsset_free(ws->testssets[i]);
    }
    free(ws->testssets);
    fast_buff_close(ws->todo->buf);
    next=ws->todo->next;
    free(ws->todo->binary);
    free(ws->todo->category);
    free(ws->todo);
    ws->todo=next;
  }
  return 1;
}

/*----[ fast_bench_workspace_to_store ]--------------------------------------*/
/* build a linked list of things to store from the workspace                 */
/*---------------------------------------------------------------------------*/
int 
fast_bench_workspace_to_store(fast_bench_workspace_t *ws,
			      fast_benchs_to_store_t **to_store)
{
  *to_store=ws->to_store;
  free(ws);
  return 1;
}
