/* $Id: bench_one.c,v 1.3 2003/06/19 16:19:30 mquinson Exp $ */

/* bench_one - run one of the needed experiments */

#include "autobench.h"
#include "config.h"

#ifdef HAVE_SYS_SIGNAL_H
#include <sys/signal.h>
#endif

#if defined(HOST_OS_OPENBSD) || defined(HOST_OS_DARWIN)
#include <sys/wait.h>
#else
#include <wait.h>
#endif

#include <unistd.h> /* getpagesize */
#include <stdio.h>
#include <errno.h> /* errno */
#include <signal.h> /* SIGSEGV, SIGFPE,  SIGTRAP */

#include "getinfo.h"

#if !HAVE_WAIT3
# include <sys/times.h>
# ifndef HZ
#  include <sys/param.h>
# endif
# if !defined(HZ) && defined(CLOCKS_PER_SEC)
#  define HZ CLOCKS_PER_SEC
# endif
# if !defined(HZ) && defined(CLK_TCK)
#  define HZ CLK_TCK
# endif
# ifndef HZ
#  define HZ 60
# endif
#endif

#include "bench_one.h"

#define CACHE_SIZE 1024*1024*4

/* msec = milliseconds = 1/1,000 (1*10e-3) second.
   usec = microseconds = 1/1,000,000 (1*10e-6) second.  */

/* Systems known to fill in the average resident set size fields:
   SunOS 4.1.3 (m68k and sparc)
   Mt. Xinu 4.3BSD on HP9000/300 (m68k)
   Ultrix 4.4 (mips)
   IBM ACIS 4.3BSD (rt)
   Sony NEWS-OS 4.1C (m68k)

   Systems known to not fill them in:
   OSF/1 1.3 (alpha)
   BSD/386 1.1 (anything derived from NET-2)
   NetBSD 1.0 (4.4BSD-derived)
   Irix 5.2 (R4000)
   Solaris 2.3
   Linux 1.0

   It doesn't matter how many clock ticks/second there are on
   systems that don't fill in those fields.

   If the avgresident (%t) we print is less than a power of 2 away from
   the maxresident (%M), then we likely are using the right number.
   Another good check is comparing the average text size with the
   output of `size' on the executable.

   According to the SunOS manual, there are 50 ticks/sec on the sun3
   and 100 on the sun4.

   Some manuals have an apparent error, claiming that units for average
   sizes are kb*sec.  Judging by the contents of `struct rusage', it
   looks like it should be kb*ticks, like on SunOS.  Ticks/sec seems
   to be (empirically):
   50 Mt. Xinu
   250 Ultrix (mips)
   50 ACIS
   100 NEWS-OS

   sysconf(_SC_CLK_TCK) is *unrelated*.  */
#if defined(sun3) || defined(hp300) || defined(ibm032)
#define TICKS_PER_SEC 50
#endif
#if defined(mips)
#define TICKS_PER_SEC 250
#endif
#ifndef TICKS_PER_SEC
#define TICKS_PER_SEC 100
#endif

/* The number of milliseconds in one `tick' used by the `rusage' structure.  */
#define MSEC_PER_TICK (1000 / TICKS_PER_SEC)

/* Return the number of clock ticks that occur in M milliseconds.  */
#define MSEC_TO_TICKS(m) ((m) / MSEC_PER_TICK)

/*****************************************************************************/
/** Type declaration & prototype of private functions                       **/
/*****************************************************************************/

/* Internal struct on the resources used by a child process.  */
struct fast_resource_use_
{
  pid_t pid;                    /* PID of the process traced */
  int tracing; /* which kind of tracing: 0:none 1:syscall 2:each step */
  int waitstatus;
  double time;
  size_t size;

  int options; /* options as defined in fast.h */
  int ok; /* true if all went ok */
};

/*----[ fast_do_bench ]------------------------------------------------------*/
/*                                                                           */
/* Calls the 3 next functions to run a bench, and get all needed infos       */
/*                                                                           */
/* #mem_trace# = 0 => do not trace the memory (let the child run freely)     */
/*             = 1 => Trace after each syscall                               */
/*             = 2 => Trace after each step                                  */
/* #argv#: the things to execv()                                             */
/* #benched_time#: where to put the timing result                            */
/* #space#: where to put the spacing result                                  */
/*                                                                           */
/* Returns: iff anything went well                                           */
/*---------------------------------------------------------------------------*/
int fast_do_bench(int options,char **argv,
		  /*OUT*/double *benched_time,double *space){
  fast_resource_use_t *resource_use;
  fast_bench_result_t *result;

  /* avoid the fork to be valgrind friendly */
  //*benched_time=*space=1;
  //return 1;

  /* launch the children, and trace it as long as needed */
  resource_use = fast_resource_use_start(options,argv);
   while ( fast_resource_use_next (resource_use) )
     ;

  /* extract the infos for my caller */
  result=fast_resource_use_end(resource_use);

  if (!result)
    return 0;

  if (options & FAST_TIME)
    *benched_time = result->time;
  
  if (options & FAST_MEM_STD || options & FAST_MEM_HIGH)
    *space = result->size;
  
  free(result);
  
  return 1;
}

/*---[ fast_resource_use_start ]---------------------------------------------*/
/*                                                                           */
/* Prepare to measure a child process.                                       */
/*                                                                           */
/* Returns a newly allocated struct to pass to the other functions           */
/*---------------------------------------------------------------------------*/
fast_resource_use_t *fast_resource_use_start(int options,char **argv) {
  fast_resource_use_t *resp;
  char *cache; /* to empty the cache */

  resp=(fast_resource_use_t *)malloc(sizeof(fast_resource_use_t));
  if (resp==NULL) {
    fprintf(stderr,"malloc error");
    exit(1);
  }

  resp->options = options;

  /* FIXME: don't trace memory on OSF5.1, I can't manage it */
#ifdef HOST_OS_OSF5_1 
   resp->tracing = 0; 
#else
  if (options & FAST_MEM_STD) {
    resp->tracing = 1;
  } else if (options & FAST_MEM_HIGH) {
    resp->tracing = 2;
  } else {
    resp->tracing = 0;
  }
#endif
  /* empties the cache */
  if ((cache=malloc(CACHE_SIZE))) {
    memset(cache,4,CACHE_SIZE);
    free(cache);
  }

  /* Runs the child */
  resp->pid= fork();

  if (resp->pid < 0) {
    fprintf(stderr,"cannot fork\n");
    exit(1);

  } else if (resp->pid == 0) {         /* If child.  */

    /* prepare the ptrace from my father if I need it */
    // if (resp->tracing)
    //  fast_ptrace_traceme();
    kill(getpid(),SIGSTOP);

    /* Don't cast execvp arguments; that causes errors on some systems,
       versus merely warnings if the cast is left off.  */
    execvp (argv[0], argv);
    fprintf(stderr, "cannot run %s\n", argv[0]);
    _exit (errno == ENOENT ? 127 : 126);
  }/*  else {
    int i;
    
    fprintf(stderr, "Launched child %d on '",resp->pid);
    for (i=0; argv[i]; i++) {
      fprintf(stderr,"%s ",argv[i]);
    }
    fprintf(stderr,"'\n");
    }*/

  /* Inits the struct */
  resp->size = 0;
  resp->time = 0;

  return resp;
}

/*---[ fast_resource_use_next ]----------------------------------------------*/
/*                                                                           */
/* Look for the next execution step of the child                             */
/* Wait for and fill in data on child process PID.                           */
/*                                                                           */
/* Retuns false if the child is done (and true if it was restarted for       */
/*  further tracing)                                                         */
/*---------------------------------------------------------------------------*/
int fast_resource_use_next (fast_resource_use_t *resp) {

  pid_t caught; /* The pid of the child catched by wait() */
  int size;     /* the size observed (in kb)              */
  double measured_time;  /* stime+utime in sec            */

  /****
   * waits for the children
   ****/
  caught = waitpid(resp->pid,
		   &(resp->waitstatus),
		   (resp->tracing ? WUNTRACED :0));

  if (caught == -1) { /* Error in waitpid */
    fprintf(stderr,"%s:%d: Error in waitpid()\n",__FILE__,__LINE__);
    resp->ok=0;
    return 0;
  }

  /**** 
   * Get info about the child
   ****/
  resp->ok=fast_get_child_sizetime(resp->pid,&size,&measured_time);

  if (resp->ok && resp->options & FAST_MEM_STD) {
    if (resp->size < size)
      resp->size = size;
  }
  if (resp->ok && resp->time<measured_time ) 
      resp->time = measured_time;

  /****
   * re-run the child if needed
   ****/
  
  if (resp->tracing) { /* we are tracing the child for memory */

    if (WIFEXITED(resp->waitstatus)) { /* the child is dead normally*/
      resp->ok = WEXITSTATUS(resp->waitstatus)==0;
      return 0; /* stop */
    }
    if (WIFSIGNALED(resp->waitstatus) 
	&& WTERMSIG(resp->waitstatus)!=SIGTRAP) { /* untrapped signal, bad */
      fprintf(stderr,
	      "Child terminated by signal %d ! abord !\n",
	      WTERMSIG(resp->waitstatus));
      resp->ok = 0;
      return 0; /* stop */
    }
    if (WIFSTOPPED(resp->waitstatus)) { /* the child stopped */
      if (WSTOPSIG(resp->waitstatus)==SIGTRAP) {
	/* the signal got is SIGTRAP => the tracing is underway */
	/* re run the child */
	if (resp->tracing == 1) {
	  fast_ptrace_big_step(resp->pid,0);
	} else {
	  fast_ptrace_small_step(resp->pid,0);
	}
	return 1; /* continue */
      } else if (WSTOPSIG(resp->waitstatus)==SIGSEGV
		 || WSTOPSIG(resp->waitstatus)==SIGFPE) {
	fprintf(stderr,
		"Child %d got %s ! abord !\n",
		(int)resp->pid,
		(WSTOPSIG(resp->waitstatus)==SIGSEGV ? "segfault" : "sigfpe"));
	resp->ok=0;
	fast_ptrace_kill(resp->pid);
	waitpid(resp->pid,&(resp->waitstatus),0);
	return 0;
      } else {
	if (WSTOPSIG(resp->waitstatus)==SIGSTOP) {
	  kill(resp->pid,SIGCONT);
	} else {
	  printf("Child got signal %d\n",WSTOPSIG(resp->waitstatus));
	}
	/* got another signal => pass it to the child */
	if (resp->tracing == 1) {
	  fast_ptrace_big_step(resp->pid,WSTOPSIG(resp->waitstatus));
	} else {
	  fast_ptrace_small_step(resp->pid,WSTOPSIG(resp->waitstatus));
	}
	return 1; /* continue */
      } 
    } 

  } else { /* we are not tracing the child for memory */
    /* if signaled, ok=false */
    resp->ok = WIFEXITED(resp->waitstatus)
               && (WEXITSTATUS(resp->waitstatus)==0);
    return 0; /* stop */
  }

  fprintf(stderr,"%s:%d: Panic: execution should not reach that point...\n",
	  __FILE__,__LINE__);
  resp->ok = 0;
  return 0;
}


/*---[ fast_resource_use_end ]-----------------------------------------------*/
/*                                                                           */
/* Frees the resources occupied to trace the child,                          */
/*  and change it to an external struct                                      */
/*---------------------------------------------------------------------------*/
fast_bench_result_t *fast_resource_use_end (fast_resource_use_t *resp){
  fast_bench_result_t *bench;

  if (!resp)  return NULL;
  if (!resp->ok) {
    free(resp);
    return NULL;
  }

  bench=(fast_bench_result_t *)malloc(sizeof(fast_bench_result_t));
  if (bench==NULL) {
    fprintf(stderr,"malloc error");
    exit(1);
  }
  bench->time = resp->time;
  bench->size = resp->size;
  fprintf(stderr,"t=%f,s=%d\n",bench->time,bench->size);
  free(resp);
  return bench;
}

