/* $Id: buffer_db2.c,v 1.3 2003/10/10 06:18:21 mquinson Exp $ */

/* buffer_db2 - A buffer to store the result of experiments with db2 */

#include <stdlib.h>
#include <sys/types.h> /* for db */
#include <db.h>        /* for db */
#include <errno.h>
#include <string.h>    /* strerror() */
#include <signal.h>    /* to trap the signals */

#include "fast_buffer_plugins.h"

/*====[ type definition ]====================================================*/
typedef struct {
  DB *handle; /* handle to the db */
  DBT key, data;  /* to manipulate the db */

  DB_ENV dbenv;
  DB_INFO dbinfo;
} fast_buff_db2_data_t;

/*====[ prototype of internal functions ]====================================*/
static fast_buff_t *buf_used;
static char *db2_make_id(char *category, const char *string_id,
			 int nb_params, double *params);
static void db2_retrieve_internal(fast_buff_db2_data_t *db,
				  char *ID,
				  double *db_time,double *space);

void handler(int);

/*====[ internal code ]======================================================*/

/*----[ db2_make_id ]---------------------------------------------(private)--*/
/*                                                                           */
/* makes an id for the given parameters                                      */
/*                                                                           */
/*---------------------------------------------------------------------------*/
static char *db2_make_id(char *category, const char *string_id,
			 int nb_params, double *params) {
  char *ID,*p;
  int i;
  
  if (!(ID=(char *)malloc(sizeof(char)*(strlen(category)
					+strlen(string_id)
					+nb_params*65
					+1)))){
    return NULL;
  }
  sprintf(ID,"%s%c%s",category,FAST_SEP,string_id);
  p=ID;
  while(*p != '\0') p++;
  for (i=0;i<nb_params;i++) {
    sprintf(p,"%c%f",FAST_SEP,params[i]);
    while(*p != '\0') p++;
  }
  return ID;
}

/*----[ db2_retrieve_internal ]-----------------------------------(private)--*/
/*                                                                           */
/* do the job of retrieve the data. It exists because _store need to call it */
/*  to see if the slot exists or not, and I want to cache the ID creation.   */
/*                                                                           */
/*---------------------------------------------------------------------------*/
static void db2_retrieve_internal(fast_buff_db2_data_t *db,
				  char *ID,
				  double *db_time,double *db_space) 
{
  char *data,*p,*q;
  int ret; /* return value of db functions */
  int cpt;

  memset(&db->key, 0, sizeof(db->key));
  memset(&db->data, 0, sizeof(db->data));
  db->key.data = ID;
  db->key.size = (strlen(ID)) * sizeof(char);
  db->data.data = NULL;

  if ((ret = db->handle->get(db->handle, NULL, 
			       &db->key, &db->data, 0)) == 0) {
    /* we have to copy this away because there is no room in data.data for 
       the ending \0 */
    if (!(data=(char *)malloc(sizeof(char)*(db->data.size +1)))) {
      fprintf(stderr,"%s:%d:malloc error, can't retrieve the buffered value\n",
	      __FILE__,__LINE__);
      return;
    }
      
    for (cpt=0;cpt<db->data.size;cpt++) 
      data[cpt]=((char*)db->data.data)[cpt];
    data[cpt]='\0';
    *db_time=strtod(data,&p);
    if (data==p) {
      fprintf(stderr,"%s:%d: can't retrieve the buffered time value\n",
	      __FILE__,__LINE__);
      *db_time=-1;
      free(data);
      return;
    }
    *db_space=strtod(p,&q);
    if (p==q) {
      fprintf(stderr,"%s:%d: can't retrieve the buffered space value\n",
	      __FILE__,__LINE__);
      *db_time=*db_space=-1;
      free(data);
      return;
    }
    //    printf("%s:%d:db2_retrieve_internal(ID=%s) -> db_time=%f;db_space=%f\n",
    //	   __FILE__,__LINE__,ID,*db_time,*db_space);
    free(data);
  } 
  /* if we can't retrive the data, do nothing and do not change the values */

  free(ID);
}

/*----[ handler ]------------------------------------------------------------*/
/* Signal handler to close the DB.                                           */
/*---------------------------------------------------------------------------*/
void handler(int s) 
{
  fprintf(stderr,"\nfast_bench: got signal %d. Abording...\n",s);
  fast_buff_db2_close(buf_used);
  exit(1);
}

/*====[ code of public functions ]===========================================*/

/*----[ fast_buff_db2_open ]-------------------------------------------------*/
/* open the DB                                                               */
/*---------------------------------------------------------------------------*/
void fast_buff_db2_open(fast_buff_t **whereto,
			const char *category,const char *file,const char *path) {
  
  fast_buff_db2_data_t *data;
  int ret; /* return value of db functions */
  char *dbName;
  
  if (!(data=(fast_buff_db2_data_t*)malloc(sizeof(fast_buff_db2_data_t)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",__FILE__,__LINE__);
    *whereto=NULL;
    return;
  }
  if (!((*whereto)=(fast_buff_t*)malloc(sizeof(fast_buff_t)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",__FILE__,__LINE__);
    free(data);
    *whereto=NULL;
    return;
  }

  if (!((*whereto)->routine = strdup(category))) {
    free(data);
    free(*whereto);
    *whereto=NULL;
    return;
  }

  (*whereto)->plugdata = (void*)data;
  (*whereto)->type = fast_buff_db2_type;
  data->handle = NULL;

  if (file==NULL) {
    const char *envfile=getenv("FAST_HOSTNAME");
    
    if (!envfile)
      envfile=getenv("HOSTNAME");

    if (!envfile) {
      envfile = "noname";
      fprintf(stderr,
	      "fast: DB2 buffer: The name of the DB was not given,"
	      "and I can't manage to access the name of this host."
	      "\"noname.db2\" used.\n");
    }

    if (!(dbName=malloc(sizeof(char)*(strlen(envfile)+5)))) {
      fprintf(stderr,"Out of memory\n");
      free(data);
      free(*whereto);
      *whereto=NULL;
      return;
    }
    sprintf(dbName,"%s.db2",envfile);
  } else {
    if (!(dbName=strdup(file))) {
      fprintf(stderr,"Out of memory\n");
      free(data);
      free(*whereto);
      *whereto=NULL;
      return;
    }
  }

  
  /* Initialize the database environment. */
  memset(&data->dbenv,0,sizeof(data->dbenv));
  data->dbenv.db_errfile = stderr;
  data->dbenv.db_errpfx = "fast_buffer"; 
 
  if ((errno = db_appinit(path, NULL, &data->dbenv, DB_CREATE)) != 0) {
    fprintf(stderr,
	    "fast: DB2 buffer: db_appinit: %s\n", strerror(errno));
    free(data);
    free(*whereto);
    *whereto=NULL;
    free(dbName);
    return;
  }
  
  /* Initialize the database. */
  memset(&data->dbinfo, 0, sizeof(data->dbinfo));
  
  /*** opens the DB ****/
  if ((ret = db_open(dbName,DB_BTREE, DB_CREATE, 0664, 
		     &data->dbenv, &data->dbinfo, &data->handle)) != 0) {
    fprintf(stderr, "fast:buffer_db2: %s: %s\n", dbName, strerror(ret));
    fast_buff_db2_close(*whereto);
    *whereto=NULL;
    free(dbName);
    return;
  }
    
  memset(&data->key, 0, sizeof(data->key));
  memset(&data->data, 0, sizeof(data->data));

  buf_used=*whereto; /* allocates the global for the sighandler */
  signal(SIGINT,handler);
  signal(SIGQUIT,handler);
  signal(SIGTERM,handler);
  free(dbName);
}

void fast_buff_db2_close(fast_buff_t *db)
{  
  fast_buff_db2_data_t *data;
 
  if (db) {
    data=(fast_buff_db2_data_t *)db->plugdata;
    if (data) {
      if (data->handle) {
	data->handle->close(data->handle, 0);
	//      free (this->handle);
      }
      free(data);
    }
    free(db);
  }
}

/*----[ fast_buff_db2_retrieve ]---------------------------------------------*/
/*                                                                           */
/* Retrive a value from the buffer                                           */
/* (this function create an ID, and call the internal function to do the job)*/
/*                                                                           */
/*---------------------------------------------------------------------------*/
void fast_buff_db2_retrieve(fast_buff_t*db,
			    const char *stringargs,int nb_params,double*params,
			    /*OUT*/double *db_time,double *db_space)
{
  fast_buff_db2_data_t *data;
  char *ID;

  *db_time=*db_space=-1;
  if (db) {
    data=(fast_buff_db2_data_t *)db->plugdata;

    if (!(ID=db2_make_id(db->routine,stringargs, nb_params, params))) {
      fprintf(stderr,"%s:%d: malloc error, can't retrieve the buffered value\n",
	      __FILE__,__LINE__);
      return;
    }
    db2_retrieve_internal(data,ID,db_time,db_space);
  }
}


/*----[ fast_buff_db2_store ]------------------------------------------------*/
/*                                                                           */
/* Store a value in the buffer                                               */
/*---------------------------------------------------------------------------*/
void fast_buff_db2_store(fast_buff_t*db,
			 const char *stringargs,int nb_params,double*params,
			 /*OUT*/double db_time,double db_space) 
     /*void fast_buf_db2_store(fast_buf_t *t,int create_slot,
			char *string_id, int nb_params, double *params,
			double db_time,double space)*/
{
  fast_buff_db2_data_t *data;
  char dataStr[130];
  char *ID;
  int ret; /* return value of db functions */
   
  static int nb_step=10; /* flush the db each time we saved 10 values */

  if (!db) {
    fprintf(stderr,"%s:%d: Can't store anything in a NULL DB.\n",
	    __FILE__,__LINE__);
    return;
  }
  data=(fast_buff_db2_data_t *)db->plugdata;
    
  if (!(ID=db2_make_id(db->routine,stringargs, nb_params, params))) {
    fprintf(stderr,"%s:%d: malloc error, can't store the value\n",
	    __FILE__,__LINE__);
    return;
  }

  data->key.data = ID;
  data->key.size = (strlen(ID)) * sizeof(char);
  
  sprintf(dataStr,"%f %f",db_time,db_space);
  data->data.data = dataStr;
  data->data.size = (strlen(dataStr)) * sizeof(char);

  //printf ("store(%s)=%s...",ID,dataStr);
  if ((ret = data->handle->put(data->handle, NULL,
			       &data->key, &data->data, 0))) {
    fprintf(stderr, "fast:buffer_db2: Can't store the value in the DB: %s\n", 
	    strerror(ret));
  }

  /* flush the DB after 10 step */
  nb_step--;
  if (nb_step<=0) {
    nb_step=10;
    printf ("Sync\n");
    ret = data->handle->sync(data->handle,0);
    if (ret) {
      fprintf(stderr, "fast:buffer_db2: Can't flush the DB: %s\n", 
	      strerror(ret));
    }
    /* FIXME: reput the last data after a sync since it seems to forget it in this case */
    if ((ret = data->handle->put(data->handle, NULL,
				 &data->key, &data->data, 0))) {
      fprintf(stderr, "fast:buffer_db2: Can't store the value in the DB: %s\n", 
	      strerror(ret));
    }
   } else {
     //    printf ("ok\n");
   }
}




