/* $Id: buffer_db41.c,v 1.4 2003/10/10 06:18:21 mquinson Exp $ */

/* buffer_db41 - A buffer to store the result of experiments with db4.1 */

#include <stdlib.h>
#include <sys/types.h> /* for db */
#include <db.h>        /* for db */
#include <errno.h>
#include <string.h>    /* strerror() */
#include <signal.h>    /* to trap the signals */

#include "fast_buffer_plugins.h"

/*====[ type definition ]====================================================*/
typedef struct {
  DB *handle; /* handle to the db */
  DBT key, data;  /* to manipulate the db */
} fast_buff_db41_data_t;

/*====[ prototype of internal functions ]====================================*/
static fast_buff_t *buf_used;
static char *db41_make_id(char *category, const char *string_id,
			  int nb_params, double *params);
static void db41_retrieve_internal(fast_buff_db41_data_t *db,
				   char *ID,
				   double *db_time,double *db_space);

void handler(int);

/*====[ internal code ]======================================================*/

/*----[ db41_make_id ]---------------------------------------------(private)--*/
/*                                                                           */
/* makes an id for the given parameters                                      */
/*                                                                           */
/*---------------------------------------------------------------------------*/
static char *db41_make_id(char *category, const char *string_id,
			  int nb_params, double *params) {
  char *ID,*p;
  int i;
  
  if (!(ID=(char *)malloc(sizeof(char)*(strlen(category)
					+strlen(string_id)
					+nb_params*65
					+1)))){
    return NULL;
  }
  sprintf(ID,"%s%c%s",category,FAST_SEP,string_id);
  p=ID;
  while(*p != '\0') p++;
  for (i=0;i<nb_params;i++) {
    sprintf(p,"%c%f",FAST_SEP,params[i]);
    while(*p != '\0') p++;
  }
  return ID;
}

/*----[ db41_retrieve_internal ]-----------------------------------(private)--*/
/*                                                                           */
/* do the job of retrieve the data. It exists because _store need to call it */
/*  to see if the slot exists or not, and I want to cache the ID creation.   */
/*                                                                           */
/*---------------------------------------------------------------------------*/
static void db41_retrieve_internal(fast_buff_db41_data_t *db,
				  char *ID,
				  double *db_time,double *db_space) 
{
  char *data,*p,*q;
  int ret; /* return value of db functions */
  int cpt;

  memset(&db->key, 0, sizeof(db->key));
  memset(&db->data, 0, sizeof(db->data));
  db->key.data = ID;
  db->key.size = (strlen(ID)) * sizeof(char);
  db->data.data = NULL;

  if ((ret = db->handle->get(db->handle, NULL, 
			     &db->key, &db->data, 0)) == 0) {
    /* we have to copy this away because there is no room in data.data for 
       the ending \0 */
    if (!(data=(char *)malloc(sizeof(char)*(db->data.size +1)))) {
      fprintf(stderr,"%s:%d:malloc error, can't retrieve the buffered value\n",
	      __FILE__,__LINE__);
      return;
    }
      
    for (cpt=0;cpt<db->data.size;cpt++) 
      data[cpt]=((char*)db->data.data)[cpt];
    data[cpt]='\0';
    printf("db41_retrieve_internal(ID=%s)=%s",ID,data);
    *db_time=strtod(data,&p);
    if (data==p) {
      fprintf(stderr,"%s:%d: can't retrieve the buffered time value\n",
	      __FILE__,__LINE__);
      *db_time=-1;
      free(data);
      return;
    }
    *db_space=strtod(p,&q);
    if (p==q) {
      fprintf(stderr,"%s:%d: can't retrieve the buffered space value\n",
	      __FILE__,__LINE__);
      *db_time=*db_space=-1;
      free(data);
      return;
    }
    printf(" -> t=%f;s=%f\n",*db_time,*db_space);
    free(data);
  } 
  /* if we can't retrive the data, do nothing and do not change the values */

  free(ID);
}

/*----[ handler ]------------------------------------------------------------*/
/* Signal handler to close the DB.                                           */
/*---------------------------------------------------------------------------*/
void handler(int s) 
{
  fprintf(stderr,"\nfast_bench: got signal %d. Abording...\n",s);
  fast_buff_db41_close(buf_used);
  exit(1);
}

/*====[ code of public functions ]===========================================*/

/*----[ fast_buff_db41_open ]------------------------------------------------*/
/* open the DB                                                               */
/*---------------------------------------------------------------------------*/
void fast_buff_db41_open(fast_buff_t **whereto,
			 const char *category,const char *file,const char *path) {
  fast_buff_db41_data_t *data;
  int ret; /* return value of db functions */
  char *dbName;
  
  if (!(data=(fast_buff_db41_data_t*)malloc(sizeof(fast_buff_db41_data_t)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",__FILE__,__LINE__);
    *whereto=NULL;
    return;
  }
  if (!((*whereto)=(fast_buff_t*)malloc(sizeof(fast_buff_t)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",__FILE__,__LINE__);
    free(data);
    *whereto=NULL;
    return;
  }

  if (!((*whereto)->routine = strdup(category))) {
    free(data);
    free(*whereto);
    *whereto=NULL;
    return;
  }

  (*whereto)->plugdata = (void*)data;
  (*whereto)->type = fast_buff_db41_type;
  data->handle = NULL;

  if (file==NULL) {
    const char *envfile=getenv("FAST_HOSTNAME");
    
    if (!envfile)
      envfile=getenv("HOSTNAME");

    if (!envfile) {
      envfile = "noname";
      fprintf(stderr,
	      "fast: DB41 buffer: The name of the DB was not given,"
	      "and I can't manage to access the name of this host."
	      "\"noname.db41\" used.\n");
    }
    fprintf(stderr,"envfile=%s",envfile);
    if (!(dbName=malloc(sizeof(char)*(strlen(envfile)+6)))) {
      fprintf(stderr,"Out of memory\n");
      free(data);
      free(*whereto);
      *whereto=NULL;
      return;
    }
    sprintf(dbName,"%s.db41",envfile);
  } else {
    if (!(dbName=strdup(file))) {
      fprintf(stderr,"Out of memory\n");
      free(data);
      free(*whereto);
      *whereto=NULL;
      return;
    }
  }

  /* Create the database handle and open the underlying database. */
  if ((ret = db_create(&data->handle, NULL, 0)) != 0) {
    fprintf(stderr, "db_create: %s\n", db_strerror(ret));
    free(data);
    free(*whereto);
    *whereto=NULL;
    return;    
  }
  if ((ret = data->handle->open(data->handle,
				NULL, dbName, NULL, DB_BTREE, DB_CREATE, 0664)) != 0) {
    data->handle->err(data->handle, ret, "%s", dbName);
    free(data);
    free(*whereto);
    *whereto=NULL;
    return;
  }
    
  memset(&data->key, 0, sizeof(data->key));
  memset(&data->data, 0, sizeof(data->data));

  buf_used=*whereto; /* allocates the global for the sighandler */
  signal(SIGINT,handler);
  signal(SIGQUIT,handler);
  signal(SIGTERM,handler);
  free(dbName);
}

void fast_buff_db41_close(fast_buff_t *db)
{  
  fast_buff_db41_data_t *data;
 
  if (db) {
    data=(fast_buff_db41_data_t *)db->plugdata;
    if (data) {
      if (data->handle) {
	data->handle->close(data->handle, 0);
	//      free (this->handle);
      }
      free(data);
    }
    free(db);
  }
}


/*----[ fast_buf_db41_retrieve ]----------------------------------------------*/
/*                                                                           */
/* Retrive a value from the buffer                                           */
/* (this function create an ID, and call the internal function to do the job)*/
/*                                                                           */
/*---------------------------------------------------------------------------*/
void fast_buff_db41_retrieve(fast_buff_t *db,
			     const char *string_id, int nb_params, double *params,
			     double *db_time,double *db_space) 
{
  fast_buff_db41_data_t *data;
  char *ID;

  *db_time=*db_space=-1;
  if (db) {
    data=(fast_buff_db41_data_t *)db->plugdata;
    
    if (!(ID=db41_make_id(db->routine,string_id, nb_params, params))) {
      fprintf(stderr,"%s:%d: malloc error, can't retrieve the buffered value\n",
	      __FILE__,__LINE__);
      return;
    }
    db41_retrieve_internal(data,ID,db_time,db_space);
  }
}


/*----[ fast_buf_db41_store ]------------------------------------------------*/
/*                                                                           */
/* Store a value in the buffer                                               */
/* (honors the create_slot boolean, ie, if the boolean is false and if the   */
/*  data is not stored yet, do not store it)                                 */
/*                                                                           */
/*---------------------------------------------------------------------------*/
void fast_buff_db41_store(fast_buff_t *db, const char *string_id,
			  int nb_params, double *params,
			  double db_time,double db_space)
{
  fast_buff_db41_data_t *data;
  char dataStr[130];
  char *ID;
  int ret; /* return value of db functions */
   
  static int nb_step=10; /* flush the db each time we saved 10 values */
  
  if (!db) {
    fprintf(stderr,"%s:%d: Can't store anything in a NULL DB.\n",
	    __FILE__,__LINE__);
    return;
  }
  data=(fast_buff_db41_data_t *)db->plugdata;

  if (!(ID=db41_make_id(db->routine,string_id, nb_params, params))) {
    fprintf(stderr,"%s:%d: malloc error, can't store the value\n",
	    __FILE__,__LINE__);
    return;
  }

  data->key.data = ID;
  data->key.size = (strlen(ID)) * sizeof(char);
  
  sprintf(dataStr,"%f %f",db_time,db_space);
  data->data.data = dataStr;
  data->data.size = (strlen(dataStr)) * sizeof(char);

  printf ("store(%s)=%s...",ID,dataStr);
  if ((ret = data->handle->put(data->handle, NULL,
			       &data->key, &data->data, 0))) {
    fprintf(stderr, "fast:buffer_db41: Can't store the value in the DB: %s\n", 
	    db_strerror(ret));
  }
  /* flush the DB after 10 step */
  nb_step--;
  if (nb_step<=0) {
    nb_step=10;
    printf ("Sync\n");
    ret = data->handle->sync(data->handle,0);
    if (ret) {
      fprintf(stderr, "fast:buffer_db41: Can't flush the DB: %s\n", 
	      db_strerror(ret));
    }
    /* FIXME: reput the last data after a sync since it seems to forget it in this case */
    if ((ret = data->handle->put(data->handle, NULL,
				 &data->key, &data->data, 0))) {
      fprintf(stderr, "fast:buffer_db41: Can't store the value in the DB: %s\n", 
	      strerror(ret));
    }
  }

  else {
    printf ("ok\n");
  }

  free(ID);
}

