/* $Id: getinfo_linux.c,v 1.1 2003/06/19 12:18:50 mquinson Exp $ */

/* bench_linux -- getting info about child in a linux optimized maner */
/* most of this code was shamelessly stolen from procps */

#include "getinfo.h" /* see this file for more info */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include <asm/param.h>  /* HZ */
#ifndef HZ
#include <netinet/in.h>  /* htons */
#endif

#include <asm/page.h>   /* PAGE_SIZE */
#ifndef PAGE_SIZE
#warning PAGE_SIZE not defined, assuming it is 4096
#define PAGE_SIZE 4096
#endif

long smp_num_cpus;     /* number of CPUs */
static char buf[1024];

#define STAT_FILE    "/proc/stat"
static int stat_fd = -1;
#define UPTIME_FILE  "/proc/uptime"
static int uptime_fd = -1;


/* This macro opens filename only if necessary and seeks to 0 so
 * that successive calls to the functions are more efficient.
 * It also reads the current contents of the file into the global buf.
 */
#define FILE_TO_BUF(filename, fd) do{                           \
    static int local_n;                                         \
    if (fd == -1 && (fd = open(filename, O_RDONLY)) == -1) {    \
        fprintf(stderr, "Error: /proc must be mounted\n");      \
        fflush(NULL);                                           \
        _exit(102);                                             \
    }                                                           \
    lseek(fd, 0L, SEEK_SET);                                    \
    if ((local_n = read(fd, buf, sizeof buf - 1)) < 0) {        \
        perror(filename);                                       \
        fflush(NULL);                                           \
        _exit(103);                                             \
    }                                                           \
    buf[local_n] = '\0';                                        \
}while(0)


/***********************************************************************
 * Some values in /proc are expressed in units of 1/HZ seconds, where HZ
 * is the kernel clock tick rate. One of these units is called a jiffy.
 * The HZ value used in the kernel may vary according to hacker desire.
 * According to Linus Torvalds, this is not true. He considers the values
 * in /proc as being in architecture-dependant units that have no relation
 * to the kernel clock tick rate. Examination of the kernel source code
 * reveals that opinion as wishful thinking.
 *
 * In any case, we need the HZ constant as used in /proc. (the real HZ value
 * may differ, but we don't care) There are several ways we could get HZ:
 *
 * 1. Include the kernel header file. If it changes, recompile this library.
 * 2. Use the sysconf() function. When HZ changes, recompile the C library!
 * 3. Ask the kernel. This is obviously correct...
 *
 * Linus Torvalds won't let us ask the kernel, because he thinks we should
 * not know the HZ value. Oh well, we don't have to listen to him.
 * Someone smuggled out the HZ value. :-)
 *
 * This code should work fine, even if Linus fixes the kernel to match his
 * stated behavior. The code only fails in case of a partial conversion.
 *
 */
static unsigned long Hertz=-1;
static void init_Hertz_value(void) __attribute__((constructor));
static void init_Hertz_value(void){
  unsigned long user_j, nice_j, sys_j, other_j;  /* jiffies (clock ticks) */
  double up_1, up_2, seconds;
  unsigned long jiffies, h;

  smp_num_cpus = sysconf(_SC_NPROCESSORS_CONF);
  if(smp_num_cpus<1) smp_num_cpus=1;
  do{
    FILE_TO_BUF(UPTIME_FILE,uptime_fd);  sscanf(buf, "%lf", &up_1);
    /* uptime(&up_1, NULL); */
    FILE_TO_BUF(STAT_FILE,stat_fd);
    sscanf(buf, "cpu %lu %lu %lu %lu", &user_j, &nice_j, &sys_j, &other_j);
    FILE_TO_BUF(UPTIME_FILE,uptime_fd);  sscanf(buf, "%lf", &up_2);
    /* uptime(&up_2, NULL); */
  } while((long)( (up_2-up_1)*1000.0/up_1 )); /* want under 0.1% error */
  
  jiffies = user_j + nice_j + sys_j + other_j;
  seconds = (up_1 + up_2) / 2;
  h = (unsigned long)( (double)jiffies/seconds/smp_num_cpus );
  /* actual values used by 2.4 kernels: 32 64 100 128 1000 1024 1200 */
  switch(h){
  case   30 ...   34 :  Hertz =   32; break; /* ia64 emulator */
  case   48 ...   52 :  Hertz =   50; break;
  case   58 ...   62 :  Hertz =   60; break;
  case   63 ...   65 :  Hertz =   64; break; /* StrongARM /Shark */
  case   95 ...  105 :  Hertz =  100; break; /* normal Linux */
  case  124 ...  132 :  Hertz =  128; break; /* MIPS, ARM */
  case  195 ...  204 :  Hertz =  200; break; /* normal << 1 */
  case  253 ...  260 :  Hertz =  256; break;
  case  295 ...  304 :  Hertz =  300; break; /* 3 cpus */ 
  case  393 ...  408 :  Hertz =  400; break; /* normal << 2 */
  case  495 ...  504 :  Hertz =  500; break; /* 5 cpus */
  case  595 ...  604 :  Hertz =  600; break; /* 6 cpus */
  case  695 ...  704 :  Hertz =  700; break; /* 7 cpus */  
  case  790 ...  808 :  Hertz =  800; break; /* normal << 3 */
  case  895 ...  904 :  Hertz =  900; break; /* 9 cpus */ 
  case  990 ... 1010 :  Hertz = 1000; break; /* ARM */
  case 1015 ... 1035 :  Hertz = 1024; break; /* Alpha, ia64 */
  case 1095 ... 1104 :  Hertz = 1100; break; /* 11 cpus */
  case 1180 ... 1220 :  Hertz = 1200; break; /* Alpha */
  default:
#ifdef HZ
    Hertz = (unsigned long)HZ;    /* <asm/param.h> */
#else
    /* If 32-bit or big-endian (not Alpha or ia64), assume HZ is 100. */
    Hertz = (sizeof(long)==sizeof(int) || htons(999)==999) ? 100UL : 1024UL;
#endif
    fprintf(stderr, "Unknown HZ value! (%ld) Assume %ld.\n", h, Hertz);
  }
  printf("Hertz=%lu\n",Hertz);
  close(uptime_fd); uptime_fd=-1;
  close(stat_fd);   stat_fd=-1;
}

/*----[ end of stolen code ]-------------------------------------------------*/
int fast_get_child_sizetime(pid_t pid,/*OUT*/int *size,double *time) {
  static pid_t tracedpid=1;
  static int fd=-1;
  char filename[256],*tmp,*tmp2;
  double utime,stime;
  int i;

  /* init */
  if (Hertz == -1) 
    init_Hertz_value();
  *time = *size  = -1;

  /* If we change pid */
  if (tracedpid != pid) {
    if (fd != -1) {
      close(fd);
      fd=-1;
    }
    sprintf(filename,"/proc/%d/stat",pid);
    tracedpid = pid;
  }
  /* read /proc file */
  FILE_TO_BUF(filename,fd);
  //printf("buf=%s",buf);
  
  /* parse it */
  tmp  = strrchr(buf, ')');      /* search end of cmdname */
  tmp += 2;                      /* skip next space, too  */

  for (i=0;i<10;i++)             /* skip 10 fields        */
    while (*(++tmp) != ' ');
  tmp2=tmp;
  while (*(++tmp2) != ' ');
  *(tmp2++)='\0';

  //printf("tmp=%s %s",tmp,tmp2);
  utime=atol(tmp);

  tmp=tmp2;
  while (*(++tmp2) != ' ');
  *(tmp2++)='\0';
  //  printf("tmp2=%s %s",tmp,tmp2);
  stime=atol(tmp);

  *time = (utime + stime)/Hertz;
  *time = *time ? *time : 0.01; /* FIXME: don't return 0, but precision */
  //printf("utime=%f stime=%f time=%f\n",utime,stime,*time);

  for (i=0;i<9;i++)              /* skip 9 fields         */
    while (*(tmp++) != ' ');
  sscanf(tmp,"%d",size); 
  //  printf("raw size=%d page_size=%d\n",*size,PAGE_SIZE);  
  *size = *size * (PAGE_SIZE / 1024);
  //  printf("time=%f size=%d\n",*time,*size);
  //fflush(stdout);
  //sprintf(buf,"ps -o 'rss,time,args' %d 2>&1",pid);
  //system(buf);
  //printf("\n");
  return 1;
}
