/* $Id: fast_avail.c,v 1.9 2003/07/09 14:12:16 mquinson Exp $ */

/* fat_avail.c - the functions allowing FAST to get the system availability  */

/* Copyright (C) 2003 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include "fast_avail_plugins.h"

/*====[ Prototypes of private functions ]====================================*/
static int fast_avail_elm_alloc(int resource,char **hosts,fast_avail_elm_t **whereto);
static void fast_avail_elm_free(fast_avail_elm_t *elm);
static int fast_avail_elm_use(fast_avail_elm_t *elm,double *value);

/*====[ Globals ]============================================================*/

/*----[ fast_nws_cache ]-----------------------------------------------------*/
/* The main cache for NWS in FAST. Each cell of this tab a generic cache     */
/*---------------------------------------------------------------------------*/
fast_cache_t *fast_avail_cache[FAST_RESOURCE_TYPE_COUNT];

/*----[ host count for each resource ]---------------------------------------*/
int fast_resource_nb_hosts[FAST_RESOURCE_TYPE_COUNT]={
  1,1,1,1, /* all resources about a unique host */
  2,2,     /* bandwidth and latency are between 2 hosts */
  1};      /* CpuCount is about a uniq host */

const char *fast_resource_name[FAST_RESOURCE_TYPE_COUNT]=
  {"available CPU","current CPU","memory","disk","bandwidth","latency",
   "CPU count"};

/*----[ Cache in which avail_cache elements are stored by ID ]---------------*/
fast_cache_t *fast_avail_cache_IDs;

/*----[ The table of the plugins ]-------------------------------------------*/
const int fast_avail_plugins_count = 2 /* intsens and pathfinder */
#ifdef HAVE_NWS
+1
#endif
;

const fast_avail_plugin_t fast_avail_plugins[]={
#ifdef HAVE_NWS
  {"NWS", &fast_nws_init, &fast_nws_elm_init, &fast_nws_refresh, 
   &fast_nws_autofetch_start, &fast_nws_autofetch_work,
   &fast_nws_edges_list, &fast_nws_restart_cpu,NULL},
#endif 
  {"Internal sensor", &fast_internal_sensor_init, &fast_internal_sensor_elm_init,
   &fast_internal_sensor_refresh,NULL,NULL,NULL,&fast_internal_sensor_restart_cpu,
   NULL},

  /* Path plug must be the last plugin handling networking */
  {"Pathfinder", NULL, &fast_pathfinder_elm_init, &fast_pathfinder_refresh,
   NULL,NULL,NULL,NULL,&fast_pathfinder_data_free}
};

/*----[ Configuration ]------------------------------------------------------*/
int fast_config_autofetch;
double fast_config_autofetch_on;
double fast_config_autofetch_off;

/**
 * fast_avail_init:
 *
 * Initialize all the availability plugins 
 */


int fast_avail_init(void) {
  int i;
  int errcode=0;

  /* auto fetching */
  if (!fast_config_getfirst_int("autofetch_use",&fast_config_autofetch))
    fast_config_autofetch=0;

  if (!fast_config_getfirst_double("autofetch_on",
				   &fast_config_autofetch_on))
    fast_config_autofetch_on=60;

  if (!fast_config_getfirst_double("autofetch_off",
				   &fast_config_autofetch_off))
    fast_config_autofetch_off=60*10;

  for (i=0; i<fast_avail_plugins_count; i++)
    if (fast_avail_plugins[i].init && !fast_avail_plugins[i].init())
      errcode++;

  return !errcode;
}

/**
 * fast_avail:
 * @resource: the fast_resource to get
 * @hostSrc: the name of the host (preferably the short name)
 * @hostDest: the name of the dest host if the test concerns two hosts
 * @value: the result
 * @Returns: true if anything went OK, false if not.
 *
 * Retrieve a @value from the NWS system of type @resource_type (on of the 
 * FAST_RESOURCE_*). If there is a resource concerning a unique host (as CPU 
 * or Memory), @hostSrc is its name (preferably the short name, without the
 * domainname of this host and @hostDest is ignored. If there is a resource
 * concerning two hosts, @hostSrc and @hostDest are their names. 
 *
 * The returned value is corrected by all bookings existing on this host
 * (using fast_cbook_use()).
 */


int fast_avail(int resource,const char *hostSrc,const char *hostDest,
	       /* OUT */ double *value) 
{
  char **keys=NULL;
  fast_avail_elm_t *elm;
  int i;

  /* Do the autofetch work */
  for (i=0; i<fast_avail_plugins_count; i++)
    if (fast_avail_plugins[i].autofetch_work)
      fast_avail_plugins[i].autofetch_work();
  
  /* Check the args */
  ASSERT(!hostSrc,"hostSrc==(null) in fast_avail");
  ASSERT(fast_resource_nb_hosts[resource] == 2 && !hostDest,
         "hostDest==(null) in fast_avail, but the resource needs 2 hosts");
  
  if (!(keys=malloc(sizeof(char*)*2)))
    FAIL("Out of memory\n");
  
  keys[0]=(char*)hostSrc;
  keys[1]=(char*)(hostDest ? hostDest: "");

  /* search the elm in the cache */
  if (!fast_multicache_retrieve(fast_avail_cache[resource],
                                fast_resource_nb_hosts[resource],
                                (const char **)keys,
                                (void*)&elm)) {
    free(keys);
    FAIL("Internal error\n");
  }


  if (!elm) { /* elm not found in the cache */
    if (fast_resource_nb_hosts[resource]==1) 
      WARN2("Data for %s on %s not found in cache, retreive it.\n",
	    fast_resource_name[resource],hostSrc);
    else
      WARN3("Data for %s between %s and %s not found in cache, retreive it.\n",
	    fast_resource_name[resource],hostSrc,hostDest);
    
    /* allocate an entry */
    if (!fast_avail_elm_alloc(resource,keys,&elm))
      return 0;

    /* store it back */
    keys[0]=strdup(keys[0]);
    keys[1]=fast_resource_nb_hosts[resource]==2 ? strdup(keys[1]) : keys[1];
    if (fast_resource_nb_hosts[resource]==1) 
      INFO3("Found %f for %s on %s. Store it.\n",
	    elm->value,fast_resource_name[resource],hostSrc);
    else
      INFO4("Found %f for %s between %s and %s. Store it.\n",elm->value,
	    fast_resource_name[resource],hostSrc,hostDest);
    if (!fast_multicache_insert(&( fast_avail_cache[resource] ),
                                fast_resource_nb_hosts[resource],
                                (char**)keys,
                                elm,(void (*)(void*))&fast_avail_elm_free)) {
      free(keys);
      FAIL("Internal error\n");
    }
    /* store it also in the cache sorted by series if autofetching */
    if (fast_config_autofetch && elm->ID)
      if (!fast_cache_insert(&fast_avail_cache_IDs,strdup(elm->ID),elm,NULL)) {
	free(keys);
        FAIL("Internal error\n");
      }
  } 

  free(keys);
  /* retrieve the value */
  if (!fast_avail_elm_use(elm,value)) {
    *value=0;
    return 0;
  }

  /* Apply the bookings */
  if (elm->resource == FAST_RESOURCE_AVAILABLE_CPU ||
      elm->resource == FAST_RESOURCE_CURRENT_CPU ||
      elm->resource == FAST_RESOURCE_MEMORY) {    

    if (!fast_cbook_use(elm->resource,elm->ctime,hostSrc,value)) {
      *value=0;
      return 0;
    }
  }
  return 1;
}

/**
 * fast_avail_reset_cpu_measurements:
 * @host: the name of the host
 * 
 * Restart the measurements about the CPU on that host. Normally, you don't
 * have to use this function directly, since the booking mecanism will do it
 * for you when needed.
 */
void fast_avail_reset_cpu_measurements(const char *host) {
  int i;

  for (i=0; i<fast_avail_plugins_count; i++)
    if (fast_avail_plugins[i].restart_cpu)
      fast_avail_plugins[i].restart_cpu(host);
}

/*----[ fast_avail_elm_alloc ]-----------------------------------------------*/
/* Allocate a new element of the cache, and initialize the constant fields   */
/*---------------------------------------------------------------------------*/
static int fast_avail_elm_alloc(int resource,char **hosts,fast_avail_elm_t **whereto) {

  int i;
  
  if (!((*whereto)=(fast_avail_elm_t *)malloc(sizeof(fast_avail_elm_t))))
    FAIL("Out of memory\n");
  (*whereto)->ID=NULL;
  (*whereto)->resource=resource;
  (*whereto)->plugin=-1;
  (*whereto)->ctime=0;
  (*whereto)->ttl=0;
  (*whereto)->value=0;
  (*whereto)->lastused=0;
  (*whereto)->autofetched = 0;
  (*whereto)->plugin_data = NULL;

  /* search a plugin able to handle the request */
  for (i=0; i<fast_avail_plugins_count; i++)
    if (fast_avail_plugins[i].elm_init && 
	fast_avail_plugins[i].elm_init(resource,hosts,*whereto)) {
      (*whereto)->plugin = i;
      return 1;
    }

  fast_avail_elm_free(*whereto);
  if (fast_resource_nb_hosts[resource]==1) 
    FAIL2("Can't retrieve the value of %s on %s.\n",
	  fast_resource_name[resource],hosts[0]);
  else
    FAIL3("Can't retrieve the value of %s between %s and %s.\n",
	  fast_resource_name[resource],hosts[0],hosts[1]);
}


/*----[ fast_avail_elm_free ]------------------------------------------------*/
/* Reclaim space used by a cache element                                     */
/*---------------------------------------------------------------------------*/
static void fast_avail_elm_free(fast_avail_elm_t *elm) {
  if (elm) {
    if (elm->plugin_data &&
	fast_avail_plugins[ elm->plugin ].data_free)
      fast_avail_plugins[ elm->plugin ].data_free( elm->plugin_data );

    if (elm->ID)   free(elm->ID);
    if (elm->autofetched) {
      fast_cache_remove(fast_avail_cache_IDs,elm->ID);
      ERROR("FIXME: Autofetched series should be stopped on free\n");
    }
    free(elm);
  }
}

/*----[ fast_avail_elm_use ]-------------------------------------------------*/
/* Get the value from a cache element, and do all maintainance work on it    */
/*---------------------------------------------------------------------------*/
static int fast_avail_elm_use(fast_avail_elm_t *elm,double *value) {
  double now=time(NULL);

  if (fast_config_autofetch) {
    if (!elm->autofetched                                 && 
	fast_avail_plugins[ elm->plugin ].autofetch_start &&
	now - elm->lastused < fast_config_autofetch_on    &&
	!(elm->autofetched=fast_avail_plugins[ elm->plugin ].autofetch_start(elm) ))
	ERROR1("Can't start the autofetching of %s\n",elm->ID);
 
    elm->lastused=now;
  }
  
  if (elm->ctime + elm->ttl < now                  &&
      fast_avail_plugins[ elm->plugin ].refresh    &&
      !fast_avail_plugins[ elm->plugin ].refresh(elm)) {

    /* FIXME: no bombing on error */
    elm->value = elm->ctime = 0;
    return 0;
  }
    
  *value = elm->value;
  return 1;
}
