/* $Id: fast_booking.c,v 1.2 2003/07/09 14:12:49 mquinson Exp $ */

/* fast_booking - how scheduler inform FAST of its decision for reactivity*/

/* Copyright (C) 2001,2002,2003 the OURAGAN project.                      */
/* Code GPL'ed (see www.gnu.org).                                         */

/* Author(s): Martin Quinson                                              */

#include "fast_internals.h"

#define FAST_CBOOK_SIZE_INC 3

static void fast_cbook_coll_free(void *c);
static void fast_cbook_free(fast_cbook_t *book);
static fast_cache_t *fast_booking_cache=NULL;

/**
 * fast_cbook_create:
 * @host: involved host
 * @s: The peek space occupation of the booking
 * @t: How long should the execution take (or <0 if unknown)
 * @cpu_load: Cpu burden induced by this execution (<1 if routine is known 
 *  to not be cpu-bound, >1 for threaded or otherwise parallel routine)
 * @whereto: where to put the created booking
 * @Returns: true if ok, false on error
 *
 * Creates a new booking.
 */
int fast_cbook_create(const char* host,
		      int s, int t, double cpu_load,
		      /*OUT*/fast_cbook_t **whereto) {
  fast_cbook_coll_t *coll;

  /* Retrieve the collection in which we should add this */
  if (!fast_cache_retrieve(fast_booking_cache,host,(void**)&coll))
    FAIL("Internal error\n");

  /* Create the collection if needed */
  if (!coll) {
    if (!(coll=(fast_cbook_coll_t *)malloc(sizeof(fast_cbook_coll_t)))) 
      FAIL("Out of memory\n");
    if (!(coll->book=(fast_cbook_t**)malloc(sizeof(fast_cbook_t*)*
					    FAST_CBOOK_SIZE_INC))) {
      free(coll);
      FAIL("Out of memory\n");
    }
    coll->size=FAST_CBOOK_SIZE_INC;
    coll->used=0;
    if (!fast_cache_insert(&fast_booking_cache,strdup(host),
			   (void*)coll,&fast_cbook_coll_free)) {
      free(coll->book);
      free(coll);
      FAIL("Internal error\n");
    }
  }

  /* create the booking itself */
  if (!(*whereto=(fast_cbook_t *)malloc(sizeof(fast_cbook_t))))
    FAIL("Out of memory\n");

  if (!((*whereto)->host=strdup(host))) {
    free(*whereto);
    FAIL("Out of memory\n");
  }

  (*whereto)->space=s;
  (*whereto)->time=t;
  (*whereto)->cpu_load=cpu_load;
  (*whereto)->status=fast_cbook_created;

  (*whereto)->ctime=time(NULL);
  (*whereto)->atime=0;
  (*whereto)->etime=0;

  /* Add the booking to the collection */
  if (coll->size <= coll->used) {
    coll->size += FAST_CBOOK_SIZE_INC;
    if (!(coll->book=(fast_cbook_t**)realloc(coll->book,
					     sizeof(fast_cbook_t*)*coll->size))) {
      free((*whereto)->host);
      free(*whereto);
      coll->size=coll->used=0;
      FAIL("Out of memory\n");
    }
  }
  coll->book[ coll->used ++ ] = *whereto;

  return 1;
}


/**
 * fast_cbook_apply:
 * @booking: the booking to apply
 * @Returns: true if ok, false on error
 *
 * Inform FAST that this booking actually started.
 */
int fast_cbook_apply(fast_cbook_t **booking) {
  if (!booking || !*booking) return 0;
  fast_avail_reset_cpu_measurements((*booking)->host);
  (*booking)->status=fast_cbook_applied;
  (*booking)->atime=time(NULL);
  return 1;
}

/**
 * fast_cbook_cancel:
 * @booking: the booking to cancel
 * @Returns: true if ok, false on error
 *
 * Inform FAST that this booking will never be started (and free it).
 */
int fast_cbook_cancel(fast_cbook_t **booking) {
  if (!booking || !*booking) return 0;
  if ((*booking)->status != fast_cbook_created) {
    WARN("The canceled booking was already fast_cbook_apply'ed. Danger\n");
  }
  fast_cbook_free(*booking);
  *booking=NULL;
  return 1;
}

/**
 * fast_cbook_end:
 * @booking: the booking to end
 * @Returns: true if ok, false on error
 *
 * Inform FAST that this booking actually ended.
 */
int fast_cbook_end(fast_cbook_t **booking) {
  if (!booking || !*booking) return 0;
  fast_avail_reset_cpu_measurements((*booking)->host);

  (*booking)->status=fast_cbook_ended;
  (*booking)->etime=time(NULL);
  return 1;
}

/**
 * fast_cbook_use:
 * @resource: one of FAST_RESOURCE_* indicating the relevant resource
 * @measurement_time: When was the last measurement realized (to update the booking if needed)
 * @host: name of the target machine
 * @value: input: the uncorrected measured value. Output: the corrected value
 * @Returns: true if ok, false on error
 *
 * WARNING: This is an internal function, presented here for completness. 
 * It is used in fast_avail() to correct the value returned by NWS and Cie.
 *
 * Use the registered bookings to correct the given measurement
 */

int fast_cbook_use(int resource,double measurement_time,const char *host,
		   double *value) {
  fast_cbook_coll_t *coll;
  fast_cbook_t *book;
  double cpucount;
  int i;
  double additional_load=0;
  int corr_factor;

  /* Retrieve the right collection */
  if (!fast_cache_retrieve(fast_booking_cache,host,(void**)&coll))
    FAIL("Internal error\n");

  if (!coll || coll->used == 0) /* no booking */
    return 1;

  /* Get the number of proc on that host */
  if (!fast_avail(FAST_RESOURCE_CPU_COUNT,host,NULL,&cpucount))
    return 0;

  for (i=0; i<coll->used; i++) {
    book= coll->book[i];
    corr_factor = 0;

    /*
    LOG4("book->{status=%d;time=%d;space=%d;load=%f}\n",
	 book->status,book->time,book->space,book->cpu_load);
    */
    /** What's the effect of this booking (and update it) **/
      /* Correct in positif if not yet applied */
      if (book->status == fast_cbook_created)
	corr_factor=1;
      
      /* If applied, check if NWS did update. 
	 If yes, change the state. If not, correct + */
      if ( book->status == fast_cbook_applied) {
	if (book->atime > measurement_time)
	  corr_factor=1;
	else
	  book->status = fast_cbook_running;
      }

      /* Correct in negatif if running and not yet ended */
      if (book->status == fast_cbook_running &&
	  book->time > 0 &&
	  book->atime + book->time < time(NULL))
	corr_factor=-1;

      /* ended but NWS did not update yet */
      if (book->status == fast_cbook_ended) {
	if (book->etime > measurement_time)
	  corr_factor=-1;
	else { /* free this booking and make sure the loop survives */
	  fast_cbook_free(book);
	  i--;
	  break;
	}
      }
      /*
      LOG5("book->{status=%d,atime=%.0f;etime=%.0f};measurement_time=%.0f => corr=%d\n",
	   book->status,book->atime,book->etime,measurement_time,corr_factor);
      */
    /** Take this booking into account **/
    additional_load += corr_factor * 
      (resource == FAST_RESOURCE_MEMORY ? book->space : book->cpu_load);
  }
  LOG2("additional load due to bookings=%f; Uncorrected value=%f\n",
       additional_load,*value);

  /* modify the value according to the additional_load */
  if (resource == FAST_RESOURCE_MEMORY) {
    *value += additional_load;

  } else if (resource == FAST_RESOURCE_AVAILABLE_CPU) {
    if (*value) {  /* The value should never be 0, but make sure we don't divide by 0 */
      *value = cpucount / *value - 1.0; /* now, *value contain the load measured by NWS */
      *value += additional_load;
      if (*value<0) *value=0;
      *value = cpucount / (*value+1.0); /* Retransform the load to a AVAIL_CPU value */
    }
  } else if (resource == FAST_RESOURCE_CURRENT_CPU) {
    if (*value==cpucount || *value==0) {
      *value=0;
    } else {
      *value=cpucount / *value;
    }
    /* now, *value contain the load measured by NWS */
    *value += additional_load;
    if (*value<0) *value=0;
    /* Retransform the load to a CURR_CPU value */
    *value = *value < 1 ? cpucount : cpucount / *value;

  } else {
    FAIL1("A booking has no effect on resource #%d. Internal error\n",resource);
  }
	 
  return 1;
}

/*----[ Local functions ]----------------------------------------------------*/
static void fast_cbook_coll_free(void *c) {
  fast_cbook_coll_t *coll=(fast_cbook_coll_t *)c;
 
  if (coll) {
    while (coll->used) fast_cbook_free(coll->book[0]);
    if (coll->book) free(coll->book);
    free(coll);
  }
}

static void fast_cbook_free(fast_cbook_t *book) {
  fast_cbook_coll_t *coll;
  int i;

  /* Remove it from its collection */
  if (!book->host) {
    ERROR("Internal error: The booking has no host (leaking)...\n");
    return;
  }
  if (!fast_cache_retrieve(fast_booking_cache,book->host,(void**)&coll) ||
      !coll) {
    ERROR("Internal error: This host has no booking (leaking)...\n");
    return;
  }
      

  for (i=0; i<coll->used && coll->book[i] != book; i++);
  if (i == coll->used) {
    ERROR("Weird, the freed booking is not known.\n");
    return;
  }

  if (i<coll->used-1)
    memmove(&(coll->book[i+1]),&(coll->book[i]),coll->used-i-1);

  coll->used --;
  if (coll->used - coll->size > FAST_CBOOK_SIZE_INC + 1) {
    coll->size = coll->used;
    if (!(coll->book=(fast_cbook_t**)realloc(coll->book,
					     sizeof(fast_cbook_t*)*coll->size))) {
      coll->size=coll->used=0;
      ERROR("Out of memory while freeing. Isn't that funny?\n");
      return;
    }
  }

  /* Free it */
  free(book->host);
  free(book);
}
