/* $Id: internal_sensor.c,v 1.3 2003/06/04 15:30:28 mquinson Exp $ */

/* internal_sensor - the availability plugin giving the load of the current
                     host without any dependency on other packages */

/* Copyright (C) 2001 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include <ctype.h>      /* isspace() */
#include <errno.h>      /* errno */
#include <sys/time.h>   /* struct timeval */

#include "fast_avail_plugins.h"

#define UPTIMELOADFIELD 10
int fast_CPU_count(void);

/* from strutil */
int
fast_strntok(char *dest,
        const char *source,
        int len,
        const char *delim,
        const char **end);
#define GETTOK(dest,source,delim,end) fast_strntok(dest,source,sizeof(dest),delim,end)
#define GETWORD(dest,source,end) GETTOK(dest,source," \t\n",end)
int
fast_strntok(char *dest,
        const char *source,
        int len,
        const char *delim,
        const char **end) {

        char *last = dest + len - 1;

        /* Skip leading delimiters. */
        while( (*source != '\0') && (strchr(delim, *source) != NULL) ) {
                source++;
        }

        if (*source == '\0') 
	  return 0;

        while( (dest < last) && (*source != '\0') &&
                        (strchr(delim, *source) == NULL) ) {
                *dest++ = *source++;
        }

        if(end != NULL) 
	  *end = (const char *)source;
        
        *dest = '\0';

        return 1;
}

/* end of local strutil */

/*
** Module variables.  #cpus# and #period# cache the number of cpus on the
** system and the expected test frequency, respectively.  In order to query
** resources as infrequently as possible (to avoid loading the system and/or
** slowing the sensor), we cache the most recent measurements taken from each
** resource and, when called, attempt to use these in our computations.
*/
static int cpus;
static int period;

int fast_CPU_count() {
   
#ifdef HAVE_SYSCONF
#       ifdef _SC_CRAY_NCPU
#               define SYSCONF_PROCESSOR_COUNT_PARAMETER _SC_CRAY_NCPU
#       elif defined(_SC_NPROC_CONF)
#               define SYSCONF_PROCESSOR_COUNT_PARAMETER _SC_NPROC_CONF
#       elif defined(_SC_NPROCESSORS_CONF)
#               define SYSCONF_PROCESSOR_COUNT_PARAMETER _SC_NPROCESSORS_CONF
#       endif
#       ifdef SYSCONF_PROCESSOR_COUNT_PARAMETER 
  /* Try to gracefully handle mis-configuration. */
  int sysconfCount = sysconf(SYSCONF_PROCESSOR_COUNT_PARAMETER);
  return (sysconfCount == 0) ? 1 : sysconfCount;
#       else
  return 1;
#       endif
#else
  return 1;
#endif
}

static struct CacheInfo {
  double timeTaken;          /* timestamp of last measurement */
  double load;               /* measured load */
  double newFraction;        /* CPU time slicing a new process would get */
  double existingFraction;   /* CPU time slicing an existing process gets */
} fast_load_cache;

int fast_internal_sensor_init(void) {
   cpus = fast_CPU_count();
   memset(&fast_load_cache, 0, sizeof(fast_load_cache));
   return 1;
}

int fast_internal_sensor_elm_init(int resource,char **hosts,fast_avail_elm_t *elm) {
  if (resource == FAST_RESOURCE_AVAILABLE_CPU ||
      resource == FAST_RESOURCE_CURRENT_CPU ||
      resource == FAST_RESOURCE_CPU_COUNT) {

      elm->ttl=10;
      WARN("FIXME: Internal sensor should check for hostname\n");
      return 1;
   }
   return 0;
}

/* fast_load_measurement:
 *
 * Measure the load using a call to uptime
 */
static int
fast_load_measurement(void) {

  FILE *pipeToResource;
  char resourceOutput[255 + 1];
  char cpuWord[99 + 1];
  const char *curr;
  int field;

  if ((pipeToResource = popen(UPTIME_PATH, "r")) == NULL)
    FAIL1("PassiveCpuGetLoad: Unable to open pipe, errno = %d\n", errno);

  /*
  ** aix needs second shot at reading... gets ECHILD on the first one but
  ** works on the second.  This is only called if the first one fails, so
  ** there's no worry about loss of data and I think fgets is predictable
  ** enough to return NULL repeatedly when there's a real problem.
  */
  if( (fgets(resourceOutput, 255, pipeToResource) == NULL) &&
      (fgets(resourceOutput, 255, pipeToResource) == NULL) ) {
    pclose(pipeToResource);
    FAIL1("GetResourceLine: read failed; errno=%d\n",  errno);
  }

  pclose(pipeToResource);

  curr = &resourceOutput[0];

  for (field = 1; field < UPTIMELOADFIELD; field++) {
    GETWORD(cpuWord, curr, &curr);
  }
  if (cpuWord[0] == 'a') {
    /* Beginning of the word "average" -- occasionally we see an extra word */
    GETWORD(cpuWord, curr, &curr);
  }
  
  fast_load_cache.load = atof(cpuWord);
  fast_load_cache.timeTaken = time(NULL);

  fast_load_cache.newFraction = (float)cpus / (fast_load_cache.load + 1.0);
  fast_load_cache.existingFraction = (fast_load_cache.load < 1.0) ? (float)cpus :
    ((float)cpus / fast_load_cache.load);

  return(1);
}

int fast_internal_sensor_refresh(fast_avail_elm_t *elm) {
   
   if ((time(NULL) - fast_load_cache.timeTaken) > (double)period)
     if (!fast_load_measurement())
       return 1;
   
   if (elm->resource == FAST_RESOURCE_AVAILABLE_CPU) elm->value=fast_load_cache.newFraction;
   if (elm->resource == FAST_RESOURCE_CURRENT_CPU) elm->value=fast_load_cache.existingFraction;
   if (elm->resource == FAST_RESOURCE_CPU_COUNT) elm->value=cpus;

   elm->ttl= period;
   elm->ctime=fast_load_cache.timeTaken;
   
   return 1;
}

void fast_internal_sensor_restart_cpu(const char*host) {
  fast_load_cache.timeTaken = 0;
}
