/* $Id: nws.c,v 1.11 2005/03/05 20:59:59 mquinson Exp $ */

/* nws - the availability plugin giving access to NWS                        */

/* Copyright (C) 2001 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include "fast_avail_plugins.h"
#define NWS_API_COMPAT /* stick to 2.0 API since there is no need to change */
#include <nws_api.h>

#define NWS_DFLT_NS_PORT 8090 /* default nameserver port */
#define NWS_DFLT_FC_PORT 8070 /* default forecaster port */

#ifndef MAX
#define MAX(a,b) (a>b ? a : b)
#endif
/*====[ datatype for this plugin ]===========================================*/
typedef struct {
  NWSAPI_ForecastState *forecaster;
  double lastChecked;
} nws_data_t;

/*====[ Prototypes ]=========================================================*/
static char *fast_nws_starify_host(const char*host);
static inline int findhost_or_add(char ***list, int *count,
				  NWSAPI_Object seriesObject,
				  /*OUT*/int *where);
static int fast_nws_hostcmp(const char *host1,const char *host2);

/*----[ Configuration ]------------------------------------------------------*/
/* do we want to use NWS at all?*/
int fast_nws_config_use;
/* do we want raw results (if true) or forecasts (if false) */
int fast_nws_config_raw;

/*----[ fast_nws_resource_default_period ]-----------------------------------*/
/* default period for each resource. Should be the same than in              */
/*   NWS/Sensor/periodic.c (#define DEFAULT_PERIOD "10") and                 */
/*   NWS/Sensor/clique_protocol.c (#define DEFAULT_CLIQUE_PERIOD "120")      */
/*---------------------------------------------------------------------------*/
double fast_nws_resource_default_period[FAST_RESOURCE_TYPE_COUNT]={
  10,10,10,10, /* all resources about a unique host */
  120,120,     /* bandwidth and latency are between 2 hosts */
  3600};       /* cpuCount is in fact permanent */

/*----[ fast_nws_resource_name ]---------------------------------------------*/
/* The name of the several resources, in the order of the numerical macros   */
/*   defined in fast_nws_api.h (cf.).                                        */
/*---------------------------------------------------------------------------*/
const char *fast_nws_resource_name[FAST_RESOURCE_TYPE_COUNT]= {
  "availableCpu","currentCpu","freeMemory","freeDisk",
  "bandwidthTcp","latencyTcp","CpuCount"};

const char *fast_nws_skillName_of_resource[FAST_RESOURCE_TYPE_COUNT]= {
  "cpuMonitor","cpuMonitor","memoryMonitor","diskMonitor",
  "tcpMessageMonitor","tcpMessageMonitor",
  "cpuCountMonitor"};


/**
 * fast_nws_starify_host:
 *
 * Add a star at the end of an hostname (before the ':' separating the port)
 */
static char *fast_nws_starify_host(const char*host) {
  char *res=malloc(strlen(host)+2);
  const char *p=host;
  int i=0;
  int seen_star=0;

  if (!res) {
    ERROR("Malloc error\n");
      return NULL;
  }

  while (*p != ':' && *p != '\0') {     /* copy until the : or the end */
    if (*p == '*' && (*(p+1)==':' || *(p+1)=='\0')) 
      seen_star=1;
    res[i++]=*(p++);
  } 
  if (!seen_star) res[i++]='*';         /* add the star if not already there */
  while (*p != '\0') res[i++]=*(p++);   /* copy until the end */
  res[i] = '\0';                        /* ends the result */
  return res;
}

/*----[ findhost_or_add ]-----------------------------------------(private)--*/
/*                                                      (used in find_path)  */
/*                                                                           */
/* Search the ranking of #elem# in the #count#-long #list#.                  */
/* If not found, add it at strdup'ed copy at the end and do 's/:.*$//'.      */
/* In case of pb (memory allocation pb), returns -1                          */
/*---------------------------------------------------------------------------*/
static inline int findhost_or_add(char ***list, int *count,
				  NWSAPI_Object seriesObject,
				  /*OUT*/int *where) {
  int i;
  char hosta[512],hostb[512];
  NWSAPI_NwsAttribute currAttribute;   /* to travel in several attributes */
  char *str1, *str2;

  hosta[0]=hostb[0]='\0';

  NWSAPI_ForEachNwsAttribute(seriesObject, currAttribute) {
    DDEBUG2("Attr(%s)=%s\n",
	    NWSAPI_NwsAttributeName_r(currAttribute),
	    NWSAPI_NwsAttributeValue_r(currAttribute));
	    
    str1 = NWSAPI_NwsAttributeName_r(currAttribute);
    if(!strcmp(str1, NWSAPI_HOST_ATTR)) {
      str2 = NWSAPI_NwsAttributeValue_r(currAttribute);
      strcpy(hosta,str2);
      free(str2);
    }
    free(str1);
    str1 = NWSAPI_NwsAttributeName_r(currAttribute);
    if(!strcmp(str1, NWSAPI_TARGET_ATTR)) {
      str2 = NWSAPI_NwsAttributeValue_r(currAttribute);
      strcpy(hostb,str2);
      free(str2);
    }
    free(str1);
  }
  DDEBUG("\n");

  if (hosta[0]=='\0')
    FAIL("NWS forgot to tell me about the host name\n");
  if (hostb[0]=='\0')
    FAIL("NWS forgot to tell me about the target name\n");

  for (i=0; i<*count && fast_nws_hostcmp((*list)[i],hosta); i++);
  if (i==*count) {
    if (!((*list)=(char **)realloc(*list,sizeof(char**)*(*count + 1))))
      FAIL("Out of memory.\n");
    (*list)[(*count)++]=strdup(hosta);
  }
  where[0]=i;

  for (i=0; i<*count && fast_nws_hostcmp((*list)[i],hostb); i++);
  if (i==*count) {
    if (!((*list)=(char **)realloc(*list,sizeof(char**)*(*count + 1))))
      FAIL("Out of memory.\n");
    (*list)[(*count)++]=strdup(hostb);
  }
  where[1]=i;

  return 1;
}

/*----[ fast_nws_hostcmp ]---------------------------------------------------*/
/* Compares the name of two hosts (if on of them is a short name, until the  */
/*  first point, and if not on the whole string)                             */
/* The return value have the same semantic than strcmp.                      */
/*---------------------------------------------------------------------------*/
static int fast_nws_hostcmp(const char *host1,const char *host2)
{
  char *p;
  int point1,point2; /* position of the first point in #host[12]# */

  p=strchr(host1,'.');
  point1=MAX(0,p-host1);

  p=strchr(host2,'.');
  point2=MAX(0,p-host2);

  if (point1+point2==0 || (point1!=0 && point2!=0)) {
    return strcmp(host1,host2);
  } else {
    return strncmp(host1,host2,MAX(point1,point2));
  }
}

/*----[ fast_nws_init ]------------------------------------------------------*/
/* Initialize whatever should be in NWS using the info passed to fast_init() */
/*                                                                           */
/* returns if anything went well.                                            */
/*---------------------------------------------------------------------------*/
int fast_nws_init(void) {
  int ns_port;
  char *ns_name;
  char **hosts;
  int *ports, size;

  NWSAPI_HostSpec nsSpec;

  /* Use NWS ? */
  if (fast_config_getfirst_int("nws_use",&fast_nws_config_use) &&
      !fast_nws_config_use) {
    LOG("NWS is configured out.\n");
    return 1;
  } else {
    fast_nws_config_use=1;
  }

  /* Use raw measurements or forecasting? */
  if (!fast_config_getfirst_int("nws_raw",&fast_nws_config_raw) ||
      !fast_nws_config_raw) {
    LOG("NWS_init: Use forecasts\n");
    fast_nws_config_raw=0;
  } else {
    LOG("NWS_init: Use raw measurements\n");
    fast_nws_config_raw=1;
  }

  /* Which nameserver and forecaster? */
  if (!fast_config_getfirst_host("nws_nameserver",&ns_name,&ns_port) ||
      !fast_thash_get_host(fast_config,"nws_forecaster",&hosts,&ports,&size)) {
    FAIL("Can't get the configuration about NWS\n");
  }

  ns_port = (ns_port==0 ? NWS_DFLT_NS_PORT : ns_port); 

  if (!ns_name) {
    FAIL("NWS_init: Please give me a valid NWS nameserver name\n");
  }
  nsSpec = *NWSAPI_MakeHostSpec (strdup(ns_name),ns_port); 
  if (!NWSAPI_UseNameServer(&nsSpec)) {
    FAIL2("NWS_init: Unable to contact the NWS name server %s:%d\n",
	  nsSpec.machineName, nsSpec.machinePort);
  }
  LOG2("NWS_init: contact %s:%d as NWS nameserver\n",
       nsSpec.machineName, nsSpec.machinePort);

  if (size) { /* The nws_forecaster option was passed */
    LOG("NWS_init: a nws_forecast host was specified, but this feature is not supported by NWS anymore. Ignoring.\n");
  }

  return 1;
}
/*----[ fast_nws_elm_init ]--------------------------------------------------*/
/* Check if NWS can answer the measurement request for the given resource on */
/*    the given hosts (exprimed as hostname, not string representation of IP)*/
/*                                                                           */
/* If yes, search the serie name and period and return true                  */
/* If not (or in case of problem), return false                              */
/*---------------------------------------------------------------------------*/
int fast_nws_elm_init(int resource,char **hosts,fast_avail_elm_t *elm) {
  char buffer[5012]; 
  NWSAPI_ObjectSet seriesObjectSet; /* set of found series */ 
  NWSAPI_Object seriesObject;       /* to travel in the above set */ 
  //  NWSAPI_Attribute currAttribute;   / * to travel in several attributes */
  char *shost[2];

  int foundCount; /* how many elements found */

  nws_data_t *data;

  /* Create the forecaster struct if needed */
  if (!fast_nws_config_raw) {
    if (!(data=malloc(sizeof(data)))) {
      FAIL("Malloc error\n");
    }
    data->forecaster=NWSAPI_NewForecastState();
    data->lastChecked = 0;
    elm->plugin_data = data;
  }

  shost[0]=fast_nws_starify_host(hosts[0]);
  shost[1]=fast_resource_nb_hosts[resource] == 2 ? fast_nws_starify_host(hosts[1]) : NULL;

  /* *** [ Get the period ] *** */

  sprintf(buffer,"(&(objectclass=nwsActivity)(skillName=%s)(|(host=%s)(member=*%s*)))",
	  fast_nws_skillName_of_resource[resource],shost[0],shost[0]);
  if(!NWSAPI_GetObjects(buffer, &seriesObjectSet)) {
    ERROR("series retrieval failed\n");
    return 0;
  }
  foundCount = 0;
  elm->ttl=0;
  NWSAPI_ForEachObject(seriesObjectSet, seriesObject) {
    elm->ttl=atof(NWSAPI_AttributeValue(NWSAPI_FindNwsAttribute(seriesObject,
								NWSAPI_PERIOD_ATTR)));    
    foundCount++;
  }
  NWSAPI_FreeObjectSet(&seriesObjectSet);
  
  if (foundCount==0) {
    ERROR1("fast_nws_seriename(): No activity found.\nFilter was :'%s'\n",buffer);
    NWSAPI_FreeObjectSet(&seriesObjectSet);
    return 0;
  } else if (foundCount>1) {
    WARN("fast_nws_seriename: Got more than one activity for the period search. Will take only the last one.\n");
  }
  if (!elm->ttl) {
    elm->ttl=fast_nws_resource_default_period[resource];
    DDEBUG1("Period of the serie not found. %f used as default.\n",
	    elm->ttl);
  } 
  
  /* *** [ Get the seriename ] *** */

  sprintf(buffer,"(&(objectclass=nwsSeries)(resource=%s)(host=%s)",
	  fast_nws_resource_name[resource],shost[0]);
  if (fast_resource_nb_hosts[resource] == 2) {
    strcat (buffer,"(target=");
    strcat (buffer, shost[1]);
    strcat (buffer, "))");
  } else {
    strcat (buffer, ")");
  }

  free(shost[0]);
  if (shost[1]) free(shost[1]);

  /* ok, the filter is built. Let's use it for the search */
  if(!NWSAPI_GetObjects(buffer, &seriesObjectSet)) {
    ERROR("series retrieval failed\n");
    return 0;
  }

  foundCount = 0;  
  /* Test all the value returned, and return the first one which works */
  NWSAPI_ForEachObject(seriesObjectSet, seriesObject) {
    if (!elm->ID) {
      elm->ID=NWSAPI_NwsAttributeValue_r(NWSAPI_FindNwsAttribute(seriesObject,
								 NWSAPI_NAME_ATTR)
					);
    }
    foundCount++;
  }
  if (foundCount==0) {
    ERROR1("No NWS serie found.\nFilter was :'%s'\n",buffer);
    NWSAPI_FreeObjectSet(&seriesObjectSet);
    return 0;
  } else if (foundCount>1) {
    WARN("Got more than one NWS serie. Will ignore the subsequent ones.\n");
  }
  if (!elm->ID) {
    ERROR1("No NWS serie name found. Filter was :'%s'\n",buffer);
    NWSAPI_FreeObjectSet(&seriesObjectSet);
    return 0;
  }
  DDEBUG1("Query: %s\n",buffer);
  DDEBUG1("Serie name: %s\n",elm->ID);
  DDEBUG1("Serie period: %f\n",elm->ttl);
  
  NWSAPI_FreeObjectSet(&seriesObjectSet);

  return 1;
}

int fast_nws_refresh(fast_avail_elm_t *elm) {  
  if (elm->ID==NULL) {
    ERROR("Name of serie is null !! What should I search for ?\n");
    elm->ctime=0;
    return 0;
  }

  elm->value=elm->ctime=-1;
  DDEBUG1("seriesNames=%s\n",elm->ID);

  if (fast_nws_config_raw) {
    NWSAPI_Measurement measurement;
  
    if (!NWSAPI_GetMeasurement(elm->ID,&measurement))
      FAIL("Error while retrieving measurement\n");
    elm->value=measurement.measurement;
    elm->ctime=measurement.timeStamp;

    LOG3("NWS measurement for %s: %f (at %.0f)\n",elm->ID,elm->value,elm->ctime);


  } else {
    static NWSAPI_Measurement *measurement_list=NULL;
    int retrievedCount;
    nws_data_t *data=(nws_data_t*)elm->plugin_data;
    NWSAPI_ForecastCollection forecasted;


    if (!measurement_list)
      if (!(measurement_list=malloc(sizeof(NWSAPI_Measurement)*1000)))
	FAIL("Malloc error\n");
    
    if (!NWSAPI_GetMeasurements(elm->ID,data->lastChecked,measurement_list,1000,
				&retrievedCount))
      FAIL("Error while retrieving measurement to feed the forecaster\n");

    NWSAPI_UpdateForecastState(data->forecaster, measurement_list,retrievedCount,
			       &forecasted, 1);

    if (forecasted.forecasts[NWSAPI_MSE_FORECAST].forecast == 0.0) {
      ERROR1("Ignore broken forecasted value from retrieve %s (is very zero)\n",
	     elm->ID);
      return 1;
    }
    if (forecasted.forecasts[NWSAPI_MSE_FORECAST].forecast < 0.0) {
      ERROR1("Ignore broken forecasted value from retrieve %s (<0)\n",
	     elm->ID);
      return 1;
    }

    elm->value=forecasted.forecasts[NWSAPI_MSE_FORECAST].forecast;
    
    elm->ctime=forecasted.measurement.timeStamp;
    LOG3("NWS forecast for %s: %f (at %.0f)\n",elm->ID,elm->value,elm->ctime);
  }
  
  return 1;
}

int fast_nws_autofetch_start(fast_avail_elm_t *elm) {
  LOG1("Start autofetching of %s\n",elm->ID);
  return NWSAPI_AutoFetchBegin(elm->ID);
}
int fast_nws_autofetch_work(void) {
  char serieName[1024];
  NWSAPI_Measurement measurementCollection;
  fast_avail_elm_t *elm;

  while (NWSAPI_AutoFetchCheck(serieName,1024,&measurementCollection,-1)) {
    if (!fast_cache_retrieve(fast_avail_cache_IDs,serieName,(void**)&elm))
      FAIL("Internal error");
    if (!elm) {
      ERROR1("Got an autofetched measurement for an untracked serie '%s'.\n",
	     serieName);
    } else {
      if (fast_nws_config_raw) {
	elm->value = (double) measurementCollection.measurement;
	elm->ctime = (double) measurementCollection.timeStamp;
	LOG3("Autofetching '%s' (%.0f: %f).\n",serieName,elm->ctime,elm->value);
      } else {
	nws_data_t *data=(nws_data_t*)elm->plugin_data;
	NWSAPI_ForecastCollection forecast;
	
	NWSAPI_UpdateForecastState(data->forecaster, &measurementCollection,1,NULL,0);

	/* piece of fun: The return value of this function is inversed */
	if (NWSAPI_ComputeForecast(data->forecaster, &forecast)) {
	  ERROR1("Unable to get a forecast for %s. Leaving inchanged\n",elm->ID);
	  continue;
	}
	if (forecast.forecasts[NWSAPI_MSE_FORECAST].forecast == 0.0) {
	  ERROR1("Ignore broken forecasted value from autofetch %s (is very zero)\n",
		elm->ID); 
	  continue;
	}
	if (forecast.forecasts[NWSAPI_MSE_FORECAST].forecast < 0.0) {
	  ERROR1("Ignore broken forecasted value from autofetch %s (<0)\n",
		 elm->ID);
	  continue;
	}
	
        elm->value = forecast.forecasts[NWSAPI_MSE_FORECAST].forecast;
	elm->ctime=(double)measurementCollection.timeStamp;
	LOG4("Autofetching+forecasting '%s' (%.0f: m=%.4f f=%.4f)\n",
	     elm->ID,elm->ctime,(double) measurementCollection.measurement,elm->value);
      }
      if (time(NULL) - elm->lastused > fast_config_autofetch_off) {
	LOG1("Serie %s unused since too long. Stop autofetching\n",elm->ID);
	NWSAPI_AutoFetchEnd(elm->ID);
      }
    }
  }
  return 1;
}

/*----[ fast_nws_edges_list ]------------------------------------------------*/
/* Extract the list of all direct test for the given resource.               */
/* The output is a pointer to a locally allocated array of strings. The dims */
/*  of this array is (2 * count).                                            */
/* To acces to the first host of the nth edge, look at (*edges_list)[2*n].   */
/*  Add 1 to the index to find the second host                               */
/*                                                                           */
/* This function will fail awfully if NWS returns more than 2 hosts per serie*/
/*                                                                           */
/* Returns if ok.                                                            */
/*---------------------------------------------------------------------------*/
int fast_nws_edges_list(int resource,
			int *vertex_count, char ***vertex_list,
			int *edges_count, int **edges_list) {

  char buffer[2048];
  NWSAPI_ObjectSet seriesObjectSet; /* set of found series */ 
  NWSAPI_Object seriesObject;       /* to travel in the above set */ 
  int i;

  /*construct a request*/
  sprintf(buffer,"(&(objectclass=nwsSeries)(resource=%s))",
	  fast_nws_resource_name[resource]);

  /* Do the search */
    
  if(!NWSAPI_GetObjects(buffer, &seriesObjectSet)) {
    FAIL("fast_nws_edges_list(): series retrieval failed\n");
  }

  /* change the result of the search to my response */  
  *edges_count = *vertex_count = 0;
  *edges_list = NULL;
  *vertex_list = NULL;

  NWSAPI_ForEachObject(seriesObjectSet, seriesObject) {

    (*edges_count)++;
    if (!(*edges_list=(int *)realloc((*edges_list),
				       (sizeof(int)*2*(*edges_count)))) ){
      FAIL("fast_nws_edges_list(): Out of memory\n");
    }

    if (!findhost_or_add(vertex_list,vertex_count,seriesObject,
			 &((*edges_list)[2 * ((*edges_count)-1)]))) {
      free(*vertex_list);
      free(*edges_list);
      NWSAPI_FreeObjectSet(&seriesObjectSet);
      FAIL("Internal error");
    }
  }

/*    DDEBUG("\n");DDEBUG1("List of vertex (count=%d):\n",*vertex_count); */
/*    for (i=0;i<*vertex_count;i++) */
/*      DDEBUG1("\t%s\n",(*vertex_list)[i]); */
/*    DDEBUG("\n"); */

   /*
  LOG1("List of edges (count=%d):\n",*edges_count);
   for (i=0;i<*edges_count;i++)  
     LOG2(" %d - %d\n", 
	     (*edges_list)[2 * i], 
	     (*edges_list)[2 * i + 1]); 
   LOG("\n");
    */
  DDEBUG2("vertex_count=%d, edges_count=%d\n",*vertex_count,*edges_count);
  for (i=0;i<*edges_count;i++) 
    DDEBUG2(" %s - %s\n",
	    (*vertex_list)[(*edges_list)[2 * i]],
	    (*vertex_list)[(*edges_list)[2 * i + 1]]);

  return 1;
}

void fast_nws_restart_cpu(const char*host) {
  char buffer[1024];
  NWSAPI_ObjectSet objectSet; /* set of found hosts */ 
  NWSAPI_Object object;       /* to travel in the above set */ 
  int port;
  NWSAPI_HostSpec hostspec;

  /* Restart the forecasters of all series about that host */
  if (!fast_nws_config_raw) {
    fast_avail_elm_t *elm;
    nws_data_t *data;
    int resource;

    for (resource=0; resource<FAST_RESOURCE_TYPE_COUNT; resource++) {
      if (resource != FAST_RESOURCE_AVAILABLE_CPU &&
	  resource != FAST_RESOURCE_CURRENT_CPU &&
	  resource != FAST_RESOURCE_MEMORY)
	continue;

      if (!fast_cache_retrieve(fast_avail_cache[resource], host, (void*)&elm))
	continue;

      if (!elm || 
	  elm->plugin != 0/*FIXME: Hardcoding NWS plugin number is a bad idea*/) 
	continue;

      data=(nws_data_t*)elm->plugin_data;
      
      NWSAPI_FreeForecastState(&data->forecaster);
      data->forecaster=NWSAPI_NewForecastState();
      /* Make sure that we never feed the forecaster with data produced before 
	 this point */
      data->lastChecked=time(NULL);
    }
  }

  /* search the hosts */
  sprintf(buffer,"(&(hostType=sensor)(name=%s*))",host);

  LOG1("Restart sensors on %s\n",host);

  if(!NWSAPI_GetObjects(buffer, &objectSet)) {
    ERROR("Host list retrieval failed\n");
    return;
  }

  /* restart first found host */
  NWSAPI_ForEachObject(objectSet, object) {
    char *str =NWSAPI_NwsAttributeValue_r(
		  NWSAPI_FindNwsAttribute(object, NWSAPI_PORT_ATTR));
    port=atoi(str);
    free(str);
    hostspec=*NWSAPI_MakeHostSpec(host,port);
    NWSAPI_RestartActivities(&hostspec);
    break;
  }
  NWSAPI_FreeObjectSet(&objectSet);

}
