/* $Id: pathfinder.c,v 1.4 2004/02/05 18:24:54 mquinson Exp $ */

/* pathplug - the availability plugin deducting the network values of a path 
              from the ones of each part */

/* Copyright (C) 2001 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include "fast_avail_plugins.h"

/*====[ local prototypes ]===================================================*/
int fast_pathfinder(int resource,const char *begin,const char *end,
		    /*OUT*/int *count, char ***hosts);

/*====[ datatype for this plugin ]===========================================*/
typedef struct {
  fast_avail_elm_t **path;
  int path_len;
} pathfinder_data_t;

/*====[ Code ]===============================================================*/

int fast_pathfinder_elm_init(int resource,char **hosts,fast_avail_elm_t *elm) {
  char **steps;       /* hosts encountered on the path */
  int step_count;     /* len of previous array */ 
  double dummyvalue;  /* a useless value needed by a prototype */
  char **keys=NULL;   /* to read in fast_avail_cache */
  pathfinder_data_t *data=NULL;
  int i;

  if (!(fast_pathfinder(resource,hosts[0],hosts[1],&step_count,&steps)))
    return 0;

  if (!(data=(pathfinder_data_t*)malloc(sizeof(pathfinder_data_t))) ||
      !(keys=(char**)malloc(sizeof(char*)*2))  ||
      !(data->path=(fast_avail_elm_t**)malloc(sizeof(fast_avail_elm_t*)*
					      (step_count-1)))) {
    if (data) free(data);
    if (keys) free(keys);
    FAIL1("Out of memory for step_count=%d\n",step_count);
  }

  data->path_len=step_count-1;
  
  /* Make sure all element of the path are created, and store them in our table */
  for (i=0 ; i<step_count-1 ; i++) {
    
    keys[0]=steps[i];
    keys[1]=steps[i+1];
    
    if (!fast_avail(resource,keys[0],keys[1],&dummyvalue)) {
      ERROR2("path_alloc(): Can't find the element corresponding to %s-%s\n",
	     keys[0],keys[1]);
      free(keys);
      free(data->path);
      free(data);
      return 0;
    }
    if (!fast_multicache_retrieve(fast_avail_cache[resource],2,(const char **)keys,
				  (void*)&( data->path[ i ] ))) {
      free(keys);
      free(data->path);
      free(data);
      FAIL("Internal error\n");
    }
  }

  elm->plugin_data = data;
  return 1;
}

int fast_pathfinder_refresh(fast_avail_elm_t *elm) {
  int i;
  pathfinder_data_t *data=(pathfinder_data_t *)elm->plugin_data;

  if (elm->ctime + elm->ttl < time(NULL)) {
    elm->value=(elm->resource==FAST_RESOURCE_BANDWIDTH ? -1 : 0);
    elm->ctime=elm->ttl=-1;
    for (i=0; i < data->path_len; i++) {      
      /* update the value of the path */
      if (elm->resource==FAST_RESOURCE_BANDWIDTH) {
	if (elm->value==-1 || data->path[i]->value < elm->value) {
	  elm->value = data->path[i]->value;
	}
      } else if (elm->resource==FAST_RESOURCE_LATENCY) {
	elm->value += data->path[i]->value;
      } else {
	FAIL("Found a resource taking 2 hosts other than bw/lat. WTF?.\n");
      }
      
      /* update the creation time and the ttl of the path */
      
      if ((elm->ctime == -1 || data->path[i]->ctime < elm->ctime))
	elm->ctime = data->path[i]->ctime;
      if ((elm->ttl == -1 || data->path[i]->ttl < elm->ttl))
	elm->ttl = data->path[i]->ttl;
    }
  }
  return 1;
}

void fast_pathfinder_data_free(void *d) {
  pathfinder_data_t *data=(pathfinder_data_t *)d;

  if (data) {
    if (data->path) free(data->path);
    free(data);
  }
}


/* ***************************************************************************
 * The pathfinder algoritm itself
 * ***************************************************************************/

/*----[ fast_pathfinder_t ]--------------------------------------------------*/
/* A struct describing a path currently builded.                             */
typedef struct fast_pathfinder_s fast_pathfinder_t;
struct fast_pathfinder_s {
  int *hosts;
  fast_pathfinder_t *next;
};

static int fast_pathfinder_alloc(fast_pathfinder_t **head,
				 fast_pathfinder_t *copy_from,
				 int new_host, int length);
static void fast_pathfinder_free(fast_pathfinder_t *path);
static int fast_pathfinder_hostcmp(const char *host1,const char *host2);

/*----[ fast_pathfinder_alloc ]-----------------------------------(private)--*/
/* Allocate a new path struct derivated from #copy_from# and put it before   */
/* #head#. The derivation consists of adding #new_host# at the end of the    */
/*  list, which is #length#-long before the operation.                       */
/* If #length# is null, #copy_from# could too.                               */
/*---------------------------------------------------------------------------*/
static int fast_pathfinder_alloc(fast_pathfinder_t **head,
				 fast_pathfinder_t *copy_from,
				 int new_host, int length)
{
  fast_pathfinder_t *res;

  if (!(res=(fast_pathfinder_t*)malloc(sizeof(fast_pathfinder_t)))) {
    FAIL("fast_pathfinder_alloc(): Out of memory\n");
  }
  if (!(res->hosts=(int*)malloc(sizeof(int)*(length+1)))) {
    free(res);
    FAIL("Out of memory\n");
  }
  res->next=*head;
  if (copy_from)
    memcpy(res->hosts,copy_from->hosts,sizeof(int)*length);
  res->hosts[length]=new_host;
  *head=res;

  return 1;
}

/*----[ fast_pathfinder_free ]------------------------------------(private)--*/
/* Free a path linked list                                                   */
/*---------------------------------------------------------------------------*/
static void fast_pathfinder_free(fast_pathfinder_t *path)
{
  if (path) {
    free(path->hosts);
    free(path);
  }
}

/*----[ fast_pathfinder_hostcmp ]--------------------------------------------*/
/* Compares the name of two hosts (if on of them is a short name, until the  */
/*  first point, and if not on the whole string)                             */
/* The return value have the same semantic than strcmp.                      */
/*---------------------------------------------------------------------------*/
int fast_pathfinder_hostcmp(const char *host1,const char *host2)
{
  char *p,*column;
  int point1,point2; /* position of the first point in #host[12]# */
  int port1=0,port2=0;

  /* Compare the ports */
  column=strchr(host1,':');
  if (column++) {
    port1=atoi(column);
  } 

  column=strchr(host2,':');
  if (column++) {
    port2=atoi(column);
  }
  if (port1 && port2) {
    if (port1 < port2) return -1; 
    if (port1 > port2) return 1; 
  } else if (port1 || port2) return 1;

  /* Compare the hostnames */
  p=strchr(host1,'.');
  point1= p ? p-host1 : 0;

  p=strchr(host2,'.');
  point2= p ? p-host2 : 0;

  if (point1+point2==0 || (point1!=0 && point2!=0)) {
    return strcmp(host1,host2);
  } else {
    return strncmp(host1,host2,MAX(point1,point2));
  }
}

/*----[ fast_pathfinder ]----------------------------------------------------*/
/* Search the shortest path between #begin# and #end# in the direct          */
/*   measurements of the various availability plugins (read: of NWS).        */
/*                                                                           */
/* This is a classic greedy algorithm with this modification:                */
/*  the classical algorithm (dijiska) handle the fact of paths converging    */
/*  on one node. It has 3 colors: node seen at previous step, node seen this */
/*  step, node never seen. In the case where a node is seen twice at the same*/
/*  step, its value is the minimum of the values observed. But I do not want */
/*  to compute the value of nodes here to not pay the overhead of a call to  */
/*  NWS, so I use only 2 colors and take only the first path found,          */
/*  discarding any others.                                                   */
/*                                                                           */
/* Returns if all went ok or not. If no path where found, returns false.     */
/*---------------------------------------------------------------------------*/
int fast_pathfinder(int resource,const char *begin,const char *end,
		    /*OUT*/int *count, char ***hosts) {
  int vertex_count;  /* These two are the list of all hosts involved in a */
  char **vertex_list;/*  direct tests in NWS for this resource */

  int *vertex_color; /* 0: never seen this vertex.
                        1: seen this vertex at previous step */

  int edges_count;   /* these two are the list of all the existing direct */
  int   *edges_list; /*  tests in NWS for this resource.                  */

  fast_pathfinder_t *paths=NULL,*new_paths,*curr_path=NULL;

  int length=0; /* length of path already found. it is the number of the step
                   we are actually doing, too. */

  int found = 0; /* true if we've found the searched path */
  int neibor; /* id of the neibor studied */
  int i,begin_id=-1,end_id=-1;

  int errcode=0;
  *count=0;

  /* get the list of all edges and vertexes from the plugins */
  vertex_count=edges_count=0;
  vertex_list=NULL;
  edges_list=NULL;
  for (i=0; i<fast_avail_plugins_count; i++) {
    /* local data to accumulate the answers of all plugins */
    int vcount;
    char **vlist;
    int ecount;
    int *elist;
    if (fast_avail_plugins[i].edges_list && 
	fast_avail_plugins[i].edges_list(resource,&vcount,&vlist,&ecount,&elist)) {
      LOG3("Got %d vertices and %d edges from plugin %s\n",vcount,ecount,fast_avail_plugins[i].name);
      /* increase the tables to accept the new results */
      if (vertex_count) {
	vertex_list=(char**)realloc(vertex_list,sizeof(char*) * (vcount+vertex_count));
      } else {
	vertex_list=(char**)malloc(sizeof(char*) * (vcount));
      }
      if (!vertex_list) {
	ERROR("Malloc error\n");
	if (vlist) free(vlist);
	if (elist) free(elist);
	goto end;
      }

      if (edges_count) {
	edges_list=(int*)realloc(edges_list,sizeof(int) * (ecount+edges_count));
      } else {
	edges_list=(int*)malloc(sizeof(int) *2* (ecount));
      }
      if (!edges_list) {
	ERROR("Malloc error\n");
	if (vlist) free(vlist);
	if (elist) free(elist);
	goto end;
      }
      /* add the results to the new ones */
      memcpy(&(vertex_list[vertex_count]),&(vlist[0]),sizeof(vlist[0]) * vcount);
      memcpy(&(edges_list[edges_count]),&(elist[0]),sizeof(elist[0]) *2* ecount);

      vertex_count += vcount;
      edges_count += ecount;
    }
  }

  /* make sure there is at least a measurement concerning each side of the link */
  for (i=0 ; i<vertex_count ; i++) {
    if (!fast_pathfinder_hostcmp(begin,vertex_list[i])) begin_id=i;
    if (!fast_pathfinder_hostcmp(end,vertex_list[i]))   end_id=i;
  }
  if (begin_id==-1 || end_id==-1) {
    ERROR2("Pathfinder: Unable to find a path between %s and %s\n",begin,end);
    goto end;
  }
  /* other initialisations */
  if (!(vertex_color=malloc(sizeof(int)*vertex_count))) {
    ERROR("Out of memory\n");
    goto end;
  }
  memset(vertex_color,0,sizeof(int)*vertex_count);

  /* Debug output */
   LOG1("Existing links (count=%d):\n",edges_count);
   for (i=0; i<edges_count; i++) {
      LOG4(" [%d]%s - [%d]%s\n",
	   edges_list[2 * i],     vertex_list[ edges_list[2 * i] ],
	   edges_list[2 * i + 1], vertex_list[ edges_list[2 * i + 1] ]);
   }
   
  if (!(fast_pathfinder_alloc(&paths,NULL,begin_id,length))) {
    goto end;
  }
  paths->next=NULL;
  vertex_color[begin_id]=1;
  

  while (!found && paths) { 
    new_paths=NULL;
    DDEBUG1("Length=%d\n",length);
    /* foreach paths in the preceding step */
    for (curr_path=paths ; 
         !found && curr_path ; 
         curr_path = curr_path->next, length++) {
      int j;
      DDEBUG("Consider path ");
      for (j=0;j<=length;j++) 
        printf("%s%s",j?"->":"",vertex_list[curr_path->hosts[j]]);
      printf ("\n");

      /* foreach member of the neiborhood of the last host of this path */
      for (i=0 ; !found && i<vertex_count ; i++) {
        neibor=-1;
        if (curr_path->hosts[length]==edges_list[2*i])
          neibor=edges_list[2*i + 1];
        if (curr_path->hosts[length]==edges_list[2*i + 1])
          neibor=edges_list[2*i];
        if (neibor != -1 && vertex_color[neibor]==0) {
          DDEBUG1("  + Add %s\n",vertex_list[neibor]);

          found = found || (neibor == end_id);
          vertex_color[neibor]=1;
          if (!(fast_pathfinder_alloc(&new_paths,curr_path,neibor,length+1))) {
	    goto end;
          }
          DDEBUG("  -> New path: ");
          for (j=0;j<=length+1;j++) 
            printf("%s%s",j?"->":"",vertex_list[new_paths->hosts[j]]);
          printf ("\n");

        }
      }
    }
    fast_pathfinder_free(paths);
    paths=new_paths;
  }
  if (found) {
    *count=length+1;
    if (!(*hosts=(char**)malloc(sizeof(char**)*(length+1)))) {
      FAIL("Out of memory.\n");
    }
    printf("Path found: ");
    for (i=0; i<*count;i++) {
      (*hosts)[i]=strdup(vertex_list[paths->hosts[i]]);
      printf("%s%s",(i?" -> ":""),(*hosts)[i]);
    }
    printf("\n");
    errcode=1;
     
  } else {
    *count=-1;
    *hosts=NULL;
    ERROR2("fast_nws_find_path(): No path found between %s and %s.\n",
             begin,end);
    if (!strcmp(begin,end)) {
      WARN("fast_nws_find_path(): The source and destination hosts are the same.\n"
	   "  It can't work if you have only one sensor on this machine.\n"
	   "  Ie, NWS can't give you the bandwidth between a sensor and itself,\n"
	   "  but can give you the bandwidth between a machine and itself it there\n"
	   "  is two sensors on this machine.\n");
    }
    errcode = 0;
  }

 end:
  if (edges_list)  free(edges_list);
  if (vertex_list) {
    for (i=0 ; i<*count ; i++) free(vertex_list[i]);
    free(vertex_list);
  }
  if (vertex_color) free(vertex_color);
  fast_pathfinder_free(paths);
    
  return errcode;

}
