/* $Id: data_arg.c,v 1.13 2003/08/09 17:27:26 pcombes Exp $ */

/* data_arg - How to describe one routine argument                        */

/* Copyright (C) 2003 the OURAGAN project.                                */
/* Code available under BSD-like licence (see COPYRIGHT file).            */
/* Author(s): Martin Quinson                                              */

#include "fast_internals.h"
fast_thash_t *data_arg_properties_get(data_arg_t *da);

fast_thash_t *data_registered_properties[data_scal_type_count][data_cons_type_count];

struct data_arg_s {
  data_scal_type_t scal;
  data_cons_type_t cons;
  fast_thash_t *properties;
  void *value; /* NULL for complex types, and defined for scalars */
};

/* Private prototypes */
static void data_arg_init (void);
static data_arg_t *data_arg_new_vargs(data_scal_type_t scal, data_cons_type_t cons, va_list pa);

/**
 * data_arg_init:
 *
 * Initialize this module (automatically called).
 */
static void data_arg_init (void) {
  int i,j;

  for (i=0; i<data_scal_type_count; i++) 
    for (j=0; j<data_cons_type_count; j++) 
      fast_thash_init(&(data_registered_properties[i][j]));

  for (i=0; i<data_scal_type_count; i++) {

    /* Scalar: value:1_to_1_<type of this scalar> */

    /* Vectors: size:1_to_1_int */
    fast_thash_register(&(data_registered_properties[i][data_cons_vect]),
			"size",fast_thc_int,1,1);
    
    /* Matrices: rows:1_to_1_int cols:1_to_1_int row_major:0_to_1_int */
    fast_thash_register(&(data_registered_properties[i][data_cons_mat]),
			"rows",fast_thc_int,1,1);
    fast_thash_register(&(data_registered_properties[i][data_cons_mat]),
			"cols",fast_thc_int,1,1);
    fast_thash_register(&(data_registered_properties[i][data_cons_mat]),
			"row_major",fast_thc_int,0,1);
    
    /* Files: size:0_to_1_int path:1_to_1_string */
    fast_thash_register(&(data_registered_properties[i][data_cons_file]),
			"size",fast_thc_int,0,1);
    fast_thash_register(&(data_registered_properties[i][data_cons_file]),
			"path",fast_thc_string,1,1);
    
  }
  fast_thash_register(&(data_registered_properties[data_char][data_cons_vect]),
		      "value",fast_thc_string,0,1);

  fast_thash_register(&(data_registered_properties[data_char][data_cons_scal]),
		      "value",fast_thc_int,1,1);    
  fast_thash_register(&(data_registered_properties[data_int][data_cons_scal]),
		      "value",fast_thc_int,1,1);    
  fast_thash_register(&(data_registered_properties[data_double][data_cons_scal]),
		      "value",fast_thc_double,1,1);    
}

/**
 * data_arg_new_vargs:
 * @scal: scalar type of the new data_arg to create
 * @cons: its constructor type
 * @pa: A NULL terminated list of properties of form {key; val}
 *
 * Creates a new data_arg.
 */
data_arg_t *data_arg_new_vargs(data_scal_type_t scal, data_cons_type_t cons, va_list pa) {

  data_arg_t *res;
  static int inited=0;

  if (!inited) {
    inited=1;
    data_arg_init();
  }
  
  if (!(res=(data_arg_t*)malloc(sizeof(data_arg_t)))) 
    FAIL("Out of memory\n");
  
  if (!(fast_thash_new_from(&(res->properties),data_registered_properties[scal][cons])))
    return NULL;

  res->scal=scal;
  res->cons=cons;
  if (!(fast_thash_set_vargs(res->properties,pa))) {
    data_arg_free(&res);
    return NULL;
  }
  
  return res;
}

/**
 * data_arg_new:
 * @scal: scalar type of the new data_arg to create
 * @cons: its constructor type
 * @Vargs: 
 *
 * Creates a new data_arg.
 */
data_arg_t *data_arg_new(data_scal_type_t scal, data_cons_type_t cons, ...) {
  data_arg_t *res;
  va_list pa;

  va_start(pa,cons);
  res=data_arg_new_vargs(scal,cons,pa);
  va_end(pa);
  return res;
}

/**
 * data_arg_free:
 * @arg: the data_arg to destroy
 * 
 * Destroys a data_arg_t.
 */
void data_arg_free(data_arg_t **arg) {
  if (*arg) {
    fast_thash_free(&(*arg)->properties);
    free(*arg);
    *arg=NULL;
  }
}

/**
 * data_arg_check:
 * @arg: the arg to check
 * @Returns: true if ok.
 *
 * Enforce the min and max of each property of this argument.
 */
int data_arg_check(data_arg_t *arg) {
  return (arg) ? fast_thash_check(arg->properties) : 0;
}

/**
 * data_arg_dump:
 * @da: What to dump
 * @name: 
 * @indent: The string to put before each printed line (for indentation for example).
 *
 * Dumps an argument for debugging purposes.
 */
void data_arg_dump(data_arg_t *da,const char *name,const char *indent) {
  const char *base,*cons;
  char subindent[64];

  if (da) {
    if (name)
      printf("%sDumping the argument '%s'\n",indent,name);
    else
      printf("%sDumping the argument\n",indent);
    switch (da->scal) {
    case data_char: base="char"; break;
    case data_int: base="int"; break;
    case data_double: base="double"; break;
    default: base="unknown";
    }
    switch (da->cons) {
    case data_cons_scal: cons="scalar"; break;
    case data_cons_vect: cons="vector"; break;
    case data_cons_mat: cons="matrix"; break;
    case data_cons_file: cons="file"; break;
    default: cons="unknown";
    }
    printf("%s  base type=%d(%s); cons type=%d(%s); properties=%p\n",
	   indent,da->scal,base,da->cons,cons,da->properties);

    snprintf(subindent,64,"%s  ",indent);
    fast_thash_dump(NULL,subindent,da->properties);

  } else {
    if (name)
      printf("%sThe argument '%s' is NULL\n", indent, name);
    else
      printf("%sThis argument is NULL\n",indent);
  }
  fflush(stdout);
}


/**
 * data_arg_char_new:
 * @value: the char used to create the new data_arg
 * @Returns: the data_arg newly created
 * 
 * creates a new data_arg_t describing a char of the given value
 */

data_arg_t *data_arg_char_new(char value) {
  return data_arg_new(data_char,data_cons_scal, "value", value, NULL);
}
/**
 * data_arg_int_new:
 * @value: the integer
 * @Returns: the new data_arg_t
 * 
 * creates a new data_arg_t describing an interger of the given @value
 */
data_arg_t *data_arg_int_new(int value) {
  return data_arg_new(data_int,data_cons_scal, "value", value, NULL);
}

/**
 * data_arg_double_new:
 * @value: the double
 * @Returns: the new data_arg_t
 * 
 * creates a new data_arg_t describing a double of the given value
 */
data_arg_t *data_arg_double_new(double value) {
  return data_arg_new(data_double,data_cons_scal, "value", value, NULL);
}

/**
 * data_arg_vect_new:
 * @type: the type of the vector
 * @size: the size of the vector
 * @Returns: the new data_arg_t
 * 
 * creates a new data_arg_t describing a vector of the given @size
 */

data_arg_t *data_arg_vect_new(data_scal_type_t type,size_t size) {
  return data_arg_new(type,data_cons_vect, "size", size, NULL);
}
  
/**
 * data_arg_string_new:
 * @val: the string
 * @Returns: the new data_arg_t
 * 
 * creates a new data_arg_t describing a string of the given @val
 */

data_arg_t *data_arg_string_new(const char *val){
  return data_arg_new(data_char,data_cons_vect, "size", strlen(val), "value", val, NULL);
}

/**
 * data_arg_mat_new:
 * @type: the type of the matrix
 * @rows: the rows
 * @cols: the cols
 * @Returns: the new data_arg_t
 * 
 * creates a new data_arg_t describing a matrix of the given sizes
 */

data_arg_t *data_arg_mat_new(data_scal_type_t type, 
			     size_t rows, size_t cols){
  return data_arg_new(type,data_cons_mat, "rows",rows, "cols",cols, NULL);
}

/**
 * data_arg_file_new:
 * @path: path to the file descriptor
 * @Returns: the new data_arg_t
 * 
 * creates a new data_arg_t describing the given path
 */

data_arg_t *data_arg_file_new(const char *path) {
  return data_arg_new(data_char,data_cons_file, "path",path, NULL);
}

/**
 * data_arg_pset_int:
 * @da: the data_arg
 * @key: the key of the value
 * @val: the value to be set
 * @Returns: True if everything went ok, false if an error occurs
 *
 * set properties in place for an integer
 */

int data_arg_pset_int   (data_arg_t *da, const char *key, int val) {
  return fast_thash_set_int((fast_thash_t*)da,key,val);
}

/**
 * data_arg_pset_double:
 * @da: the data_arg
 * @key: the key of the value
 * @val: the value to be set
 * @Returns: True if everything went ok, false if an error occurs
 *
 * set properties in place for a double
 */

int data_arg_pset_double(data_arg_t *da, const char *key, double val) {
  return fast_thash_set_double((fast_thash_t*)da,key,val);
}

/**
 * data_arg_pset_string:
 * @da: the data_arg
 * @key: the key of the value
 * @val: the value to be set
 * @Returns: True if everything went ok, false if an error occurs
 *
 * set properties in place for a string
 */

int data_arg_pset_string(data_arg_t *da, const char *key, const char *val) {
  return fast_thash_set_string((fast_thash_t*)da,key,val);
}
/**
 * data_arg_scaltype_get:
 * @da: the data_arg
 * @Returns: the data_arg of scal
 *
 * Return the type of data_arg 
 */ 

data_scal_type_t data_arg_scaltype_get(data_arg_t *da) {
  if (da) {
    return da->scal;
  } else {
    return data_scal_type_count;
  }
}

/**
 * data_arg_constype_get:
 * @da: the data_arg
 * @Returns: The data_arg of cons
 *
 * Return the type of data_arg
 */ 

data_cons_type_t data_arg_constype_get(data_arg_t *da){
  if (da) {
    return da->cons;
  } else {
    return data_cons_type_count;
  }
}

/**
 * data_arg_properties_get:
 * @da: The data_arg
 * @Returns: The properties
 *
 * Get the typed hash containing the properties of this argument
 */
fast_thash_t *data_arg_properties_get(data_arg_t *da) {
  return da ? da->properties : NULL;
}
