/* $Id: data_argstack.c,v 1.15 2003/09/05 14:30:27 pcombes Exp $ */

/* data_description - How to describe data                                */

/* Copyright (C) 2001 the OURAGAN project.                                */
/* Code available under BSD-like licence (see COPYRIGHT file).            */
/* Author(s): Martin Quinson                                              */

#include "fast_internals.h"
/* defined in data_arg.c */
fast_thash_t *data_arg_properties_get(data_arg_t *da);

struct data_argstack_s {
  char *routine;
  int size;
  int used;
  data_arg_t **args;
};

/**
 * data_argstack_new:
 * @routine_name: the name of the routine
 * @Returns: the data_argstack
 * 
 * Creates a new empty stack.
 */
data_argstack_t *data_argstack_new(const char *routine_name) {
  data_argstack_t *res = malloc(sizeof(data_argstack_t));

  if (!res) return res;

  res->routine = strdup(routine_name);
  res->size = res->used = 0;
  res->args = NULL;
  
  return res;
}

/**
 * data_argstack_free:
 * @as: the arg_stack
 * 
 * Free an argstack
 */
void data_argstack_free(data_argstack_t *as) {
  int i;

  if (as) {
    if (as->used) {
      for (i=0; i<as->used; i++) {
	data_arg_free(&(as->args[i]));
      }
      free(as->args);
    }
    free(as->routine);
    free(as);
  }
}

/**
 * data_argstack_check:
 * @as: the arg_stack
 * @Returns: true if OK, false if not 
 * 
 * Enforce the value count limitations on each argument of this stack.
 */


int data_argstack_check(data_argstack_t *as) {
  int i;

  if (!as) return 0;

  for (i=0; i< as->used; i++) 
    if (!(data_arg_check(as->args[i]))) 
      return 0;

  return 1;
}

/**
 * data_argstack_create:
 * @as: the argstack to create
 * @routine_name: the name of the corresponding routine
 * @argcount: how many arguments should be added to the stack
 * @varargs: the arguments to add to the stack
 * @Returns: True if everything went ok, false if an error occurs
 * 
 * Creates a new argstack, populate it and check it.
 */

int data_argstack_create(data_argstack_t **as, const char *routine_name, int argcount,...) {
  va_list ap;
  int i;

  if (!(*as=data_argstack_new(routine_name))) 
    return 0;

  if (!( (*as)->args=(data_arg_t **)malloc(sizeof(data_arg_t *) * argcount))) {
    data_argstack_free(*as);
    return 0;
  }
  (*as)->used = (*as)->size = argcount;
  va_start(ap,argcount);
  for (i=0; i<argcount; i++) {
    (*as)->args[i] = va_arg(ap,data_arg_t*);
  }
  va_end(ap);
  return data_argstack_check(*as);
}

/**
 * data_argstack_push:
 * @as: the argstack
 * @arg: the argument to push
 * @Returns: True if everything went ok, false if an error occurs
 * 
 * Push an argument on top of the stack
 */

int data_argstack_push(data_argstack_t *as, data_arg_t *arg) {
  if (!as) return 0;
  if (as->used++ == as->size) {
    as->size += 5;
    if (!(as->args = realloc(as->args, sizeof(data_arg_t *) * as->size))) {
      as->size = as->used = 0;
      FAIL("Malloc error\n");
    }
  }
  as->args[as->used -1] = arg;
  return 1;
}

/**
 * data_argstack_pop:
 * @as: the arg_stack
 * @Returns: 
 *
 */


data_arg_t *data_argstack_pop(data_argstack_t *as) {
  if (as && as->used > 1) {
    return as->args[ as->used-- ];
  } else {
    return NULL;
  }
}

/**
 * data_argstack_get:
 * @as: the arg_stack
 * @pos: the position
 * @Returns: the arg_stack
 *
 */


data_arg_t *data_argstack_get(data_argstack_t *as, int pos) {
  if (as && as->used > pos) {
    return as->args[pos];
  } else {
    return NULL;
  }
}

/**
 * data_argstack_routine_get:
 * @as: the arg_stack
 * @Returns: the routine name of this argstack
 * 
 * Don't modify the result, since it's the pointer to the internal data.
 */
char *data_argstack_routine_get(data_argstack_t *as) {
  return as ? as->routine : NULL;
}

/** 
 * data_argstack_split:
 * @as: the argstack
 * @strings: the string args
 * @nums: 
 * @num_used:
 * @Returns: True if everything went ok, false if an error occurs
 *
 * Split the string args from the nums ones.
 */
int data_argstack_split(data_argstack_t *as, char **strings, 
			double **nums, int *num_used) {
  int string_size=0;
  int string_used=0;
  int num_size=0;

  int i,integer,size;
  double dbl;
  char *str;
  char c;
  char num_string[4096];

  (*strings)=NULL;
  (*nums)=NULL;
  *num_used=0;

  if (!as || !(data_argstack_check(as)))
    FAIL("The instance passed to data_argstack_split() do not pass the data_argstack_check().\nPlease be kind enough to go to hell ;)\n");

  for (i=0; i<as->used; i++) {
    switch (data_arg_scaltype_get(as->args[i])) {
    case data_char:
      if (data_arg_constype_get(as->args[i]) == data_cons_scal) {
	if (!(fast_thash_getfirst_int(data_arg_properties_get(as->args[i]),"value",(int*)&c)))
	  FAIL1("Internal error while getting value of arg %d (@%p)\n",i);
	size=2;
      } else {
	if (!(fast_thash_getfirst_string(data_arg_properties_get(as->args[i]),"value",&str)))
	  FAIL1("Internal error while getting value of arg %d\n",i);
	size = strlen(str)+1;
      }

      if (string_size <= string_used + size) {
	string_used += size; 
	string_size += size;
	if (!(*strings)) {
	  if (!(*strings=malloc(sizeof(char)*(string_size))))
	    FAIL("Out of memory\n");
	  (*strings)[0]='\0';
	} else {
	  if (!(*strings=realloc(*strings,sizeof(char)*(string_size))))
	    FAIL("Out of memory\n");
	}
      }
      if ((*strings)[0] != '\0') strcat(*strings,"!");
      if (data_arg_constype_get(as->args[i]) == data_cons_scal) {
	char *p=*strings;
	while (*p != '\0') p++; /* goto EOS */
	*(p++) = c; /* put value in place */
	*p='\0';    /* ends the string */
      } else {
	strcat(*strings,str);
      }
      break;

    case data_int:
    case data_double:
      if (data_arg_constype_get(as->args[i]) != data_cons_scal) {
	WARN("data_argstack_split() does only accept chars, strings, scalar int and scalar double\n");
	break;
      }

      if (num_size <= *num_used) {
	num_size+=5;
	if (!nums) {
	  *nums=(double*)malloc(sizeof(double)*num_size);
	} else {
	  *nums=(double*)realloc(*nums,sizeof(double)*num_size);
	}
	if (!*nums) FAIL("Out of memory\n");
      }

      if (data_arg_scaltype_get(as->args[i]) == data_int) {
	if (!(fast_thash_getfirst_int(data_arg_properties_get(as->args[i]),"value",&integer)))
	  FAIL1("Internal error while getting value of arg %d\n",i);
	(*nums)[ (*num_used)++ ] = (double)integer;
      } else {
	if (!(fast_thash_getfirst_double(data_arg_properties_get(as->args[i]),"value",&dbl)))
	  FAIL1("Internal error while getting value of arg %d\n",i);
	(*nums)[ (*num_used)++ ] = dbl;
      }
      break;
      
    default:
      FAIL("Invalid argument in the argstack\n");
    }
  }
  LOG2("String args='%s' (len=%d)\n",*strings,string_used);
  if (*num_used) {
    char *p;
    sprintf(num_string,"[%.2f",*nums[0]);
    p=&(num_string[0]);
    while (*p != '\0') p++;
    for (i=1;i<*num_used;i++) {
      sprintf(p,", %.2f",(*nums)[ i ]);
      while (*p != '\0') p++;
    }
    sprintf(p,"] (len=%d)",*num_used);
  } else {
    strcat(num_string,"[] (len=0)");
  }
  LOG1("Num args=%s\n",num_string);
  return 1;
}

/**
 * data_argstack_dump:
 * @as: the arg_stack
 *
 * Dumps an argstack for debuging purposes.
 */
void data_argstack_dump(data_argstack_t *as) {
  int i;
  char name[32];

  if (as) {
    printf("Dumping of the argstack %p:\n  routine=%s\n  size=%d\n  used size=%d\n",
	   as,as->routine,as->size,as->used);
    for (i=0; i<as->used; i++) {
      sprintf(name,"arg %d",i);
      data_arg_dump(as->args[i],name,"    ");
    }
  } else {
    printf("The dumped argstack is NULL!!\n");
  }    
  fflush(stdout);
}
