#include <ctype.h> /* isspace() */

#include "fast_internals.h" 

#ifdef HAVE_GSL
#include <gsl/gsl_math.h>
#include <gsl/gsl_multifit.h>
#define GSL_RANGE_CHECK_OFF
#endif

/*******************
 * Type definition *
 *******************/
struct fast_estimator_ {
  double *coef;
  int order;
  int nb_args;
  int nb_coefs;
  double chisq; /* observed chi square while fitting */
};

 
/**********************************
 * Prototype of private functions *
 **********************************/
static int estimator_size(int max,int nb_param);
static inline int sum_tab(int *args,int nb_args);
static int next_args(int *args,int nb_args,int max);
/* this is a private function because I do not want to make a dependance on gsl to use estimators */
#ifdef HAVE_GSL
fast_estimator_t *fast_estimator_alloc_gsl(gsl_matrix *knots, gsl_vector *y,
					   int maxorder);
fast_estimator_t *fast_estimator_alloc_gsl_real(gsl_matrix *knots,
						gsl_vector *y, 
						int order);
double fast_estimator_gsl_test(fast_estimator_t *estimator,
			       gsl_matrix *knots, gsl_vector *y);
#endif
/* compute the maximal size a sprintf will need */
int fast_printf_string_upper_bound(const char *format);
/*----[ dump_args ]----------------------------------------------------------*/
static void dump_args(char *name,int *args,int nb_args,int max);

/********
 * Code *
 ********/
/*----[ estimator_size ]-----------------------------------------------------*/
/*                                                                           */
/* compute how many (i,j,k) exists for which  0<= i+j+k <= max is true       */
/* For 3 parameters, this is ( max )                                         */
/*                           (max+3)                                         */
/*  and this is equal to : \frac{(max+1)...(max+nb_param)}                   */
/*                               ------------------------                    */
/*                                   {nb_param !}                            */
/*                                                                           */
/* Naturally, the function handel any number of parameters.                  */
/*                                                                           */
/*---------------------------------------------------------------------------*/
static int estimator_size(int max,int nb_param) 
{
  long res = 1;
  int i;

  for (i=1;i<=nb_param;i++) {
    res *= max+i;
  }
  for (i=2;i<=nb_param;i++) {
    res /= i;
  }

  //  printf("estimator_size(max=%d ,nb_param=%d)=%d\n",max,nb_param,(int)res);
  return (int)res;
}

/*----[ sum_tab ]------------------------------------------------------------*/
/* compute the sum of all values in the table                                */
/*---------------------------------------------------------------------------*/
static inline int sum_tab(int *args,int nb_args) 
{
  int i,total=0;
  for (i=0 ; i<nb_args ; i++) total += args[i];
  return total;
}

/*----[ dump_args ]----------------------------------------------------------*/
/* dump the args on the screen to debug.                                     */
/*---------------------------------------------------------------------------*/
static void dump_args(char *name,int *args,int nb_args,int max)
{
  int i;

  fprintf(stderr,"%s (total:%d max:%d):",name,sum_tab(args,nb_args),max);
  for (i=0; i<nb_args; i++)
    fprintf(stderr," %d",args[i]);
  fprintf(stderr,"\n");
}

/*----[ next_args ]----------------------------------------------------------*/
/* given a configuration args[0]...args[nb_args-1], search the next one so   */
/*   that \sum_{i=0}^{nb_args-1} args[i] <= p                                */
/*                                                                           */
/* returns true if we managed to build the next config, and false if there   */
/*   is no more                                                              */
/*---------------------------------------------------------------------------*/
static int next_args(int *args,int nb_args,int max) 
{
  int arg=0;

  //FIXME: initialisation should be done by a function ad hoc

  //FIXME: this trick is to use this function in a for(;next_args;) statement
  // This is a bad idea, it should be a do {} while (next_args)
  if (args[0]==-1) { // not initialized
    memset(args,0,sizeof(int)*nb_args);
    return 1;
  }

  args[0]++; // next obvious solution
  //  dump_args("next obvious solution",args,nb_args,max);
  
  arg=0;
  while (sum_tab(args,nb_args) > max && arg<nb_args) {
    // while (this solution is not good) && (we didn't reach the last solution)
    args[arg]=max;
    
    //handel the carry
    while (args[arg] >= max && arg < nb_args-1) { 
      args[arg]=0;
      args[++arg]++;
    }

    if (sum_tab(args,nb_args) > max) {
      // if carry overflowed 
      return 0;
    }
    
  }
  //return 1;
  return arg<nb_args-1 || (arg=nb_args-1 && sum_tab(args,nb_args) <= max);
}

/*---[ estimator_use ]--------------------------------------------(private)--*/
/* does the job of computing the estimator on one point, without output      */
/*---------------------------------------------------------------------------*/
static double estimator_use(fast_estimator_t *estimator,double *args)
{
  /* approximation of the function f(args) */

  double res=0.0,tmp;
  int *power_of_args,arg; /* to which power we should set the args */
  int coef_estim;

  /* malloc some working place */
  if (!(power_of_args=(int *)malloc(sizeof(int)*estimator->nb_args))) {
    fprintf(stderr,"%s:%d: malloc() error\n",__FILE__,__LINE__);
    return -1;
  }

  /* To understand the next for statement, please refer to */
  /* fast_estimator_alloc(), were the same thing is also   */
  /*  used (and documented) there */

  for (coef_estim=0,power_of_args[0]=-1;
       /* next_args search the next step (and returns if found or not) */
       next_args(power_of_args,estimator->nb_args,estimator->order) ; 
       coef_estim++ ) {
     
    tmp=estimator->coef[coef_estim];
    for (arg=0 ; arg<estimator->nb_args ; arg++) {
      tmp *= pow(args[arg],(double)power_of_args[arg]);
    }
    res += tmp;
  }

  free(power_of_args);
  return res;
}

/****[ Code of public functions ]*********************************************/
/*---[ fast_estimator_alloc_empty ]------------------------------------------*/
/* Allocate a null estimator.                                                */
/*---------------------------------------------------------------------------*/
fast_estimator_t *fast_estimator_alloc_empty(void) 
{
  fast_estimator_t *result;

  if (!(result=(fast_estimator_t *)malloc(sizeof(fast_estimator_t)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    return NULL;
  }
  if (!(result->coef=(double *)malloc(sizeof(double)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    return NULL;
  }
  result->coef[0]=0;
  result->order = 0;
  result->nb_coefs = 1;
  result->chisq = 0;
  result->nb_args = 1;

  return result;
}

#ifdef HAVE_GSL
/*---[ fast_estimator_alloc_gsl_real ]---------------------------------------*/
/* Fit to a #order#th order polynomial, so that #y# = #knots# * #result#     */
/*  so, according to the gsl info file, I do: X_{ij} = x_i^j                 */
/*---------------------------------------------------------------------------*/
fast_estimator_t *fast_estimator_alloc_gsl_real(gsl_matrix *knots,
						gsl_vector *y, 
						int order) {

  fast_estimator_t *result;
  gsl_vector *gsl_coef;
  int nb_args=knots->size1;
  int nb_lines=knots->size2;
  int i;
  int line; /* which experiement from knots are we looking for ? (and which line of X are we filling) */

  int *args,arg; /* to build all possibility of nb_params integers so that the sum is less than #order# */

  int size=estimator_size(order,nb_args);
  int col_X; /* iterator on X's columns while building X */
  float val; /* What we will put in X */

  gsl_matrix * X;
  gsl_matrix * cov;
  gsl_multifit_linear_workspace *WORK;

  /* Argument verification */
  if (nb_lines != y->size) {
    fprintf(stderr,
	    "%s:%d:fast_estimator_alloc_gsl():\n"
	    "dimensions mismatch in polynomial regression\n"
	    "parameters have %d lines, and results %d.\n",
	    __FILE__,__LINE__,
	    nb_lines,(int)y->size);
    return NULL;
  }

  /* allocate the needed space */
  if (!(result=(fast_estimator_t *)malloc(sizeof(fast_estimator_t)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    return NULL;
  }
  if (!(result->coef=(double *)malloc(sizeof(double)*size))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    free(result);
    return NULL;
  }

  if (!(args=(int *)malloc(sizeof(int)*nb_args))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    free(result->coef);
    free(result);
    return NULL;
  }

  if (!(X = gsl_matrix_alloc ( nb_lines, size ))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    free(result->coef);
    free(result);
    free(args);
    return NULL;
  }

  if (!(cov = gsl_matrix_alloc (size,size))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    free(result->coef);
    free(result);
    free(args);
    gsl_matrix_free(X);
    return NULL;
  }

  if (!(gsl_coef = gsl_vector_alloc (size))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    free(result->coef);
    free(result);
    free(args);
    gsl_matrix_free(X);
    gsl_matrix_free(cov);
    return NULL;
  }
  
  if (!(WORK=gsl_multifit_linear_alloc(nb_lines,size))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    free(result->coef);
    free(result);
    free(args);
    gsl_matrix_free(X);
    gsl_matrix_free(cov);
    gsl_vector_free(gsl_coef);
    return NULL;
  }
    
  /* some initialisations */
  result->order = order;
  result->nb_args = nb_args;

  /* do the interpolation */
  for(line=0 ; line<nb_lines ; line++) {
    /*
     * In the next for, I want to go through all possible configuration of 
     *  parameters so that the sum of them is less than #order#. 
     *
     * It sounds obscure, but for 3 parameters, we will have :
     *   for (i=0 ; i<=p ; i++) {
     *     for (j=i ; j<=p-i ; j++) {
     *        for (l=i+j ; l<=p-i-j ; l++){
     *              DO PAYLOAD
     *        }
     *     }
     *   }
     *
     * Furthermore, we want to count the cases (in col_X variable).
     *
     */
    fflush(stdout);
    for (col_X=0,args[0]=-1;
	 /* next_args search the next step (and returns if found or not) */
	 next_args(args,nb_args,order) ; 
	 ) {

      //      if (line==0) dump_args("while allocation",args,nb_args,order);
      val=1;
      for (arg=0 ; arg<nb_args ; arg++) {
	val *= pow(gsl_matrix_get(knots,arg,line) , args[arg]);
      }
      gsl_matrix_set(X,line,col_X++,val);
    }
  }  

  //  printf ("Fitting ...\n");fflush(stdout);
  /* search c so that  y = X c */
  gsl_multifit_linear (X,y,gsl_coef,cov, &(result->chisq),WORK);

  //  printf ("Fitting done\n");fflush(stdout);
  result->nb_coefs=gsl_coef->size;
  for (i=0;i<result->nb_coefs;i++) {
    result->coef[i]=gsl_vector_get(gsl_coef,i);
  }

  /* release the used space */
  free(args);
  gsl_matrix_free(X);
  gsl_matrix_free(cov);
  gsl_vector_free(gsl_coef);
  gsl_multifit_linear_free(WORK);

  return result;
}

/*----[ fast_estimator_gsl_test ]--------------------------------------------*/
/* return the average error of this estimator on this data set               */
/*---------------------------------------------------------------------------*/
double fast_estimator_gsl_test(fast_estimator_t *estimator,
			       gsl_matrix *knots, gsl_vector *y)
{
  int i,j;
  double *args;
  double val,theo,err,max_err=-1,avg_err=0,avg2_err=0,ecart=0;

  int nbparams=knots->size1;
  int nbxp=knots->size2,avg2_nbxp=0;
  /* avg2 is the average when you forget points for which the value is < 0.1 */

  if (!(args=(double *)malloc(sizeof(double)*nbparams))) {
    fprintf(stderr,"%s:%d: malloc error\n",__FILE__,__LINE__);
    return -1;
  }
  // compute the maximal and average error, 
  for (i=0; i<nbxp; i++) {
    for (j=0 ; j<nbparams ; j++) {
      args[j]=gsl_matrix_get(knots,j,i);
    }

    theo=estimator_use(estimator,args);
    val=gsl_vector_get(y,i);
    err=fabs((theo-val)/val) * 100;
    if (err > max_err)
      max_err=err;
    avg_err += fabs(err);

    if (fabs(theo)>0.1) {
      avg2_err+=err;
      avg2_nbxp++;
    }
  }
  avg_err /= nbxp;
  avg2_err /= avg2_nbxp;

  // compute the ecart type (FIXME: translation)
  for (i=0; i<nbxp; i++) {
    for (j=0 ; j<nbparams ; j++) {
      args[j]=gsl_matrix_get(knots,j,i);
    }

    theo=estimator_use(estimator,args);
    val=gsl_vector_get(y,i);
    err=fabs((theo-val)/val) * 100;
    ecart += fabs(err - avg_err);
  }
  ecart /= nbxp;

  // print the sumary
/*    printf("\n\n CHI 2: %f\n" */
/*  	 "MAXIMAL ERROR: %f%%\nAVERAGE ERROR: %f%% AVERAGE2 ERROR: %f%%\n" */
/*  	 "ECART TYPE: %f\nVARIANCE: %f\n", */
/*  	 estimator->chisq, max_err, avg_err, avg2_err, ecart, ecart*ecart); */

  free(args);

  return avg2_err;
}

/*---[ fast_estimator_alloc_gsl ]--------------------------------------------*/
/* Fit to a polynomial, so that #y# = #knots# * #result#, determining        */
/*  automagically the order (max #maxorder#)for which the average error is   */
/*  minimal.                                                                 */
/*---------------------------------------------------------------------------*/
fast_estimator_t *fast_estimator_alloc_gsl(gsl_matrix *knots,gsl_vector *y,
					   int maxorder)
{
  int order;
  fast_estimator_t *res=NULL,*this;
  double err_res=-1,err_this;

  for (order=2 ; order<maxorder ; order++) {
    printf("Building estimator of order %d ",order);
    this=fast_estimator_alloc_gsl_real(knots,y,order);
    if (this) {
      err_this=fast_estimator_gsl_test(this,knots,y);
      printf("; average error = %.2e %% ",err_this);
      if (res==NULL || err_this<err_res) {
	if (res)
	  fast_estimator_free(res);
	res=this;
	err_res=err_this;
	printf ("; Selecting this estimator\n");
      } else {
	fast_estimator_free(this);
	printf ("; Rejecting this estimator\n");
      }
    } else {
      printf("Error while allocating estimator\n");
    }
  }
  printf("\n");
  return res;
} 

/*---[ fast_estimator_alloc_array ]------------------------------------------*/
fast_estimator_t *fast_estimator_alloc_array(int nbparams,int nbxp,
					     double *xp,double *y,
					     int maxorder) {
  int i,j;
  fast_estimator_t *res;
  
  gsl_matrix *knots;
  gsl_vector *gsl_y;

  /* Some arg verification */
  if (!nbxp) {
    fprintf(stderr,"%s:%d: I can't interpolate a set of zero experiments\n",
	    __FILE__,__LINE__);
    return NULL;
  }

  /* Some allocation */
  if (!(knots=gsl_matrix_alloc (nbparams,nbxp))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    return NULL;
  }

  if ((!(gsl_y=gsl_vector_alloc (nbxp)))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    gsl_matrix_free(knots);
    return NULL;
  }

  /* fill the args for ...alloc_gsl() */
  for (i=0; i<nbxp; i++) {
    for (j=0; j<nbparams; j++) {      
      gsl_matrix_set(knots,j,i,xp[nbparams*i+j]);
    }
    gsl_vector_set(gsl_y,i,y[i]);
  }

  /* do the job */
  res=fast_estimator_alloc_gsl(knots, gsl_y,maxorder);

  /* free the allocated memory */
  gsl_matrix_free(knots);
  gsl_vector_free(gsl_y);
  return res;
}


/*---[ fast_estimator_alloc_vargs ]------------------------------------------*/
fast_estimator_t *fast_estimator_alloc_vargs(int nbparams,int nbxp,
					     int maxorder,...) {
  int i,j;
  va_list pa;
  double *args;
  
  gsl_matrix *knots;
  gsl_vector *y;

  fast_estimator_t *res;

  /* Some allocation */
  if (!(knots=gsl_matrix_alloc (nbparams,nbxp))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    return NULL;
  }

  if (!(y=gsl_vector_alloc (nbxp))) {
    fprintf(stderr,"%s:%d: Malloc error\n",
	    __FILE__,__LINE__);
    gsl_matrix_free(knots);
    return NULL;
  }

  /* fill the args for ...alloc_gsl() */
  va_start(pa,maxorder);
  for (j=0; j<nbparams; j++) {      
    args=va_arg(pa,double*);
    for (i=0; i<nbxp;i++) {
      gsl_matrix_set(knots,j,i,args[i]);
    }
  }
  args=va_arg(pa,double*);
  for (i=0; i<nbxp;i++) {
    gsl_vector_set(y,i,args[i]);
  }
  va_end(pa);

  /* do the job */
  res=fast_estimator_alloc_gsl(knots, y,maxorder);

  /* free the allocated memory */
  gsl_matrix_free(knots);
  gsl_vector_free(y);

  return res;
}
#else /*****[ ! HAVE_GSL ]****************************************************/

fast_estimator_t *fast_estimator_alloc_array(int nbparams,int nbxp,
					     double *xp,double *y,
					     int maxorder) {

  fprintf(stdout,"Well, I can't do data fiting without GSL...\n");
  if (0) estimator_size(0,0); /* simulate a call to avoid a warning 'not used' */
  return fast_estimator_alloc_empty();
}
fast_estimator_t *fast_estimator_alloc_vargs(int nbparams,int nbxp,
					     int maxorder,...) {
  fprintf(stdout,"Well, I can't do data fiting without GSL...\n");
  return fast_estimator_alloc_empty();
}
#endif /****[ HAVE_GSL ]******************************************************/

/*---[ fast_estimator_free ]-------------------------------------------------*/
void fast_estimator_free(fast_estimator_t *estimator) {
  if (estimator) {
    if (estimator->coef)
      free(estimator->coef);
    free(estimator);
  }
}

/*---[ fast_estimator_chisq ]------------------------------------------------*/
inline double fast_estimator_chisq(fast_estimator_t *estimator) {
  return estimator->chisq;
}

/*---[ fast_estimator_use_array ]--------------------------------------------*/
double fast_estimator_use_array(fast_estimator_t *estimator,double *args)
{
  /* approximation of the function f(args) */

  double res=estimator_use(estimator,args);
  int arg; 
   
  return res; /* FIXME: no output */

  printf("fast_estimator_use_array(");
  for (arg=0 ; arg<estimator->nb_args ; arg++) {
    printf("%s%f",(arg?",":""),args[arg]);
  }
  printf(")=%f\n",res);

  return res;
}

/*---[ fast_estimator_use_vargs ]--------------------------------------------*/
double fast_estimator_use_vargs(fast_estimator_t *estimator,...) {
  double *args;
  int arg;
  va_list pa;

  double res;

  if (!(args=(double *)malloc(sizeof(double)*estimator->nb_args))) {
    fprintf(stderr,"%s:%d: malloc() error\n",__FILE__,__LINE__);
    return -1;
  }

  /* parse from the arguments on which point we should do the interpolation */
  va_start(pa,estimator);
  for (arg=0 ; arg<estimator->nb_args ; arg++) {
    args[arg]=va_arg(pa,double);
  }
  va_end(pa);

  res = fast_estimator_use_array(estimator,args);
  free(args);
  return res;
}

/*---[ fast_estimator_test_array ]-------------------------------------------*/
void fast_estimator_test_array(fast_estimator_t *estimator,
			       int nbparams,
			       int nbxp,
			       double *xp,double *y)
{
  int i,j;
  double *args;
  double val,theo,err,err_max=-1,err_moy=0,ecart=0;
  double err_abs,err_abs_moy=0, err_abs_max=-1; /* Absolute error */
  int print;

  if (!(args=(double *)malloc(sizeof(double)*nbparams))) {
    fprintf(stderr,"%s:%d: malloc error\n",__FILE__,__LINE__);
    return;
  }

  // compute the maximal and average error, 
  // printe the values for square matrices
  for (i=0; i<nbxp; i++) {
    print=1; 
    for (j=0 ; j<nbparams ; j++) {
      args[j]=xp[nbparams*i+j];
      //      if (j>0) {
      //	print = print && args[j] == args[j-1];
      //      }
    }

    val=y[i];
    if (val==-1) continue;
    theo=estimator_use(estimator,args);
    /* Handle absolute error */
    err_abs=fabs(theo-val);
    err=(err_abs/val) * 100;

    if (err_abs > err_abs_max) {
      err_abs_max=err_abs;
      err_max=err;
    }

    err_abs_moy+=err_abs;
    if (err > err_max) err_max=err;
    err_moy += fabs(err);

    if (print) {
      printf("line %3d ",i);
      for (j=0 ; j<nbparams ; j++) 
	printf("%c%4d",(j?';':'('),(int)args[j]); 
      printf("): err=%f%% ",err);
      printf("estimator_use=%f ; data=%f\n",theo,val);
    }
  }
  err_moy /= nbxp;
  err_abs_moy /= nbxp;

  // compute the ecart type (FIXME: translation)
  for (i=0; i<nbxp; i++) {
    for (j=0 ; j<nbparams ; j++) {
      args[j]=xp[nbparams*i+j];
    }

    theo=estimator_use(estimator,args);
    val=y[i];
    err=fabs((theo-val)/val) * 100;
    ecart += fabs(err - err_moy);
  }
  ecart /= nbxp;

  // print the summary
  printf("\nSUMMARY: chi 2: %f ; Max err: %f (%f%%) ; avg err: %f (%f%%)\n"
	 "         Ecart type: %f ; Variance: %f\n",
	 estimator->chisq, err_abs_max, err_max, err_abs_moy, err_moy, ecart, ecart*ecart);

  free(args);
}

/*---[ fast_estimator_dump ]-------------------------------------------------*/
void fast_estimator_dump(FILE *stream,fast_estimator_t *estimator) {
  int *power_of_args,arg; /* to which power we should set the args */
  int coef_estim;
  char name_of_args[6]="abcdef";

  if (0) dump_args(NULL,NULL,0,0); // trick to use -Werror -Wunused
  fprintf(stream,
	  "The maximal order of the estimator is %d ;"
	  " its chisq is %f\n",
	  estimator->order,estimator->chisq);
  /* malloc some working place */
  if (!(power_of_args=(int *)malloc(sizeof(int)*estimator->nb_args))) {
    fprintf(stderr,"%s:%d: malloc() error\n",__FILE__,__LINE__);
    return;
  }

  /* To understand the next for statement, please refer to */
  /* fast_estimator_alloc(), were the same thing is also   */
  /*  used (and documented) there */

  for (coef_estim=0,power_of_args[0]=-1;
       /* next_args search the next step (and returns if found or not) */
       next_args(power_of_args,estimator->nb_args,estimator->order) ; 
       coef_estim++ ) {
    
    fprintf(stream,"%s % 09.9f ",
	    (coef_estim?"+":" "), 
	    estimator->coef[coef_estim]);
    for (arg=0 ; arg<estimator->nb_args ; arg++) {
      fprintf(stream," x %c ^ %d",name_of_args[arg],power_of_args[arg]);
    }
    fprintf(stream,"\n");
  }
  free(power_of_args);
}

/*---[ fast_printf_string_upper_bound ]--------------------------------------*/
int fast_printf_string_upper_bound(const char *format) {
/* inspired (ok, stolen) from g_printf_string_upper_bound from gnome libs */
/* limitations: 
    - '*' is not a valid format modifier 
    - %s,%c,%p,%n are not allowed
*/
  int len=0;
  
  while (*format) {
    char c;
    
    c = *format++;
    
    if (c == '%') {
      int done = 0;
      int extra_long=0,long_int=0;
      
      while (*format && !done) {
	switch (*format++) {
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	  /* add specified format length, since it might exceed the
	   * size we assume it to have.
	   */
	  format -= 1;
	  len += strtol (format, (char**) &format, 10);
	  break;
	case 'h':
	  /* ignore short int flag, since all args have at least the
	   * same size as an int
	   */
	  break;
	case 'l':
	  if (long_int)
	    extra_long = 1; /* linux specific */
	  else
	    long_int = 1;
	  break;
	case 'q':
	case 'L':
	  long_int = 1;
	  extra_long = 1;
	  break;
	case 'd':
	case 'i':
	case 'o':
	case 'u':
	case 'x':
	case 'X':
	  len += extra_long ? 64 : 32;
	  done = 1;
	  break;
	case 'D':
	case 'O':
	case 'U':
	  len += 32;
	  done = 1;
	  break;
	case 'e':
	case 'E':
	case 'f':
	case 'g':
	  len += extra_long ? 128 : 64;
	  done = 1;
	  break;
	case '%':
	  len += 1;
	  done = 1;
	  break;
	default:
	  /* ignore unknow/invalid flags */
	  break;
	}
      }
    }
    else
      len += 1;
  }
  
  return len;
}


/*---[ fast_estimator_to_string ]--------------------------------------------*/
char *fast_estimator_to_string(fast_estimator_t *estimator) {
  size_t n = estimator->nb_coefs ;
  
  size_t i;
  char *buffer,*p,*res;
  int status;

  
  buffer =(char *)malloc(n*fast_printf_string_upper_bound("%.32e")
			 + (n+4)*sizeof(char) /* \t between elements */ 
			 /*order,nb_args,nb_coefs*/
			 + 3*fast_printf_string_upper_bound("%d") 
			 + fast_printf_string_upper_bound("%e")); /* chisq */
  if (buffer == NULL) {
    fprintf(stderr,"%s:%d: malloc() error\n",__FILE__,__LINE__);
    return NULL;
  }
  
  p=buffer;

  status = sprintf (p,"%d\t%d\t%d\t",
		    estimator->order,estimator->nb_args,estimator->nb_coefs);
  if (status < 0){
    fprintf(stderr,"%s:%d: sprintf failed",__FILE__,__LINE__);
    free(buffer);
    return NULL;
  }
  while (*p!='\0') p++;

  for (i = 0; i < n; i++) {    
    status = sprintf (p,
		      "%.32e",
		      estimator->coef[i]);
    if (status < 0) {
      fprintf(stderr,"%s:%d: sprintf failed",__FILE__,__LINE__);
      free(buffer);
      return NULL;
    }
    
    while (*p!='\0') p++;

    status = sprintf(p,"\t");

    if (status < 0){
      fprintf(stderr,"%s:%d: sprintf failed",__FILE__,__LINE__);
      free(buffer);
      return NULL;
    }
    while (*p!='\0') p++;
  }

  status = sprintf (p,"\t%e",estimator->chisq);
  if (status < 0){
    fprintf(stderr,"%s:%d: sprintf failed",__FILE__,__LINE__);
    free(buffer);
    return NULL;
  }

  //  res=realloc(buffer,strlen(buffer));
  if (!(res=(char *)malloc(sizeof(char)*((int)strlen(buffer)+1)))) {
    return buffer;
  }
  strcpy(res,buffer);
  free(buffer);
  return res;
}

fast_estimator_t *fast_estimator_alloc_from_string(char *string) {
  fast_estimator_t *result;
  char *p=string,*new_p;
  size_t i;
  int status;

  if (!(result=(fast_estimator_t *)malloc(sizeof(fast_estimator_t)))) {
    fprintf(stderr,"%s:%d: Malloc failed",__FILE__,__LINE__);
    return NULL;
  }

  status = sscanf (p,"%d\t%d\t%d\t",
		   &(result->order),&(result->nb_args),&(result->nb_coefs));
  /* check if the scanf was ok */
  if (status != 3) {
    fprintf(stderr,"%s:%d: sscanf failed (got %d)\n",__FILE__,__LINE__,status);
    free(result);
    return NULL;
  }
  /* pass what we've read */
  while (!isspace((int) *(++p)));    p++; /* order */
  while (!isspace((int) *(++p)));    p++; /* nb_args */
  while (!isspace((int) *(++p)));    p++; /* nb_coefs */
  /* alloc the matrix for the coefs */
  result->coef=(double *)malloc(sizeof(double)*result->nb_coefs);
  if (!(result->coef)) {
    fprintf(stderr,"%s:%d: Malloc failed",__FILE__,__LINE__);
    free(result);
    return NULL;
  }

  /* read the coefs */
  for (i = 0; i < result->nb_coefs; i++) {
    double tmp ;

    while (isspace((int) *p)) p++;
    tmp=strtod(p,&new_p);
          
    if (p == new_p) {
      fprintf(stderr,"%s:%d: strtod failed",__FILE__,__LINE__);
      free(result->coef);
      free(result);
      return NULL;
    }
    result->coef[i] = tmp;
    p=new_p;

  }
  
  /* read the chisq */
  result->chisq=strtod(p,&new_p);
  if (p == new_p) {
    fprintf(stderr,"%s:%d: strtod failed",__FILE__,__LINE__);
    free(result->coef);
    free(result);
    return NULL;
  }

  return result;
}
