/* $Id: fast.c,v 1.12 2003/06/19 16:22:44 mquinson Exp $ */

/* fast.c - external functions of the Fast Agent's System Timer           */

/* Copyright (C) 2001,2002,2003 the OURAGAN project.                      */
/* Code GPL'ed (see www.gnu.org).                                         */

/* Author(s): Martin Quinson                                              */

#include <netdb.h> /* gethostbyaddr() FIXME: rien a foutre la */
#include <netinet/in.h> /* struct in_addr */
#include <arpa/inet.h> /* inet_* */

#include "fast_internals.h"

/*####[ Code of private functions ]#######################################*/
/*@null@*/ char *getMachineName (ip_t host);

/*####[ Code of public functions ]########################################*/

/*====[ Operations on ip_t ]=================================================*/
inline ip_t ip_str2ip_t(const char *ip) {
  return (ip_t)inet_addr(ip);
}
inline char *ip_t2ip_str(ip_t ip) {
  struct in_addr addrAsInAddr;
  addrAsInAddr.s_addr = ip;
  return inet_ntoa(addrAsInAddr);
}

/*====[ fast_comm_time ]==================================================*/
/*                                                                        */
/* Compute the needed time (in seconds) to send #howmany# bytes from the  */
/*   #source# to the #dest#.                                              */
/*                                                                        */
/* The result is a fast_comm_result_t structure, described below. The     */
/*  value returned is volatile and will be overwritten by subsequent calls*/
/*                                                                        */
/*....[ fast_comm_result_t ]..............................................*/
/*                                                                        */
/* One day, FAST will be multi protocol (for now, only TCP is available). */
/* This day, the result of fast_comm_time will be a #nb# long list of:    */
/*   {#protocol# ; #time# over this protocol}                             */
/* This list is sorted, the shortest time first.                          */
/*                                                                        */
/*========================================================================*/

/*----[ fast_comm_result_free ]-------------------------------------------*/

/**
 * fast_comm_result_free:
 * @res: a previously computate results 
 *
 * free a previously computate result
 */

void fast_comm_result_free(/*@only@*//*@null@*/fast_comm_result_t *res) 
{
  if (res) {
    if (res->protocol) free (res->protocol);
    if (res->time) free (res->time);
    free(res);
  }
}

/**
 * getMachineName:
 * @host: the ip of the host
 * @Returns: The name of the machine
 *
 * Give the name of a host
 */

char *getMachineName (ip_t host) 
{
  struct in_addr addrAsInAddr;
  struct hostent *addrEntry;
  char *res;

  /* search machine name (FIXME: rien a foutre la) */
  addrAsInAddr.s_addr = (unsigned long)host;

#if defined(HOST_OS_OPENBSD) || defined(HOST_OS_DARWIN)
  addrEntry=gethostbyaddr((char *)&addrAsInAddr,sizeof(addrAsInAddr),0);
#else
  addrEntry=gethostbyaddr((char *)&addrAsInAddr,sizeof(addrAsInAddr),AF_INET);
#endif
  if (!addrEntry) {
      ERROR1("Can't get the name of machine %s\n",ip_t2ip_str(host));
      return NULL;
  } 
  res=strdup(addrEntry->h_name);
  
  free(addrEntry->h_name);
  if (addrEntry->h_aliases) free(addrEntry->h_aliases);
  if (addrEntry->h_addr_list) free(addrEntry->h_addr_list);
  free(addrEntry);
   
  return res;
}

/*----[ fast_comm_time_list ]-----------------------------------------------*/

/**
 * fast_comm_time_list:
 * @source: the name of the source
 * @destination: the name of the host to reach
 * @howbig: the size of the packets in bytes
 * @where: the results
 * @Returns: True if everything went OK, false if an error occurs
 * 
 * Compute the needed time (in seconds) to send @howbig bytes from the
 * @source to the @destination, and return the list of all results for each
 * possible protocol
 */

int fast_comm_time_list(const char *source,const char *destination,size_t howbig,
			/*@out@*//*@only@*//*@null@*/fast_comm_result_t **where)
{
  double bandwidth,latency;
  char *src,*dest;

  //  if (!where) return 0;
  if (!source || !destination) {
    *where=NULL;
    return 0;
  }

  if (!(*where=(fast_comm_result_t*)malloc(sizeof(fast_comm_result_t)))) 
    FAIL("Out of memory\n");

  if (!(src=malloc(sizeof(char)*(strlen(source)+2)))) {
    free (*where); 
    *where=NULL;
    FAIL("Out of memory\n");
  }
   
  if (!(dest=malloc(sizeof(char)*(strlen(destination)+2)))) {
    free (*where); 
    free (src);
    *where=NULL;
    FAIL("Out of memory\n");
  }
  snprintf(src,strlen(source)+2,"%s*",source);
  snprintf(dest,strlen(destination)+2,"%s*",destination);
  
  /* FIXME: As long as there is only one protocol, our life is easier here */
  /* FIXME: for now, #protocol# is unspecified */
  (*where)->nb=1;
  (*where)->protocol = (int*)malloc(sizeof(int));
  (*where)->time = (double*)malloc(sizeof(double));

  if (!(*where)->protocol || !(*where)->time) {
    if ((*where)->protocol)
      free((*where)->protocol);
    free((*where));
    FAIL("Out of memory\n");
  }

  if (!strcmp(src,dest)) {
    //FIXME: NWS can't monitor link from one host to itself
    LOG("sendTime=0 from a machine to itself (FIXME)\n");
    (*where)->time[0] = 0;
  } else {
    if (!fast_avail_bandwidth(src,dest,&bandwidth) 
	|| !fast_avail_latency(src,dest,&latency)) {
      free((*where));
      return 0;
    }

    (*where)->time[0] = latency + ((double)howbig) / bandwidth  / 1024 / 1024;

    LOG4("sendTime= lat=%f (s) + size=%f (Mb) / bw=%f (Mb/s) )=%f\n",
	 latency, ((double)howbig)/1024/1024, bandwidth, (*where)->time[0]);
  }
  
  return 1;
}

/*----[ fast_comm_time_best ]---------------------------------------------*/
/* Returns the time of the best forecast (and -1 on error).               */
/*------------------------------------------------------------------------*/

/**
 * fast_comm_time_best:
 * @source: the source machine
 * @dest: the host to reach
 * @howbig: the size of the packets in bytes
 * @value: the time of the best forcast
 * @Returns: True if everything went ok, false if an error occurs
 *
 * determine the time of the best forcast
 */

int fast_comm_time_best(const char *source,const char *dest,size_t howbig,
			/*@OUT@*/double *value)
{
  fast_comm_result_t *comm_res;

  if (!fast_comm_time_list(source,dest,howbig,&comm_res))
    return 0;
  
  if (!comm_res || !(comm_res->nb)) {
    return 0;
  }
  *value=comm_res->time[0];

  fast_comm_result_free(comm_res);
  return 1;
}


/*====[ fast_comp_time ]==================================================*/
/*                                                                        */
/* Compute the needed time (in seconds) to execute #function# on #host#   */
/*   or -1 if impossible                                                  */
/*                                                                        */
/* FIXME: we need a result type as for comm time to give the configuration*/
/*  of the parallal task which can achive the given forecast, but I've so */
/*  few clue about all this stuff that I prefer skip it for now...        */
/*                                                                        */
/*========================================================================*/

/**
 * fast_comp_time_best:
 * @host: the host to reach
 * @function: the function to evaluate
 * @value: the needed time in sec to excecute @function
 * @Returns: True if everything went ok, false if an error occurs
 *
 * Compute the needed time (in seconds) to execute @function on @host 
 */

int fast_comp_time_best(const char *host,data_argstack_t *function,
			/*OUT*/double *value)
{
  double avaispace,avaitime;
  double needspace,needtime;
  char *hoststar; /* host + '*' at the end */

  if (!(hoststar=malloc(sizeof(char)*(strlen(host)+2))))
    FAIL("Out of memory\n");
  sprintf(hoststar,"%s*",host);

  /* get the parts of the forecast */
  if (!fast_avail(FAST_RESOURCE_AVAILABLE_CPU,hoststar,NULL,&avaitime) ||
      !fast_need(FAST_NEED_TIME,host,function,&needtime))
    return 0;

  if (fast_need(FAST_NEED_SPACE,host,function,&needspace) &&
      fast_avail(FAST_RESOURCE_MEMORY,hoststar,NULL,&avaispace)) {
    avaispace *= 1024*1024; // convert Mb to b
    if (needspace>avaispace) {
      LOG2("Not enought space to do the computation (need %f, %f avail).\n",
	   needspace,avaispace);
      *value=-1;
      return 1;
    }
  } /* if NWS or LDAP can't tell about mem, do not check it */

  /* final computation */
  if (needtime < 0)
    needtime=0;
  *value=(avaitime==0?65325:(1/avaitime))*needtime;
  LOG5("Retrieve time for %s on %s from both LDAP and NWS."
       " Found %fs (=%f/%f).\n",
       data_argstack_routine_get(function),host,*value,needtime,avaitime);
  return 1;
}

/*----[ fast_init ]-------------------------------------------------------*/
/**
 * fast_init:
 * @first: 
 * @Varargs:
 * @Returns: true if everything went OK, false if an error occurs
 *
 * Push some option settings to fast. Each option comes in several arguments: 
 * name and value see in the code of fast_config_register to see the list of 
 * valid names (or the tobewritten documentation ;)  Each value can be of 
 * type string, int, double or host (host values are pairs of hostname 
 * (string) and port (integer)). Last argument must be NULL (unless you 
 * want to segfault ;) Note that previously existing settings are not removed 
 * by this function use fast_config_set if you want to *change* a value
 */

int fast_init(const char *first,...) {
  va_list pa;

  DirectDiagnostics(DIAGINFO, stdout);
  DirectDiagnostics(DIAGLOG, stdout);
  DirectDiagnostics(DIAGWARN, stderr);
  DirectDiagnostics(DIAGERROR, stderr);
  DirectDiagnostics(DIAGFATAL, stderr);

  va_start(pa,first);
  if (!fast_config_init(first,pa))
    return 0;
  va_end(pa);
   
  LOG("Initialize FAST\n");
  if (!fast_avail_init())
    FAIL("Can't initialize the system availability monitoring submodule.\n");
  
  if (!fast_need_init())
    FAIL("Can't initialize the routine needs forecasting submodule.\n");

  LOG("Initialization of FAST done\n");
  return 1;
}

/*====[ lower level interface ]===========================================*/
/*                                                                        */
/* Get the status of the system in case you want to redo the computation, */
/*  or if you do a visualisation tool. (hello Eddy ;)                     */
/*                                                                        */
/* The input data are always the name of the machines, for example "pixies". */
/*  it could be the long name (ie "pixies.ens-lyon.fr"), but it's safer to*/
/*  use the short name (I build a request for nws with (host=#hostname#*),*/
/*  so the '*' will match the long name if you give the short one,        */
/*  but not the contrary.                                                 */
/*                                                                        */
/* All these functions return true if all went ok, and false on error.    */
/*                                                                        */
/*========================================================================*/


/**
 * fast_avail_load:
 * @hostName: the name of the machine
 * @value: get the load of the system
 * @Returns: true if all went ok, falst on error
 * 
 * Get the load of the system in case you want to redo the computation,
 * or if you do a visualisation tool. 
 */

int fast_avail_load(const char *hostName,double *value) 
{
  return fast_avail(FAST_RESOURCE_AVAILABLE_CPU,hostName,NULL,value);
}

/**
 * fast_avail_cpucount:
 * @hostName: the name of the machine
 * @value: get the number of cpu on the system
 * @Returns: true if all went ok, falst on error
 *
 * Get the number of cpu on the system in case you want to redo the computation,
 * or if you do a visualisation tool. 
 */


int fast_avail_cpucount(const char *hostName,int *value) 
{
  int ok;
  double dval;
  if ((ok=fast_avail(FAST_RESOURCE_AVAILABLE_CPU,hostName,NULL,&dval))) {
    *value=(int)dval;
  }
  return ok;
}

/**
 * fast_avail_memory:
 * @hostName: the name of the machine
 * @value: get the available memory on the system
 * @Returns: true if all went ok, falst on error
 * 
 * Get the avaiable memory on the system in case you want to redo the 
 * computation, or if you do a visualisation tool. 
 */


int fast_avail_memory(const char *hostName,double *value)
{
  return fast_avail(FAST_RESOURCE_MEMORY,hostName,NULL,value);
}

/**
 * fast_avail_latency:
 * @source: the name of the source machine
 * @dest: the name of the machine to reach
 * @value: get the latency between @source and @dest 
 * @Returns: true if all went ok, falst on error
 * 
 * Get the latency between the @source and @dest in case you want to redo the 
 * computation, or if you do a visualisation tool. 
 */


int fast_avail_latency(const char *source,const char *dest,double *value)
{
  return fast_avail(FAST_RESOURCE_LATENCY,source,dest,value);
}

/**
 * fast_avail_bandwidth:
 * @source: the name of the source machine
 * @dest: the name of the machine to reach
 * @value: get the available brandwidth between @source and @dest 
 * @Returns: true if all went ok, falst on error
 * 
 * Get the brandwidth between the @source and @dest in case you want to 
 * redo the computation, or if you do a visualisation tool. 
 */ 


int fast_avail_bandwidth(const char *source,const char *dest,double *value)
{
  return fast_avail(FAST_RESOURCE_BANDWIDTH,source,dest,value);
}
