/* $Id: fast_need.c,v 1.5 2003/06/10 13:36:11 pmachard Exp $ */

/* fast_need.c - the functions allowing FAST to compute the routines needs   */

/* Copyright (C) 2001 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include "fast_need_plugins.h"

/*====[ Globals ]============================================================*/
/*----[ The table of the plugins ]-------------------------------------------*/
const int fast_need_plugins_count = 0 /* no other need plugin so far */
#ifdef HAVE_LDAP
+1
#endif
;

const fast_need_plugin_t fast_need_plugins[]={
#ifdef HAVE_LDAP
  {"LDAP", &fast_ldap_init, &fast_ldap_getestimator},
#endif 
  {"Dummy entry", NULL,NULL } /* needed for syntax */
};

/*----[ fast_need_cache ]----------------------------------------------------*/
/* The main cache for LDAP in FAST. Each cell of this tab a generic cache    */
/*---------------------------------------------------------------------------*/
fast_cache_t *fast_need_cache[FAST_NEED_TYPE_COUNT];

const char *fast_need_name[]={"time", "space"};

/**
 * fast_need_init:
 * @Returns: True if everything went ok, false if an error occurs
 *
 * Initialize the routine needs forecasting submodule.
 */

int fast_need_init(void) {
  int i;
  int errcode;

  for (i=0; i<fast_need_plugins_count; i++)
    if (fast_need_plugins[i].init && !fast_need_plugins[i].init())
      errcode++;

  return !errcode;
}


/**
 * fast_need:
 * @need: what we need
 * @host: the name of the host
 * @argstack: the argstack
 * @value: the returned value
 * @Returns: true if everything wenk OK, false if not
 *
 * Search the needs of a routine.
 */

int fast_need(int need, const char *host, data_argstack_t *argstack,
	      /*OUT*/double *value) 
{
  fast_estimator_t *estim;
  /* The data about this instance. 
     Should not be freed because their are static else where in the code */
  char *stringargs;
  double *numargs; 
  /* end of data to not free */
  int numargs_count;
  int i;

  if (!data_argstack_split(argstack,&stringargs,&numargs,&numargs_count))
    FAIL("Can't separate string args from num ones to retrieve the estimator.\n");

  if (!fast_cache_retrieve(fast_need_cache[need],host,(void**)&estim))
    FAIL("Internal error in cache while searching the estimator\n");

  if (!estim) { /* not cached. Retrieve it */
    fast_estimator_t **retrieved=NULL;
    
    /* search a plugin able to handle the request */
    for (i=0; i<fast_need_plugins_count; i++)
      if (fast_need_plugins[i].getestimator && 
	  fast_need_plugins[i].getestimator(host,
					    data_argstack_routine_get(argstack),
					    stringargs,
					    &retrieved)) {
	break;
      } else { /* error in plugin */
	retrieved=NULL; 
      }
    
    if (retrieved) { /* Store it back */
      for (i=0; i<FAST_NEED_TYPE_COUNT; i++) 
	if (retrieved[i]) {
	  if (!fast_cache_insert(&( fast_need_cache[i] ),strdup(host),
				 retrieved[i] ,
				 (void (*)(void*))&fast_estimator_free))
	    return 0;
	}
      estim=retrieved[need];
      free(retrieved);
    }
    if (!estim)
      FAIL4("Cannot get the %s estimator for %s(%s) on %s from all registered storage plugins\n",
	    fast_need_name[need], data_argstack_routine_get(argstack),stringargs,host);
  }

  /* Here, we either got the estimator from the cache, or retrieved and cached it successfully */
  *value=fast_estimator_use_array(estim,numargs);
  if (*value < 0)
    *value = 0;
  return 1;
}

/**
 * fast_need_time:
 * @host: the name of the host
 * @argstack: the argstack
 * @value: the returned value
 * @Returns: true if everything wenk OK, false if not
 * 
 * Search the needed time.
 */ 

int fast_need_time(const char *host, data_argstack_t *argstack,double *value) {
  return fast_need(FAST_NEED_TIME,host,argstack,value);
}

/**
 * fast_need_space:
 * @host: the name of the host
 * @argstack: the argstack
 * @value: the returned value
 * @Returns: true if everything wenk OK, false if not
 * 
 * Search the needed space.
 */ 

int fast_need_space(const char *host, data_argstack_t *argstack,double *value) {
  return fast_need(FAST_NEED_SPACE,host,argstack,value);
}
