/* $Id: ldap.c,v 1.5 2003/10/14 16:28:46 mquinson Exp $ */

/* fast_ldap.c - the functions allowing FAST to use LDAP                     */

/* Copyright (C) 2001 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include <ldap.h>

#include "fast_need_plugins.h"

/*----[ fast_ldap_socket, fast_ldap_mask ]-----------------------------------*/
/* Things to connect to LDAP                                                 */
/*---------------------------------------------------------------------------*/
LDAP *fast_ldap_socket; 
char *fast_ldap_basedn;

/* Configuration */
int fast_ldap_use;
int fast_ldap_debug;

/*----[ fast_ldap_need_name ]------------------------------------------------*/
/* The name of the attribute containing the fast_need in LDAP.               */
/*---------------------------------------------------------------------------*/
/* FIXME: I should make a real LDAP schema there */
const char *fast_ldap_need_name[FAST_NEED_TYPE_COUNT+1]= {
  "description","street",(char*)NULL};

int fast_ldap_search(char *filter,char *base,/*OUT*/char ***where);

/*====[ Code of functions ]===============================================*/
/*----[ fast_ldap_init ]--------------------------------------------------*/
int fast_ldap_init() {
  char *host,*basedn;
  int port,errcode;
  
  if (!fast_config_getfirst_int("ldap_use",&fast_ldap_use) || !fast_ldap_use) {
    LOG("LDAP was configured out\n");
    return 1; /* nothing to do if user don't want to use LDAP */
  }

  if (!fast_config_getfirst_host("ldap_server",&host,&port) ||
      !fast_config_getfirst_string("ldap_binddn",&basedn) ||
      !fast_config_getfirst_int("ldap_debug",&fast_ldap_debug)) {
    FAIL("Can't get the configuration about LDAP\n");
  }
  if (fast_ldap_debug) {
    if( ber_set_option( NULL, LBER_OPT_DEBUG_LEVEL, &fast_ldap_debug )
	!= LBER_OPT_SUCCESS ) {
      FAIL1("LDAP_init: Could not set LBER_OPT_DEBUG_LEVEL %d\n", fast_ldap_debug );
    }
    if( ldap_set_option( NULL, LDAP_OPT_DEBUG_LEVEL, &fast_ldap_debug )
	!= LDAP_OPT_SUCCESS ) {
      FAIL1( "LDAP_init: Could not set LDAP_OPT_DEBUG_LEVEL %d\n", fast_ldap_debug );
    }
    LOG1("LDAP_init: Debug level: %d\n",fast_ldap_debug);
  }

  LOG3("LDAP_init: Contact %s:%d as LDAP server, using %s as basedn\n",host,port,basedn);
  if (!(fast_ldap_socket=ldap_init(host,port)))
    FAIL1("LDAP_init: Can't init the connection: %s\n",
	  ldap_err2string(ldap_result2error(fast_ldap_socket, NULL,0)));


  fast_ldap_basedn=(char*)basedn; 
  errcode=ldap_simple_bind_s( fast_ldap_socket,NULL/*who*/,NULL/*cred*/);
  if (errcode != LDAP_SUCCESS ) { 
    ERROR1("Error %d while connecting to LDAP: \n",errcode);
    ldap_perror( fast_ldap_socket, "ldap_bind" );
    FAIL1("%s\n",ldap_err2string(ldap_result2error(fast_ldap_socket, NULL,0)));
  } 

  free(host);
  return 1;
}

/*----[ fast_ldap_estimator ]---------------------------------------------*/
/* Retrieve an estimator from the LDAP server.                            */
/* Note: this function prefetchs all estimators (time and space).         */
/*------------------------------------------------------------------------*/
int fast_ldap_getestimator(const char *host,
			   const char *functionpath,
			   const char *stringargs,
			   /*OUT*/fast_estimator_t ***retrieved)
{
  char *filter=NULL; /* not used, stay to null and is passed so. Maybe on day*/
  char *base=NULL; /*for example ou=n!n!1!0,ou=dgemm,ou=BLAS,ou=blaise,ou=...*/
  char *p;/* pointer where to write in base */
  char *pathcpy;/*to modify it*/
  char *q; /* pointer to the last '/' in pathcpy */
  char **estimstring;
  int i;

  if (!fast_ldap_use) /* Don't use LDAP */
    return 0;

  if (!(*retrieved=malloc(sizeof(fast_estimator_t*)*FAST_NEED_TYPE_COUNT))) 
    FAIL("Malloc error\n");

  /* not found, allocate an entry */
  if (!(pathcpy=strdup(functionpath)) ||
      !(base=malloc(sizeof(char)*
		    ((stringargs ? strlen(stringargs) : 0)+
		     strlen(fast_ldap_basedn)+
		     strlen(functionpath)+256))))
    FAIL("Out of memory\n");
  /* add the stringargs to the base of the query */
  p=base;
  if (stringargs && stringargs[0]!='\0') {
    sprintf(base,"ou=%s,",stringargs);
    while (*(++p) != '\0');
  }
  DDEBUG1("base:%s\n",base?base:"(null)");
  /* add each part of the function path to the base */
  while ((q=strrchr(pathcpy,'/'))) {
    sprintf(p,"ou=%s,",q+1);
    DDEBUG2("base:%s;q:%s\n",base,q);
    *q='\0';
    while (*(++p) != '\0');      
  }
  sprintf(p,"ou=%s,",pathcpy);
  while (*(++p) != '\0');
  free(pathcpy);
  /* add the hostname and Hardware */
  if ((q=strchr(host,'.'))) // remove anything after the first '.'
    snprintf(p,4+q-host,"ou=%s",host);
  else
    sprintf(p,"ou=%s",host);
  while (*(++p) != '\0');
  sprintf(p,",ou=Hardware");
  
  /* Launch the query */
  if (!fast_ldap_search(filter,base,&estimstring)) {
    free(base);
    return 0;
  }
  free(base);
  
  /* store the results (nb: we search (=prefetch) all need estimator) */
  for (i=0;i<FAST_NEED_TYPE_COUNT;i++) {
    if (estimstring[i]) {
      (*retrieved)[i]=fast_estimator_alloc_from_string(estimstring[i]);
    } else {
      (*retrieved)[i]=NULL;
    }
  }
   
  return 1;
}
  

/*----[ fast_ldap_search ]---------------------------------------------------*/
/* Do a search                                                               */
/*---------------------------------------------------------------------------*/
int fast_ldap_search(char *filter,char *base,
		     /*OUT*/char ***where)
{
  char *entry; /* attribute name while printing a search entry */
  BerElement *ber = NULL; /* to print a search entry */
  char **result; /* result of the search */
  /* to print the search result */
  int err;
  char *matcheddn = NULL;
  char *text = NULL;
  char **refs = NULL;
  LDAPControl **ctrls = NULL;

  LDAPMessage	*msg,*reponse;
  ber_int_t msgid;
  int rc; /* return code of functions */
  char *basedn; /* #base#+fast_ldap_defaultdn */
  int i;

  /* malloc place for the result */
  if (!(*where=malloc(sizeof(char*)*FAST_NEED_TYPE_COUNT)))
    FAIL("Out of memory\n");
  for (i=0;i<FAST_NEED_TYPE_COUNT;i++) 
    (*where)[i]=NULL;

  /* build the new base by aggregating the basedn of fast to the given base */
  if (base) {
    if (!(basedn=malloc(sizeof(char)*(strlen(fast_ldap_basedn)+
				      strlen(base) + 
				      2 /* ',' + '\0' */))))
      FAIL("Out of memory\n");
    sprintf(basedn,"%s,%s",base,fast_ldap_basedn);
  } else {
    basedn=fast_ldap_basedn;
  }

  LOG2("Do a LDAP search. Filter=`%s' and basedn=`%s'.\n",filter,basedn);

  /* Run the search */
  rc=ldap_search_ext(fast_ldap_socket, basedn, LDAP_SCOPE_SUBTREE, filter, 
		     (char **)fast_ldap_need_name, 0/* 1 => only the attributes */,
		     /* the next args are so in ldapsearc.c*/
		     NULL, NULL, NULL, -1,
		     &msgid);

  if (rc != LDAP_SUCCESS) {
    ERROR3("Error in ldap with filter `%s' and dn `%s': %s'\n",
	   filter,basedn,
	   ldap_err2string(rc));
    if (base) free(basedn);
    return 0;
  }
  /* extract the results */
  while ((rc = ldap_result( fast_ldap_socket, LDAP_RES_ANY, LDAP_MSG_ONE,
			    NULL, &reponse )) > 0 ) {
    /* for each part of the result */
    for ( msg = ldap_first_message( fast_ldap_socket, reponse );
	  msg != NULL;
	  msg = ldap_next_message( fast_ldap_socket, msg ) ) {
      DDEBUG("Got a message from LDAP\n");
      switch( ldap_msgtype( msg ) ) {
      case LDAP_RES_SEARCH_ENTRY:

	for ( entry = ldap_first_attribute( fast_ldap_socket, msg, &ber ); 
	      entry != NULL;
	      entry = ldap_next_attribute( fast_ldap_socket, msg, ber ) ) {

	  /* search the corresponding need */
	  for (i=0;
	       i<FAST_NEED_TYPE_COUNT &&
		 strcmp(entry,fast_ldap_need_name[i]);
	       i++); /* nothing more to do */
	  if (i==FAST_NEED_TYPE_COUNT) {
	    if (base) free(basedn);
	    FAIL1("LDAP returned an unwanted entry `%s'\n",entry);
	  }

	  /* put the answer in place */
	  result=ldap_get_values(fast_ldap_socket, msg, entry);
	  if (ldap_count_values(result) == 1) {
	    DDEBUG2("Read %s=%s\n",entry,result[0]);
	    (*where)[i]=strdup(result[0]);
	  } else {
	    WARN4("%s%d %s%s",
		  "LDAP gave ",ldap_count_values(result),
		  "value(s) instead of 1. ",
		  "Are you sure of the spelling of the question ?\n");
	  }
	  ldap_value_free(result);
	}
	if( ber != NULL ) {
	  ber_free( ber, 0 );
	}
	
	break;
	
      case LDAP_RES_SEARCH_RESULT:

	rc = ldap_parse_result( fast_ldap_socket, msg,
				&err, &matcheddn, &text, &refs, &ctrls, 0 );

	if( rc != LDAP_SUCCESS ) {
	  if (base) free(basedn);
	  FAIL1("Error in LDAP query: %s\n",ldap_err2string(rc));
	}
	if ( err != LDAP_SUCCESS ) {
	  if (base) free(basedn);
	  ERROR2("ldap_parse_result: %s (%d)\n", ldap_err2string(err), err );
	}

	if( matcheddn && *matcheddn ) {			
	  WARN1("matchedDN= %s\n",matcheddn);
	  /*WARN2("LDAP search was: Filter=`%s' and basedn=`%s'.\n",
		filter,basedn);*/

	  ber_memfree( matcheddn );
	}

	if( text && *text ) {
	  WARN1( "Additional information: %s\n", text );
	  ber_memfree( text );
	}

	if( refs ) {
	  for( i=0; refs[i] != NULL; i++ ) {
	    WARN1("Referral: %s\n", refs[i]);
	  }

	  ber_memvfree( (void **) refs );
	}
	
	if( ctrls ) {
	  WARN("The controls are ignored\n");
	  ldap_controls_free( ctrls );
	}
	goto done; /* no message after result */
      }
      /* FIXME: other cases in ldapsearch.c which I ignore :
	 case LDAP_RES_SEARCH_REFERENCE:
	 case LDAP_RES_EXTENDED: 
	 case LDAP_RES_EXTENDED_PARTIAL: */
    }

    ldap_msgfree( msg );
  }
 done:

  /* sanity check of the result */
  for (i=0;i<FAST_NEED_TYPE_COUNT;i++)
    if ((*where)[i]==NULL) {
      WARN1("LDAP did not give the entry corresponding to need %s\n",
	    fast_ldap_need_name[i]);
    }

  /* free allocated mem */
  if (base)
    free(basedn);
  return 1;
}


/* [from openldap2.0.7/libraries/libldap/search.c 
 *   other search functions are marked deprecated in ldap.h]
 * 
 * ldap_search_ext - initiate an ldap search operation.
 *
 * Parameters:
 *
 *      ld              LDAP descriptor
 *      base            DN of the base object
 *      scope           the search scope - one of LDAP_SCOPE_BASE,
 *                          LDAP_SCOPE_ONELEVEL, LDAP_SCOPE_SUBTREE
 *      filter          a string containing the search filter
 *                      (e.g., "(|(cn=bob)(sn=bob))")
 *      attrs           list of attribute types to return for matches
 *      attrsonly       1 => attributes only 0 => attributes and values
 *
 * Example:
 *      char    *attrs[] = { "mail", "title", 0 };
 *      ldap_search_ext( ld, "c=us,o=UM", LDAP_SCOPE_SUBTREE, "cn~=bob",
 *          attrs, attrsonly, sctrls, ctrls, timeout, sizelimit,
 *              &msgid );
 *
 * Prototype:
 *   int ldap_search_ext(
 *        LDAP *ld,
 *        LDAP_CONST char *base,
 *        int scope,
 *        LDAP_CONST char *filter,
 *        char **attrs,
 *        int attrsonly,
 *        LDAPControl **sctrls,    (set to NULL in ldapsearch.c)
 *        LDAPControl **cctrls,    (set to NULL in ldapsearch.c)
 *        struct timeval *timeout, (set to NULL in ldapsearch.c)
 *        int sizelimit,           (set to -1 in ldapsearch.c)
 *        int *msgidp )
 *
 */

/* [from openldap2.0.7/libraries/libldap/result.c]
 *
 * ldap_result - wait for an ldap result response to a message from the
 * ldap server.  If msgid is LDAP_RES_ANY (-1), any message will be
 * accepted.  If msgid is LDAP_RES_UNSOLICITED (0), any unsolicited
 * message is accepted.  Otherwise ldap_result will wait for a response
 * with msgid.  If all is LDAP_MSG_ONE (0) the first message with id
 * msgid will be accepted, otherwise, ldap_result will wait for all
 * responses with id msgid and then return a pointer to the entire list
 * of messages.  In general, this is only useful for search responses,
 * which can be of three message types (zero or more entries, zero or
 * search references, followed by an ldap result).  An extension to
 * LDAPv3 allows partial extended responses to be returned in response
 * to any request.  The type of the first message received is returned.
 * When waiting, any messages that have been abandoned are discarded.
 *
 * Example:
 *      ldap_result( s, msgid, all, timeout, result )
 * 
 * Prototype:
 *  int 
 *  ldap_result(
 *          LDAP *ld, 
 *          int msgid,
 *          int all, 
 *          struct timeval *timeout,
 *          LDAPMessage **result )
 */
