/* $Id: fast_xml.c,v 1.1 2003/05/26 10:53:00 mquinson Exp $ */

/* fast_xml.c -- functions to read FAST data from xml file                   */

/* Copyright (C) 2001 the OURAGAN project. Code GPL'ed (see www.gnu.org).    */
/* Author(s): Martin Quinson                                                 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


#include "fast_diagnostic.h"
#include "fast_xml.h" /* prototypes of this file */
#include "config.h"

#include "../../Extract/slimfast_api.h"

#ifdef HAVE_XML2
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
/* To ask xml to check validity against the dtd */
extern int xmlDoValidityCheckingDefaultValue;



/* the order of tags in a XML file is not specified. But I need some part
   to be defined before I can parse some other parts. That is to say that I
   need the high level description before I can read the benching stuff.
   That's why I have to parse the file in a big struct, and then convert it
   to the regular smaller ones */

typedef enum {
  C = 0
} fast_xml_language_t;
#define fast_xml_language_count 1
const char *fast_xml_language_name[fast_xml_language_count]={"C"};

typedef struct {
  char *name;
  /* description part: the data */
  char           **descInput_name;
  sf_data_desc_t **descInput;
  int              descInput_count;

  char           **descOutput_name;
  sf_data_desc_t **descOutput;
  int              descOutput_count;

  /* FIXME: descCode (ie, high level code) ignored for now */

  /* Implementation part: the data */
  sf_impl_desc_t *impl;
  char               **implInput_name;
  sf_impl_desc_arg_t **implInput;
  int                  implInput_count;
  

  /* Implementation part: the code */
  fast_xml_language_t language;
  char *implDecl;
  char *benchBody;
  char *implBody;
} fast_xml_problem_t;

struct fast_xml_problemSet_ {
  char *name;
  int count;
  fast_xml_problem_t **pbs;
};

/*----[ fast_xml_problem_parse ]---------------------------------------------*/
static int fast_xml_problem_parse(fast_xml_problemSet_t *pbSet,
				  xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
				  /*OUT*/fast_xml_problem_t **where);

/*----[ fast_xml_problem_free ]----------------------------------------------*/
static void fast_xml_problem_free(fast_xml_problem_t *pb);
/*----[ fast_xml_problemSet_free ]-------------------------------------------*/
static void fast_xml_problemSet_free(fast_xml_problemSet_t *pbs);

/*====[ Code ]===============================================================*/
#define CUR_NAME_IS(cur,search_name,search_ns)\
    (!xmlStrcmp(cur->name,(const xmlChar *) search_name) &&\
	(cur->ns == search_ns))

/*----[ fast_xml_problem_parse ]---------------------------------------------*/
static int fast_xml_problem_parse(fast_xml_problemSet_t *pbSet,
				  xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
				  /*OUT*/fast_xml_problem_t **where)
{
  char *dummy;
  int i;
  xmlNodePtr lvl1,lvl2;

  if (!((*where)=malloc(sizeof(fast_xml_problem_t))))
    FAIL("Out of memory\n");
  memset(*where,0,sizeof(fast_xml_problem_t));

  (*where)->name=xmlGetProp(cur,"name");
  printf("Read %s\n",(*where)->name);
  cur = cur->xmlChildrenNode;
  while (cur && xmlIsBlankNode (cur))
         cur = cur->next;
  while (cur != NULL) {

    DDEBUG1("See %s.\n",cur->name);

    /* Parse the implementation */
    if (CUR_NAME_IS(cur,"implementation",ns)) {
      dummy=xmlGetProp(cur,"language");
      for (i=0;i<fast_xml_language_count;i++)
	if (!strcmp((char *)fast_xml_language_name[i],dummy))
	  break;
      if (i==fast_xml_language_count) {
	ERROR3("Problem %s/%s has an unknown implementation language: %s\n",
	      pbSet->name,(*where)->name,dummy);
	free(dummy);
	return 0;
      }
      free(dummy);
      (*where)->language=i;

      lvl1=cur->xmlChildrenNode;
      while (lvl1 && xmlIsBlankNode(lvl1)) lvl1=lvl1->next;
      
      while (lvl1) {
	DDEBUG2("See %s:%s.\n",cur->name,lvl1->name);

	/* Parse the implementation description */
	if (CUR_NAME_IS(lvl1,"implCode",ns)) {
	  lvl2=lvl1->xmlChildrenNode;
	  while (lvl2 && xmlIsBlankNode(lvl2)) lvl2=lvl2->next;
	  while (lvl2) {

	    DDEBUG3("See %s:%s:%s.\n",cur->name,lvl1->name,lvl2->name);
	    /* Get the declarations */
	    if (CUR_NAME_IS(lvl2,"implDecl",ns)) {
	      (*where)->implDecl = 
		xmlNodeListGetString(doc, lvl2->xmlChildrenNode, 1);
	    /* Get the bench addendum */ 
	    } else if (CUR_NAME_IS(lvl2,"benchBody",ns)) {
	      (*where)->benchBody = 
		xmlNodeListGetString(doc, lvl2->xmlChildrenNode, 1);
	    /* Get the body */
	    } else if (CUR_NAME_IS(lvl2,"implBody",ns)) {
	      (*where)->implBody = 
		xmlNodeListGetString(doc, lvl2->xmlChildrenNode, 1);
	    }
	    lvl2=lvl2->next;
	  }
	}
      lvl1=lvl1->next;
      }
    }
    cur=cur->next;
  }
  return 1;
}

/*----[ fast_xml_parseFile ]-------------------------------------------------*/
/* Parse #file# and returns if anything went ok                              */
/*---------------------------------------------------------------------------*/
int fast_xml_parseFile(const char *filename,
		       /*OUT*/fast_xml_problemSet_t **where)
{
  xmlDocPtr doc;
  xmlNsPtr ns;
  xmlNodePtr cur;
  
  DDEBUG1("Parse file `%s'\n",filename);

  /*
   * XML checks
   */
  xmlDoValidityCheckingDefaultValue = 1;
  LIBXML_TEST_VERSION;
  xmlKeepBlanksDefault(0);


  /*
   * build an XML tree from a the file;
   */
  if (!(doc = xmlParseFile(filename)))
    FAIL1("Unable to parse file %s\n",filename);
  
  /*
   * Check the document is of the right kind
   */
    
  if (!(cur = xmlDocGetRootElement(doc))) {
    xmlFreeDoc(doc);
    FAIL1("File %s contains an empty document\n",filename);
  }

  if (xmlStrcmp(cur->name, (const xmlChar *) "problemSet")) {
    xmlFreeDoc(doc);
    FAIL1("File %s malformed, root node != problemSet",filename);
  }

  ns = xmlSearchNsByHref(doc, cur,
            (const xmlChar *) "http://www.ens-lyon.fr/~mquinson/fast");
  if (ns == NULL) {
    xmlFreeDoc(doc);
    FAIL1("File %s of the wrong type, FAST Namespace not found\n",
	  filename);
  }

  /* 
   * Parse the tree
   */
  if (!(*where=(fast_xml_problemSet_t*)malloc(sizeof(fast_xml_problemSet_t))))
    FAIL("Out of memory");

  (*where)->name=xmlGetProp(cur, (const xmlChar *) "name");
  (*where)->count=0;
  (*where)->pbs=NULL;

  cur = cur->xmlChildrenNode;
  while (cur && xmlIsBlankNode (cur))
         cur = cur->next;

  /* FIXME: general settings ignored for now */
  while (cur != NULL) {

    if (CUR_NAME_IS(cur,"problem",ns)) {
      if (!((*where)->pbs=
	    (fast_xml_problem_t **)realloc((*where)->pbs,
					   sizeof(fast_xml_problem_t *)*
					   (++(*where)->count)))) {
	xmlFreeDoc(doc);
	(*where)->count--;
	fast_xml_problemSet_free(*where);
	FAIL("Out of memory\n");
      }

      if (!fast_xml_problem_parse(*where,doc, ns, cur,
				  &((*where)->pbs[(*where)->count-1]))) {
	fast_xml_problemSet_free(*where);
	xmlFreeDoc(doc);
	return 0;
      }

    }
    cur = cur->next;
  }
  
  xmlFreeDoc(doc);
  return 1;
}

/*----[ fast_xml_problem_free ]----------------------------------------------*/
void fast_xml_problem_free(fast_xml_problem_t *pb)
{
  if (pb) {
    free(pb->name);
    free(pb->implDecl);
    free(pb->benchBody);
    free(pb->implBody);
  }
}

/*----[ fast_xml_problemSet_free ]-------------------------------------------*/
void fast_xml_problemSet_free(fast_xml_problemSet_t *pbs)
{
  int i;

  if (pbs) {
    free(pbs->name);
    for (i=0; i<pbs->count; i++)
      fast_xml_problem_free(pbs->pbs[i]);
    free(pbs->pbs);
  }
}

/*----[ fast_xml_print ]-----------------------------------------------------*/
/* debuging output                                                           */
/*---------------------------------------------------------------------------*/
int fast_xml_print(fast_xml_problemSet_t *cur)
{
  int i;
  printf("ProblemSet %s: %d member%s\n",
	 cur->name,cur->count,cur->count>1?"s":"");
  for (i=0; i<cur->count; i++) {
    printf("----[ Problem %s ]-----------------------\n",cur->pbs[i]->name);
    printf("void to_%s_%s(){\n",cur->name,cur->pbs[i]->name);
    if (cur->pbs[i]->implDecl)
      printf("  /*implDecl*/\n%s\n",cur->pbs[i]->implDecl);
    if (cur->pbs[i]->benchBody)
      printf("  /*benchBody*/\n%s\n",cur->pbs[i]->benchBody);
    if (cur->pbs[i]->implBody)
      printf("  /*implBody*/\n%s\n",cur->pbs[i]->implBody);
    
    printf("}\n");
    printf("----[ End of Problem %s ]----------------\n",cur->pbs[i]->name);
  }
  return 1;
}

#else /* don't have XML2 */
/*----[ fast_xml_parseFile ]-------------------------------------------------*/
/* Parse #file# and returns if anything went ok                              */
/*---------------------------------------------------------------------------*/
int fast_xml_parseFile(const char *filename,
		       /*OUT*/fast_xml_problemSet_t **where)
{
  FAIL("Can't use XML if XML is not installed (installation problem)\n");
}
/*----[ fast_xml_print ]-----------------------------------------------------*/
/* debuging output                                                           */
/*---------------------------------------------------------------------------*/
int fast_xml_print(fast_xml_problemSet_t *cur)
{
  FAIL("Can't use XML if XML is not installed (installation problem)\n");
}
#endif /* HAVE_XML2 */
