/* $Id: cache.c,v 1.4 2005/03/05 17:55:57 mquinson Exp $ */

#include "cache.h" /* prototypes of this module */

#include <stdlib.h> /* malloc() */
#include <string.h> /* strlen() */

#include "diagnostic.h" /* FAIL() */

/*####[ Type definition ]####################################################*/
struct fast_cache_ {
  char *key;
  void *content;
  void (*free_ctn)(void*); /*pointer to the function to call to free this ctn*/

  fast_cache_t **sub; /* the sub entries */
  int sub_count;
};

/*####[ Private prototypes ]#################################################*/
/*----[ fast_cache_alloc ]---------------------------------------------------*/
/* Alloc a node with no child                                                */
/* Returns if it went ok or not.                                             */
/*---------------------------------------------------------------------------*/
static int fast_cache_alloc(char *key,void *data,void (*free_ctn)(void*),
			    /*OUT*/fast_cache_t **where);
			    
/*----[ fast_string_prefix_lgr ]---------------------------------------------*/
/* Returns the length of the common prefix of #str1# and #str2#              */
/*---------------------------------------------------------------------------*/
static int fast_string_prefix_lgr(const char *str1,const char *str2);

/*----[ fast_cache_dump_recursiv ]-------------------------------------------*/
/* private function to do the job of fast_cache_dump recursively             */
/* Returns if it was ok or not                                               */
/*---------------------------------------------------------------------------*/
static int fast_cache_dump_recursiv(fast_cache_t *head,int prefix_lgr,
				    void (*output)(void*));

/*####[ utility functionse ]#################################################*/
/*----[ fast_string_prefix_lgr ]---------------------------------------------*/
/* Returns the length of the common prefix of #str1# and #str2#              */
/*---------------------------------------------------------------------------*/
static int fast_string_prefix_lgr(const char *str1,const char *str2)
{
  int i=0;

  if (!str1 || !str2)
    return 0;
  
  while (str1[i] == str2[i] && str1[i] != '\0')
    i++;

  return i;
}


/*####[ Memory management functions ]########################################*/
/*----[ fast_cache_alloc ]---------------------------------------------------*/
/* Alloc a node with no child                                                */
/* Returns if it went ok or not.                                             */
/*---------------------------------------------------------------------------*/
static int fast_cache_alloc(char *key,void *data,void (*free_ctn)(void*),
			    /*OUT*/fast_cache_t **where)
{
  if (!((*where)=malloc(sizeof(fast_cache_t)))) {
    if (free_ctn && data)
      free_ctn(data);
    FAIL("Out of memory\n");
  }

  (*where)->key=key;
  (*where)->content=data;
  (*where)->free_ctn=free_ctn;

  (*where)->sub=NULL;
  (*where)->sub_count=0;

  return 1;
}

/**
 * fast_cache_free:
 *
 * @cache: the cache to be freed
 *
 * Frees a cache structure with all its childs. 
 */


int fast_cache_free(fast_cache_t **cache) 
{
  int i;

  if (*cache) {
    for (i=0;i<(*cache)->sub_count;i++)
      if (!fast_cache_free(&( (*cache)->sub[i] )))
	return 0;
    if ((*cache)->key) free((*cache)->key);
    if ((*cache)->sub) free((*cache)->sub);
    if ((*cache)->free_ctn && (*cache)->content) 
      (*cache)->free_ctn((*cache)->content);

    free(*cache);
    *cache=NULL;
  }

  return 1;
}

/*----[ fast_cache_add_child ]-----------------------------------------------*/
/* Allocate a place in the list of childs of #father#, and let it point to   */
/*  child.                                                                   */
/* returns if it was ok or not.                                              */
/*---------------------------------------------------------------------------*/
static int fast_cache_add_child(fast_cache_t *father,
				fast_cache_t *child)
{
   father->sub=realloc(father->sub,
		       sizeof(fast_cache_t**)*(++ father->sub_count));
   if (!father->sub)
     FAIL("Out of memory\n");

   father->sub[father->sub_count -1] = child;

   return 1;
}

/**
 * fast_cache_insert:
 *
 * @head: the head of the cache
 * @key: the key to insert the new data
 * @data: the data to add in the cache
 * @Returns: true, if it was ok, false if not
 *
 * Insert the @data in the structure under the @key. The @key is destroyed
 * in the process. Think to strdup it before.
 */

int fast_cache_insert(fast_cache_t **head,
		      char *key,
		      void *data,void (*free_ctn)(void*)) 
{
  int i;
  int prefix=0;
  char *ancestor_key;
  fast_cache_t *ancestor;
  fast_cache_t *toadd; /* the child we want to add. */
  fast_cache_t *child; /* the child having a common prefix */

  /* Well, the algorithm of this function is not so trivial, so I'll explain
     it in comments. To insert something in the struct, you have to check if
     an existing child having a common prefix with the key to add exists, and 
     distinguish a few cases. Let's take some examples.

     First, trivial cases:
       if head does not exists yet, create it and place #to_add# as child.
       if there is no key, change the head's value.
       if no child have a common prefix with #to_add#, create a new child.

     Then, the not so trivial cases:
       [abc;abg] create a common ancestor 'ab', put #to_add# and #child# as
         child of the common ancestor (cutting their key)
       [ab;abg] #to_add# become a child of head, and #child# a child of
         #to_add# (cutting its key)
       [abc;ab] #to_add# is a child of #child# => recusion (with cutted key)

       [ab;ab] (recurse and then change the head (ie, child's) value)

     Note:
       [abc;abg] means that I want to add 'abc', and a child 'abg' exists.
  */

  /* there is no head, create it */
  DDEBUG3("--> Insert '%s' after '%s' in tree %p\n",
	  key,(((*head) && ((*head)->key)) ? ((*head)->key):"(head)"),
	  (head ? (void*)*head : (void*)head));
  if (!(*head)) {
    DDEBUG("-> Add a head\n");
    if (!fast_cache_alloc(strdup(""),NULL,NULL,head) ||
	!fast_cache_alloc(key,data,free_ctn,&toadd) || 
	!fast_cache_add_child(*head,toadd))
      return 0;
    
    return 1;
  }

  /* there is no key, change the value of head */
  if (!key || key[0]=='\0') {
    DDEBUG("--> Change the value of head\n");
    if ((*head)->free_ctn)
      (*head)->free_ctn((*head)->content);
    (*head)->free_ctn = free_ctn;
    (*head)->content = data;

    return 1;
  }
    
  /* search the relevant entry between my sub trees */
  for (i=0;i<(*head)->sub_count;i++) {
    DDEBUG2("---> Compare %s and %s\n",key,(*head)->sub[i]->key);
    if ((prefix=fast_string_prefix_lgr(key,(*head)->sub[i]->key)))
      break;
  }

  /* no child have a common prefix with the one added. */
  if (prefix==0) {
    DDEBUG("-> Add a child\n");
    (*head)->sub=realloc((*head)->sub,
			 sizeof(fast_cache_t**)*(++ (*head)->sub_count));
    if (!(*head)->sub)
      FAIL("Out of memory\n");

    return fast_cache_alloc(key,data,free_ctn,
			    &((*head)->sub[(*head)->sub_count -1]));
  }

  child=(*head)->sub[i];

  /* the common prefix is the value of the child. Recurse on this child */
  if (prefix==strlen(child->key)) {
    memmove(key,key+prefix,strlen(key)-prefix+1);
    DDEBUG1("-> Recurse (newkey=%s)\n",key);

    return fast_cache_insert(&( (*head)->sub[i] ), key,data,free_ctn);
  } 

  /* the common prefix is the key I want to add. child become child of toadd */
  if (prefix==strlen(key)) {
    DDEBUG("-> The child become child of toadd\n");
    memmove(child->key, child->key+prefix, strlen(child->key)-prefix+1);
    
    if (!fast_cache_alloc(key,data,free_ctn, &( (*head)->sub[i]) )) {
      (*head)->sub[i]=child;
      return 0; /* output comes from cache_alloc */
    }
    
    if (!fast_cache_add_child((*head)->sub[i],child)) {
      (*head)->sub[i]=child;
      return 0;
    }
    return 1;
  }
      
  /* we need to make a common ancestor */
  DDEBUG("-> Make a common ancestor\n");
  if (!(ancestor_key=malloc(sizeof(char)*(prefix+1))))
    FAIL("Out of memory\n");
  strncpy(ancestor_key,key,prefix);
  ancestor_key[prefix]='\0';

  memmove(key,key+prefix,strlen(key)-prefix+1);
  memmove(child->key, child->key+prefix, strlen(child->key)-prefix+1);
  if (!fast_cache_alloc(ancestor_key,NULL,NULL,&ancestor) ||

      !fast_cache_alloc(key,data,free_ctn,&toadd) ||
      !fast_cache_add_child(ancestor,toadd) || 

      !fast_cache_add_child(ancestor,child) )
    return 0;

  (*head)->sub[i]=ancestor;
  return 1;
}

/**
 * fast_cache_retrieve:
 *
 * @head: the head of the cache
 * @key: the key to find data
 * @data: the data that we are looking for
 * @Returns: True if everythink went OK, false on internal error
 *
 * Search the given @key. data=NULL when not found. 
 */


int fast_cache_retrieve(fast_cache_t *head,const char *key,
			/* OUT */void **data)
{
  fast_cache_t *curr=head;
  int idx=0; /* index in key, We are in the right subtree concerning 
		  0..idx, and are looking for idx+1..end */
  int sub_nb; /* loop counter on child indexes */
  char *sub_key;

  int i; /* loop counter on chars in keys */

  *data=NULL; // Make it ready to answer 'not found' in one operation

  if (!head)
    return 1;
  while (1) { /* recurse in the tree until found or end of tree */

    /* Search in which subtree we will recurse*/
    /* Note: We use any branch seen with a NULL key */
    sub_nb=0; i=0;
    while (sub_nb<curr->sub_count && 
	   (sub_key=curr->sub[sub_nb]->key)) {
      i=0;
      while(sub_key[i] == key[idx+i] && 
	    sub_key[i] != '\0' && key[idx+i] != '\0')
	i++;

      if (sub_key[i] == '\0') // Follow this branch
	break;

      if (key[idx+i] == '\0') // Key were not found
	return 1;

      // Test next child
      sub_nb++;
    }
    if (sub_nb == curr->sub_count)  // No child correspond
      return 1;

    if (key[idx+i] == '\0'){/* Key and sub_key reached \0 at the same point
				 It's found !! */
      *data=curr->sub[sub_nb]->content;
      return 1;
    }

    //Recurse on this child
    idx += i;
    curr=curr->sub[sub_nb];
  }
  return 0;
}

/*
 * fast_cache_remove:
 *
 * @head: the head of the cache
 * @key: the key of the data to be removed
 * @Returns: True if everything went OK, false on internal error
 *
 * Remove the entry associated with the given @key
 */


int fast_cache_remove(fast_cache_t *head,const char *key)
{
  fast_cache_t *curr=head;
  fast_cache_t *child; /* used to collapse tree when needed */
  int idx=0; /* index in key, We are in the right subtree concerning 
		  0..idx, and are looking for idx+1..end */
  int sub_nb; /* loop counter on child indexes */
  char *sub_key;

  int i; /* loop counter on chars in keys */

  if (!head)
    return 1;
  while (1) { /* recurse in the tree until found or end of tree */

    /* Search in which subtree we will recurse*/
    /* Note: We use any branch seen with a NULL key */
    sub_nb=0; i=0;
    while (sub_nb<curr->sub_count && 
	   (sub_key=curr->sub[sub_nb]->key)) {
      i=0;
      while(sub_key[i] == key[idx+i] && 
	    sub_key[i] != '\0' && key[idx+i] != '\0')
	i++;

      if (sub_key[i] == '\0') // Follow this branch
	break;

      if (key[idx+i] == '\0') // Key were not found
	return 0;
     
      if (i>0) // Common prefix, but does not match completly. Wont found
	return 0;

      // Test next child
      sub_nb++;
    }
    if (sub_nb == curr->sub_count)  // No child correspond
      return 0;

    if (key[idx+i] == '\0'){/* Key and sub_key reached \0 at the same point
				 It's found !! */
      DDEBUG3("Found! curr=['%s','%s',%d]\n",
	      curr->key,curr->content?curr->content:"(null)",curr->sub_count);

      if (curr->sub[sub_nb]->content && curr->sub[sub_nb]->free_ctn) {
	/* free content if possible */
	curr->sub[sub_nb]->free_ctn(curr->sub[sub_nb]->content);
      } else {
	DDEBUG("Can't remove internal entry on user behalf\n");
	/* it's an internal entry (because there is no associated data)
	   I can't free it on user behalf */
	return 0;
      }

      child=curr->sub[sub_nb];

      /* free the corresponding child iff it's a leaf */
      if (child->sub_count == 0) {
	DDEBUG("Remove a leaf\n");
	free (child->key);
	free (child);
	curr->sub[sub_nb] = curr->sub[-- curr->sub_count];
	if (curr->sub_count)
	  curr->sub=realloc(curr->sub,sizeof(fast_cache_t**)*curr->sub_count);
	else {
	  free(curr->sub);
	  curr->sub=NULL;
	}
      } else {
	/* we don't have to free the child. Make sure it's valid */
	curr->sub[sub_nb]->free_ctn =  NULL;
	curr->sub[sub_nb]->content = NULL;
      }
      
      /* Collapse the tree between curr and child if needed.
	 Ie, if child had at most a brother */
      if (curr->sub_count == 1 && (curr->content==NULL) 
	  && curr->key[0]!='\0') {
	DDEBUG("Collapse curr and child\n");

	child=curr->sub[0];
	if (child->content || child->sub_count>1) {
	  /* Concatenate the key of the child to current */
	  if (!(sub_key=malloc(sizeof(char)*
			       (strlen(curr->key)+strlen(child->key)+1)))) {
	    /* we can't collapse the tree, but the struct remains valid */
	    return 1;
	  }
	  sprintf(sub_key,"%s%s",curr->key,child->key);
	  free(curr->key);
	  free(child->key);
	  curr->key=sub_key;
	} else {
	  /* concatenate the key of child to its uniq child if exists */
	  if (child->sub_count==1) {
	    if (!(sub_key=malloc(sizeof(char)*
				 (strlen(child->key)+
				  strlen(child->sub[0]->key)+
				  1)))) {
	      /* we can't collapse the tree, but the struct remains valid */
	      return 1;
	    }
	    sprintf(sub_key,"%s%s",child->key,child->sub[0]->key);
	    free(child->sub[0]->key);
	    free(child->key);
	    child->sub[0]->key=sub_key;
	  }
	}

	if (child->content) {
	  curr->content = child->content; 
	  curr->free_ctn = child->free_ctn; 
	}

      	free(curr->sub);
      	curr->sub = child->sub;
      	curr->sub_count = child->sub_count;

      	free(child);
      }

      /* Collapse the tree between child and its child if needed.
	 Ie, if child had a unique child */

      return 1;
    }

    //Recurse on this child
    idx += i;
    curr=curr->sub[sub_nb];
  }
}

/*----[ fast_cache_dump_recursiv ]-------------------------------------------*/
/* private function to do the job of fast_cache_dump recursively             */
/* Returns if it was ok or not                                               */
/*---------------------------------------------------------------------------*/
static int fast_cache_dump_recursiv(fast_cache_t *head,int prefix_lgr,
				    void (*output)(void*))
{
  int i;

  if (!head) return 1;
  printf("#");
  for (i=0;i<prefix_lgr;i++)
    printf(" ");
  if (head->key)
    printf("%s(%p) -> ",head->key,head);
  if (output && head->content) {
    output(head->content);
  } else {
    printf("(null)");
  }
    printf("    \t\t\t[ %d child(s) ]",head->sub_count);
  printf("\n");
  for (i=0;i<head->sub_count;i++)
    if (!fast_cache_dump_recursiv(head->sub[i], prefix_lgr +
				  (head->key ? strlen(head->key) : 0),
				  output))
      return 0;
  return 1;
}

/**
 * fast_cache_dump:
 *
 * @head: the head of the cache
 * @Returns: True if enverything went OK, false if not
 * 
 * Ouputs the content of the structure. (for debuging purpose). @ouput is a 
 * function to output the data. If NULL, data won't be displayed.
 */

int fast_cache_dump(fast_cache_t *head,
		    void (*output)(void*))
{
  printf("Tree %p:\n",head);
  return fast_cache_dump_recursiv(head,0,output);
}

/**
 * fast_cache_print:
 *
 * @data:
 *
 * To dump multicache, this function dumps a cache
 */

void fast_cache_print(void *data) {
  printf("tree %p",data);
}

/**
 * fast_cache_prints:
 *
 * @data:
 *
 * To dump multicache, this one dumps a string
 */

void fast_cache_prints(void *data) {
  printf("(somedata)");
}

/*####[ Multi cache functions ]##############################################*/
/* The are cache of cache of data. Any function there works the same way     */
/*  than their simple cache counterpart.                                     */
/*###############################"###########################################*/
/*----[ free_cache ]---------------------------------------------------------*/
/* a wrapper to free cache with the right prototype.                         */
/*---------------------------------------------------------------------------*/
static void free_cache(void*);

/*====[ Code of private functions ]==========================================*/
static void free_cache(void* cache) {
  fast_cache_free( (fast_cache_t**) (&cache) );
}

/**
 * fast_muticache_insert:
 *
 * @head: the head of cache
 * @keycount: the number of the key
 * @key: the key
 * @data: the data to insert
 * @Returns: True if anything went Ok, false if not.
 *
 * Insert the data in the structure under the @keycount @key. The 
 * keys are destroyed in the process. Think to strdup it before.
 */

int fast_multicache_insert(fast_cache_t **head,
			   int keycount,char **key,
			   void *data,void (*free_ctn)(void*))
{
  fast_cache_t *cache;
  int i;
  void *cachedata=NULL;

  /*  printf("fast_multicache_insert(%p,%d",*head,keycount);
      for (i=0;i<keycount;i++)
      printf(",\"%s\"",key[i]);
      printf(")\n");*/
  if (keycount <1)
    FAIL("Can't insert less than one key in a multicache\n");

  if (keycount == 1)
    return fast_cache_insert(head,key[0],data,free_ctn);

  if (*head == NULL)
    if (!fast_cache_alloc(strdup(""),NULL,NULL,head))
      return 0;
  cache=*head;

  for (i=0 ; i<keycount-1 ; i++) {
    /* search the cache of next level */
    if (!fast_cache_retrieve(cache,key[i],&cachedata)) {
      ERROR1("Internal error while retrieving cache level %d\n",i);
      return 0;
    }
    /* make sure the cache of next level exists */
    if (cachedata==NULL) {
      if (!fast_cache_alloc(strdup(""),NULL,NULL,(fast_cache_t**)&cachedata) ||
	  !fast_cache_insert(&cache,key[i],cachedata,&free_cache))
	return 0;
    }

    cache=(fast_cache_t*)cachedata;
  }


  return fast_cache_insert(&cache,key[i],data,free_ctn);
}

/**
 * fast_muticache_retrieve:
 *
 * @head: the head of cache
 * @keycount: the number of the key
 * @key: the key
 * @data: the data to insert
 * @Returns: True if anything went Ok, false if not.
 * 
 * Search the given @key. data=NULL when not found
 */ 


int fast_multicache_retrieve(fast_cache_t *head,
			     int keycount,const char **key,
			     /* OUT */void **data)
{
  fast_cache_t *cache=head;
  int i=0;

  /*  printf("fast_multicache_retrieve(%p,%d",head,keycount);
      for (i=0;i<keycount;i++)
      printf(",\"%s\"",key[i]);
      printf(")\n");
      i=0;*/
  while (cache && i<keycount-1) {
    if (!fast_cache_retrieve(cache, key[i], (void**)&cache))
      FAIL("Internal error\n");
    if (cache) {
      DDEBUG3("Found level %d for key %s in multitree %p\n",i,key[i],head);
    } else {
      DDEBUG3("NOT found level %d for key %s in multitree %p\n",i,key[i],head);
    }
    i++;
  }
  if (cache) { // Found all caches to the data
    //    fast_cache_dump(cache,&fast_cache_prints);
    return fast_cache_retrieve(cache, key[i], data);
  } else {
    *data=NULL;
    return 1;
  }
}

/**
 *  fast_muticache_remove:
 *  
 *  @head: the head of cache
 *  @keycount: the number of the key
 *  @key: the key
 *  @Returns: True if anything went Ok, false if not.
 *  
 * Remove the entry associated with the given @key
 * Removing a non-existant key is ok. 
 */ 

int fast_multicache_remove(fast_cache_t *head,
			   int keycount,const char **key)
{
  fast_cache_t *cache=head;
  int i=0;

  while (cache && i<keycount-1) {
    if (!fast_cache_retrieve(cache, key[i], (void**)&cache))
      return 0;
  }
  if (cache) // Found all caches to the data
    return fast_cache_remove(cache, key[i]);
  else 
    return 1;
}

/*####[ Cache cursor functions ]#############################################*/
/* To traverse (simple) caches                                               */
/* Don't add or remove entries to the cache while traversing !!!             */
/*###########################################################################*/
struct fast_cache_cursor_ {
  /* we store all level encountered until here, to backtrack on next() */
  fast_cache_t **cache;
  int *num;  /* num[0] is the pos of cache[1] between childs of cache[0] */
  int maxsize; /* allocated size */
  int size;    /* used size */
};

#undef FAST_CACHE_CURSOR_DEBUG 
//#define FAST_CACHE_CURSOR_DEBUG 1

/* creator/destructor */

/**
 * fast_cache_cursor_new:
 *
 * @head: the head of the cache
 * @cursor: the curent position in the cache
 */


int fast_cache_cursor_new(fast_cache_t *head,
			  /*OUT*/fast_cache_cursor_t **cursor) {
  fast_cache_cursor_t *res=malloc(sizeof(fast_cache_cursor_t));

  if (!res)
    FAIL("Malloc error: out of memory\n");

  if (!(res->cache = malloc(sizeof(fast_cache_t*)*10)) ||
      !(res->num = malloc(sizeof(int)*10))) {
    free (res);
    FAIL("Malloc error: out of memory\n");
  }

  res->maxsize=10;
  res->cache[0]=head;
  *cursor=res;
  return fast_cache_cursor_rewind(*cursor);
}

void fast_cache_cursor_free(fast_cache_cursor_t *cursor) {
  if (cursor) {
    free(cursor->cache);
    free(cursor->num);
    free(cursor);
  }
}

/* back to first element 
   it is not enough to reinit the cache after an add/remove in cache*/

/**
 * fast_cache_cursor_rewind:
 *
 * @cursor: the cursor
 * @Returns: true if everything went Ok, false if an error occur
 *
 * back to the first element
 */

int fast_cache_cursor_rewind(fast_cache_cursor_t *cursor) {
#ifdef FAST_CACHE_CURSOR_DEBUG
  printf("fast_cache_cursor_rewind\n");
#endif

  cursor->size=1;
  if (!cursor->cache[0])
    return 1; /* head=NULL */

  /* While there is no data here (ie, internal node) and there is childs */
  while (!(cursor->cache[cursor->size -1]->content) &&
	 cursor->cache[cursor->size -1]->sub_count) {

#ifdef FAST_CACHE_CURSOR_DEBUG
    printf("rewind goes to cursor->size=%d\n",cursor->size);fflush(stdout);
#endif

    if (cursor->size == cursor->maxsize) {
      cursor->maxsize += 10;
      cursor->cache=realloc(cursor->cache,sizeof(fast_cache_t*)*cursor->maxsize);
      cursor->num=realloc(cursor->cache,sizeof(int)*cursor->maxsize);
      if (!cursor->cache || !cursor->num) 
	FAIL("Malloc error: out of memory\n");
    }
    cursor->cache[ cursor->size ]=cursor->cache[ cursor->size -1 ]->sub[0];
    cursor->num[ cursor->size ] = 0;
    cursor->size++;
  }
  //printf("Cursor on first entry\n");
  return 1;
}
/* next element. Returns true if ok, and false on last entry (or error) */

/**
 * fast_cache_cursor_next:
 *
 * @cursor: the cursor
 * @Returns: True if everything went Ok, false if an error occur
 *
 * Move to the next element
 */

int fast_cache_cursor_next(fast_cache_cursor_t *cursor) {
#ifdef FAST_CACHE_CURSOR_DEBUG
  printf ("fast_cache_cursor_next()\n");
  printf ("cursor=%p\n",cursor);
  printf ("cursor->size=%d\n",cursor->size);
  printf ("current=%p\n",cursor->cache[cursor->size-1]);
  if (cursor->size) {
    printf("Our father have %d childs\n",cursor->cache[cursor->size-2]->sub_count);
  }
  fflush(stdout);
#endif

  if (!(cursor->cache[0])) { /* cursor on null cache */
#ifdef FAST_CACHE_CURSOR_DEBUG
    printf("Cursor on NULL cache\n");
#endif
    return 0;
  }
  /* If we're on an internal node associated with data, next=child */
  if (cursor->cache[cursor->size-1]->content &&
      cursor->cache[cursor->size-1]->sub_count) {

#ifdef FAST_CACHE_CURSOR_DEBUG
    printf("we're on an internal node[%s] with associated data\n",cursor->cache[cursor->size-1]->key);
#endif

    /* search next child having data */
    do {
      if (cursor->size == cursor->maxsize) {
	cursor->maxsize += 10;
	cursor->cache=realloc(cursor->cache,sizeof(fast_cache_t*)*cursor->maxsize);
	cursor->num=realloc(cursor->cache,sizeof(int)*cursor->maxsize);
	if (!cursor->cache || !cursor->num) 
	  FAIL("Malloc error: out of memory\n");
      }

      cursor->cache[cursor->size]=cursor->cache[ cursor->size-1]->sub[0];
      cursor->num[cursor->size] =0;
      cursor->size++;
    } while (!(cursor->cache[cursor->size-1]->content) &&
	     cursor->cache[cursor->size-1]->sub_count);

    return 1;
  }
  /* if we're at the root and there is no data attached, no next */
  if (cursor->size == 1) {
#ifdef FAST_CACHE_CURSOR_DEBUG
    printf("We're at root, and there is no data => DONE\n");
#endif
    return 0;
  }

  /* if we're the last child of our father, next=uncle */
  if (cursor->size > 1 &&
      cursor->num[ cursor->size-1 ] == cursor->cache[ cursor->size -2]->sub_count -1){
#ifdef FAST_CACHE_CURSOR_DEBUG
    printf("We're last child of our father\n");
#endif
    /* backtrace while we're on last entry of our father */
    while (cursor->size > 1 && 
	   cursor->num[ cursor->size-1 ] == cursor->cache[ cursor->size -2]->sub_count -1){
      cursor->size--;
#ifdef FAST_CACHE_CURSOR_DEBUG
      if (cursor->size > 1) {
	printf("Backtraced to level %d ; We're child %d of %d\n",cursor->size,cursor->num[cursor->size-1]+1,
	       cursor->cache[ cursor->size -2]->sub_count);     
      } else {
	printf("Backtraced to the root\n");
      }
#endif
    }
    /* if backtraced until root, no next child */
    if (cursor->size == 1) {
#ifdef FAST_CACHE_CURSOR_DEBUG
      printf("We backtracked until root=>DONE\n");
#endif
      return 0;
    }
  }
  /* get our next brother */
  cursor->num[cursor->size-1]++;
#ifdef FAST_CACHE_CURSOR_DEBUG
  printf("Get next brother %d. Size=%d\n",cursor->num[cursor->size-1],cursor->size);fflush(stdout);
#endif
  cursor->cache[cursor->size-1]= cursor->cache[cursor->size-2]->sub[ cursor->num[cursor->size-1] ];

  /* search next child having data if needed */
  while (!(cursor->cache[cursor->size-1]->content) &&
	 cursor->cache[cursor->size-1]->sub_count) {
#ifdef FAST_CACHE_CURSOR_DEBUG
    printf("We're on an internal node. Search next child with data\n");
#endif
    if (cursor->size == cursor->maxsize) {
      cursor->maxsize += 10;
      cursor->cache=realloc(cursor->cache,sizeof(fast_cache_t*)*cursor->maxsize);
      cursor->num=realloc(cursor->cache,sizeof(int)*cursor->maxsize);
      if (!cursor->cache || !cursor->num) 
	FAIL("Malloc error: out of memory\n");
    }
    
    cursor->cache[cursor->size]=cursor->cache[ cursor->size-1]->sub[0];
    cursor->num[cursor->size] =0;
    cursor->size++;
  } 
  return 1;
}

/* Get current element */

/**
 * fast_cache_cursor_get_key:
 *
 * @cursor: the cursor
 * @key: the current element
 * @Returns: True if everything went OK, false if not
 *
 * Get current key
 */

int fast_cache_cursor_get_key(fast_cache_cursor_t *cursor,
			      /*OUT*/char **key) {
  int i;
  int size=1; /* '\0' */
  char *res;
  if (!cursor->cache[0])
    return 0;
  /* Compute the length of the key */
  for (i=0; i<cursor->size; i++)
    size += strlen(cursor->cache[i]->key);
  /* Allocate it */
  if (!(res=malloc(sizeof(char)*size))) 
    FAIL("Malloc error: out of memory\n");
  res[0]='\0';
  for (i=0; i<cursor->size; i++)
    strcat(res,cursor->cache[i]->key);
  *key=res;
  return 1;
}

/**
 * fast_cache_cursor_get_data:
 * 
 * @cursor: the cursor
 * 
 * Get current data
 */


void* fast_cache_cursor_get_data(fast_cache_cursor_t *cursor) {
  return cursor->cache[cursor->size-1]->content;
}
