/* $Id: typed_hash.c,v 1.7 2003/06/05 13:59:53 pmachard Exp $ */

/* typed_hash - Hash table where all cell is of a determined type.        */

/* This is useful to build named structs, like option or property sets.   */

/* Copyright (C) 2001,2002,2003 the OURAGAN project.                      */
/* Code available under BSD-like licence (see COPYRIGHT file).            */
/* Author(s): Martin Quinson                                              */

#include "fast_internals.h"

/* fast_thc_t: the typedef corresponding to a Typed Hash Cell.

   Both data en DTD are mixed, but fixing it now would prevent me to ever
   defend my thesis. */

typedef struct {
  /* Allowed type of the cell */
  fast_thc_type_t type;
  int min,max;

  /* actual content 
     (cannot be an union because type host use both str and i) */
  int size;
  char **str;
  int *i;
  double *d;
} fast_thc_t;

static const char*fast_thc_type_name[fast_thc_type_count]=
  {"int","double","string","host"};

/* Internal stuff used in cache to free a cell */
static void fast_thash_cell_free(void *data);

/* Retrieve the cell we'll modify */
static int fast_thash_cell_get(fast_thash_t *th, const char *name,
			       fast_thc_type_t type,
			       /* OUT */ fast_thc_t **whereto);


/*----[ Memory management ]-----------------------------------------------*/

/**
 * fast_thash_init:
 *
 * @whereto: 
 *
 * Initialise an Hash table
 */


int fast_thash_init(fast_thash_t **whereto) {
  *whereto=NULL;
  return 1;
}

/**
 * fast_thash_new_from:
 *
 * @whereto: the hash table to be created
 * @tocopy: the source data
 * @Returns: true if everything went OK, false if an error occur
 *
 */

int fast_thash_new_from(fast_thash_t **whereto, fast_thash_t *tocopy) {
  fast_cache_cursor_t *cursor; 
  fast_thc_t *cell;
  char *name;
  int errcode;
  
  *whereto=NULL;
  if (!tocopy) return 1;
  
  if (!fast_cache_cursor_new(tocopy,&cursor)) {
    ERROR("error during the typed hash creation: cursor creation failed\n");
    return 0;
  }

  do {
    cell=(fast_thc_t*)fast_cache_cursor_get_data(cursor);
    if (!(errcode=fast_cache_cursor_get_key(cursor,(char**)&name))) {
      fast_cache_cursor_free(cursor);
      return errcode;
    }

    if (!(fast_thash_register(whereto,name, cell->type, cell->min, cell->max))) {
      free(name);
      fast_thash_free(whereto);
      return 0;
    }
    
    free (name);
  } while (fast_cache_cursor_next(cursor)); 

  fast_cache_cursor_free(cursor);
  //  fast_thash_dump("newly created",*whereto);
  return 1;
}
void fast_thash_free(fast_thash_t **th) {
  fast_cache_free((fast_cache_t**)th);
}

/**
 * fast_thash_dump:
 * @name: The name to give to this typed hash
 * @indent: what to write at the begining of each line (right number of spaces)
 * @th: the hash table
 *
 * Dumps a typed hash for debuging purpose
 */
void fast_thash_dump(const char *name,const char *indent,fast_thash_t *th) {
  fast_cache_cursor_t *cursor; 
  fast_thc_t *cell;
  char *key;
  int i;
  
  if (!fast_cache_cursor_new(th,&cursor)) return;
if (name) printf("%s>> Dumping of the typed hash '%s':\n",indent,name);

  do {
    if (!(fast_cache_cursor_get_key(cursor,(char**)&key))) {
      printf("Internal error while retrieving the key associated to the cursor\n");
      fast_cache_cursor_free(cursor);
      return;
    }
    printf("%s  %s:",indent,key);
    fflush(stdout);
    cell=(fast_thc_t*)fast_cache_cursor_get_data(cursor);

    printf("%d_to_%d_%s. Actual size=%d. List of values:\n",
	   cell->min,cell->max,fast_thc_type_name[cell->type],cell->size);
    free(key);

    switch (cell->type) {
    case fast_thc_int:
      for (i=0; i<cell->size; i++) printf ("%s    %d\n",indent,cell->i[i]);
      break;

    case fast_thc_double:
      for (i=0; i<cell->size; i++) printf ("%s    %f\n",indent,cell->d[i]);
      break;

    case fast_thc_string:
      for (i=0; i<cell->size; i++) printf ("%s    %s\n",indent,cell->str[i]);
      break;

    case fast_thc_host:
      for (i=0; i<cell->size; i++) printf ("%s    %s:%d\n",indent,cell->str[i],cell->i[i]);
      break;

    default:
      printf("%s    Invalid type!!\n",indent);
    }
    
  } while (fast_cache_cursor_next(cursor)); 

  if (name) printf("%s<< End of the typed hash '%s'\n",indent,name);
  fast_cache_cursor_free(cursor);
  return;
}

/**
 * fast_thash_cell:
 *
 * @data: the data to be freed
 *
 * free an hash table cell
 */

void fast_thash_cell_free(void *data) {
  fast_thc_t *c=(fast_thc_t *)data;
  int i;

  if (c->size >0){
    if (c->type == fast_thc_host ||
	c->type == fast_thc_string) {
      for (i=0; i<c->size; i++) {
	free(c->str[i]);
      }
      free(c->str);
    }
    if (c->type == fast_thc_host ||
	c->type == fast_thc_int) {
      free(c->i);
    }
    if (c->type == fast_thc_double) {
      free(c->d);
    }
  }
  free(c);
}

/*----[ Registering stuff ]-----------------------------------------------*/

/**
 * fast_thash_register:
 *
 * @th: the hash table
 * @type: the type of the hash table cell
 * @min: the minimum
 * @max: the maximum
 *
 * register a cell within an hash table
 */

int fast_thash_register(fast_thash_t **th,
			const char *name, fast_thc_type_t type,
			int min, int max){
  fast_thc_t *res;

  if (!fast_cache_retrieve((fast_cache_t*)*th,name,(void**)&res))
    FAIL("Internal error in cache while searching in a thash\n");

  if (res) {
    WARN1("Cell tyoe %s registered twice.\n",name);
    /* Will be removed by the insertion of the new one */
  } 

  if (!(res=(fast_thc_t*)malloc(sizeof(fast_thc_t))))
    FAIL("Malloc error: out of memory\n");

  res->type=type;
  res->min=min;
  res->max=max;

  res->size=0;
  res->str=NULL;
  res->i=NULL;
  res->d=NULL;

  return fast_cache_insert((fast_cache_t**)th,strdup(name),res,&fast_thash_cell_free);
}

/**
 * fast_thash_unregister:
 * 
 * @th: the hash table
 * @name: the name of the cell to be freed
 * @Returns: true if everything went ok, false if an error occur
 * 
 * unregister an hash table
 */


int fast_thash_unregister(fast_thash_t *th,const char *name) {
  return fast_cache_remove((fast_cache_t*)th,name);
}

/**
 * fast_thash_register_str:
 *
 * @th: the hash table
 * @entry: the entry to register
 * @Returns: True if everything went ok, false if an error occur
 */


int fast_thash_register_str(fast_thash_t **th,const char *entry) {
  char *entrycpy=strdup(entry);
  char *tok;

  int min,max;
  fast_thc_type_t type;

  tok=strchr(entrycpy, ':');
  if (!tok) {
    ERROR3("%s%s%s",
	  "Invalid typed hash cell descriptor: ",entry,
	  "\nShould be <name>:<min nb>_to_<max nb>_<type>\n");
    free(entrycpy);
    return 0;
  }
  *(tok++)='\0';

  min=strtol(tok, &tok, 10);
  if (!tok) {
    ERROR1("Invalid minimum in typed hash cell descriptor %s\n",entry);
    free(entrycpy);
    return 0;
  }

  if (!strcmp(tok,"_to_")){
    ERROR3("%s%s%s",
	  "Invalid typed hash descriptor: ",entry,
	  "\nShould be <name>:<min nb>_to_<max nb>_<type>\n");
    free(entrycpy);
    return 0;
  }
  tok += strlen("_to_");

  max=strtol(tok, &tok, 10);
  if (!tok) {
    ERROR1("Invalid maximum in typed hash descriptor %s\n",entry);
    free(entrycpy);
    return 0;
  }

  if (*(tok++)!='_') {
    ERROR3("%s%s%s",
	  "Invalid typed hash descriptor: ",entry,
	  "\nShould be <name>:<min nb>_to_<max nb>_<type>\n");
    free(entrycpy);
    return 0;
  }

  for (type=0; 
       type<fast_thc_type_count && strcmp(tok,fast_thc_type_name[type]); 
       type++);
  if (type == fast_thc_type_count) {
    ERROR3("%s%s%s",
	  "Invalid type in typed hash descriptor: ",entry,
	  "\nShould be one of 'string','int', 'host' or 'double'.\n");
    free(entrycpy);
    return 0;
  }

  if (!fast_thash_register(th,entrycpy,type,min,max)) {
    free(entrycpy);
    return 0;
  }
  free(entrycpy); /* strdup'ed by cache mechanism, but cannot be const */
  return 1;
}

/**
 * fast_thash_check:
 *
 * @th: the hash table
 * @Returns: True if the hash is ok, false if not
 * 
 * Check the hash table
 */


int fast_thash_check(fast_thash_t *th) {
  fast_cache_cursor_t *cursor; 
  fast_thc_t *cell;
  char *name;
  int res;
  int problem=0;

  if (!th) 
    FAIL("NULL typed hash.\n");
  
  if (!fast_cache_cursor_new(th,&cursor))
    return 0;
  
  do {
    cell=(void*)fast_cache_cursor_get_data(cursor);
    
    if (cell->min > cell->size) { 
      if (!(res=fast_cache_cursor_get_key(cursor,(char**)&name))) {
	fast_cache_cursor_free(cursor);
	return res;
      }
      fprintf(stderr,"Cell type %s needs at least %d %s, but there is only %d values.\n",
	      name,cell->min,fast_thc_type_name[cell->type],
	      cell->size);
      free(name);
      problem++;
    }

    if (cell->max < cell->size) {
      if (!(res=fast_cache_cursor_get_key(cursor,(char**)&name))) {
	fast_cache_cursor_free(cursor);
	return res;
      }
      fprintf(stderr,"Cell type %s accepts at most %d %s, but there is %d values.\n",
	      name,cell->max,fast_thc_type_name[cell->type],
	      cell->size);
      free(name);
      problem++;
    }
      
  } while (fast_cache_cursor_next(cursor)); 

  fast_cache_cursor_free(cursor);
  return !problem;
}

static int fast_thash_cell_get(fast_thash_t *th,
			       const char *name,
			       fast_thc_type_t type,
			       /* OUT */ fast_thc_t **whereto){
  if (!fast_cache_retrieve((fast_cache_t*)th,name,(void**)whereto))
    FAIL1("Internal error while accessing to the configuration entry %s\n",
	  name);

  if (!*whereto) {
    fprintf(stderr,"No registered cell %s in this typed hash\n",name);
    return 0;
  }

  if ((*whereto)->type != type)
    FAIL3("You tried to access to the configuration entry %s as an %s, but its type is %s.\n",
	  name,
	  fast_thc_type_name[type],
	  fast_thc_type_name[(*whereto)->type]);

  return 1;
}

/**
 * fast_thash_get_type:
 *
 * @th: the hash table
 * @name: the name of the cell 
 * @type: the result
 * @Results: True if everything went ok, false if an error occur
 *
 * Give the type of the hash table
 */


int fast_thash_get_type(fast_thash_t *th, const char *name, 
			/* OUT */fast_thc_type_t *type) {

  fast_thc_t *cell;

  if (!fast_cache_retrieve((fast_cache_t*)th,name,(void**)&cell))
    FAIL1("Internal error while accessing to the typed hash cell '%s'\n",
	  name);

  if (!cell) {
    //    abort();
    FAIL1("Can't get the type of '%s' since this cell does not exist\n",name);
  }

  *type=cell->type;

  return 1;
}

/*----[ Setting ]---------------------------------------------------------*/
/** 
 * fast_thash_set_vargs(): 
 * @th: typed hash to fill
 * @varargs: NULL-terminated list of pairs {(const char*)key, value}
 *
 * Add some values to the typed hash.
 * @warning: if the list isn't NULL terminated, it will segfault. 
 */
int fast_thash_set_vargs(fast_thash_t *th, va_list pa) {
  char *str,*name;
  int i;
  double d;
  fast_thc_type_t type;
  
  while ((name=va_arg(pa,char *))) {

    if (!fast_thash_get_type(th,name,&type)) {
      FAIL1("Can't set the property '%s' since it's not registered\n",name);
    }

    switch (type) {
    case fast_thc_host:
      str = va_arg(pa, char *);
      i=va_arg(pa,int);
      if (!fast_thash_set_host(th,name,str,i))
	return 0;
      break;
      
    case fast_thc_string:
      str=va_arg(pa, char *);
      if (!fast_thash_set_string(th, name, str))
	return 0;
      break;

    case fast_thc_int:
      i=va_arg(pa,int);
      if (!fast_thash_set_int(th,name,i))
	return 0;
      break;

    case fast_thc_double:
      d=va_arg(pa,double);
      if (!fast_thash_set_double(th,name,d))
	return 0; 
      break;

    default: 
      FAIL1("Typed hash cell %s not valid.",name);
    }
  }
  return 1;
}

/** 
 * fast_thash_set():
 * @th: typed hash to fill
 * @varargs: NULL-terminated list of pairs {(const char*)key, value}
 * @Returns: true if eveything went ok, false if an error occur
 *
 * Add some values to the typed hash.
 * @warning: if the list isn't NULL terminated, it will segfault. 
 */
int fast_thash_set(fast_thash_t *th, ...) {
  va_list pa;
  int errcode;

  va_start(pa,th);
  errcode=fast_thash_set_vargs(th,pa);
  va_end(pa);
  return errcode;
}

/**
 * fast_thash_set_parse():
 * @th: typed hash to fill
 * @options: a string containing the content to add to the typed hash. This
 * is a '\t',' ' or '\n' separated list of cells. Each individual cell is
 * like "<name>:<value>" where <name> is the name of an already registred 
 * cell, and <value> conforms to the data type under which this cell was
 * registred.
 * @Return: 1 if ok, 0 if not.
 *
 * Add the cells described in a string to a typed hash.
 */

int fast_thash_set_parse(fast_thash_t *th, const char *options) {
  int i;
  double d;
  char *str;

  fast_thc_t *cell;
  char *optionlist_cpy;
  char *option,  *name,*val;

  int len;


  if (!options || !strlen(options)) { /* nothing to do */
    return 1;
  }
  optionlist_cpy=strdup(options);

  DDEBUG1("Options list='%s'\n",options);
  option=optionlist_cpy;
  while (1) { /* breaks in the code */

    if (!option) 
      break;
    name=option;
    len=strlen(name);
    DDEBUG3("Parse list '%s'. len=%d; option-name=%d\n",name,len,option-name);

    /* Pass the value */
    while (option-name<=(len-1) && *option != ' ' && *option != '\n' && *option != '\t') {
      //fprintf(stderr,"Take %c.\n",*option);
      option++;
    }
    if (option-name == len) {
      //fprintf(stderr,"Boundary=EOL\n");
      option=NULL; /* don't do next iteration */

    } else {
      //fprintf(stderr,"Boundary on '%c'. len=%d;option-name=%d\n",*option,len,option-name);

      /* Pass the following blank chars */
      *(option++)='\0';
      while (option-name<(len-1) && (*option == ' ' || *option == '\n' || *option == '\t')) {
	//      fprintf(stderr,"Ignore a blank char.\n");
	option++;
      }
      if (option-name == len-1)
	option=NULL; /* don't do next iteration */
    }
    DDEBUG2("this='%s';rest='%s'\n",name,option);

    if (name[0] == ' ' || name[0] == '\n' || name[0] == '\t')
      continue;
    if (!strlen(name))
      break;
    
    val=strchr(name,':');
    if (!val) {
      ERROR1("Malformated option: '%s'\n should be of the form 'name:value'\n",
	    name);
      free(optionlist_cpy);
      return 0;
    }
    *(val++)='\0';

    DDEBUG2("name='%s';val='%s'\n",name,val);
    if (!fast_cache_retrieve((fast_cache_t*)th,name,(void**)&cell)) {
      ERROR1("Internal error while accessing to the typed hash cell %s\n",
	    name);
      free(optionlist_cpy);
      return 0;
    }

    if (!cell) {
      fprintf(stderr,"No registrated cell corresponding to '%s'.\n",name);
      free(optionlist_cpy);
      return 0;
    }

    switch (cell->type) {
    case fast_thc_string:
      if (!fast_thash_set_string(th, name, val))
	return 0;
      break;

    case fast_thc_int:
      i=strtol(val, &val, 0);
      if (val==NULL) {
	fprintf(stderr,"Value of option %s not valid. Should be an integer\n",
	      name);
	free(optionlist_cpy);	
	return 0;
      }
	
      if (!fast_thash_set_int(th,name,i)) {
	free(optionlist_cpy);
	return 0;
      }
      break;

    case fast_thc_double:
      d=strtod(val, &val);
      if (val==NULL) {
	fprintf(stderr,"Value of option %s not valid. Should be a double.\n",
	      name);
	free(optionlist_cpy);
	return 0;
      }
      
      if (!fast_thash_set_double(th,name,d)) {
	free(optionlist_cpy);
	return 0;
      }
      break;

    case fast_thc_host:
      str=val;
      val=strchr(val,':');
      if (!val) {
	fprintf(stderr,"Value of option %s not valid. Should be an host (machine:port)\n",
		name);
	free(optionlist_cpy);
	return 0;
      }

      *(val++)='\0';
      i=strtol(val, &val, 0);
      if (val==NULL) {
	fprintf(stderr,"Value of option %s not valid. Should be an host (machine:port)\n",
		name);
	free(optionlist_cpy);
	return 0;
      }
      if (!fast_thash_set_host(th,name,str,i)) {
	free(optionlist_cpy);
	return 0;
      }
      break;      

    default: 
      ERROR1("Type of typed hash cell %s is not valid.",name);
      free(optionlist_cpy);
      return 0;
    }
    
  }
  free(optionlist_cpy);
  return 1;
}

/**
 * fast_thash_set_int:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the value of the cell
 * @Returns; true if eveything went ok, false if an error occur
 *
 * Set the value of the cell @name in @th with the provided value.
 */ 
int fast_thash_set_int(fast_thash_t *th,const char*name, int val) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_int,&cell))
    return 0;

  if (cell->max > 1) {
    cell->size++;
  } else {
    cell->size=1;
  }
  cell->i=realloc(cell->i,sizeof(int)*cell->size);
  if (!cell->i)
    FAIL("Malloc error: out of memory\n");
  cell->i[cell->size - 1] = val;
  return 1;
}

/**
 * fast_thash_set_double:
 * @th: the hash table
 * @name: the name of the cell
 * @val: the doule to set
 * @Returns: true if everything went ok, false if an error occur
 * 
 * Set the value of the cell @name in @th with the provided value.
 */ 

int fast_thash_set_double(fast_thash_t *th,const char*name, double val) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_double,&cell))
    return 0;

  if (cell->max > 1) {
    cell->size++;
  } else {
    cell->size=1;
  }
  cell->d=realloc(cell->i,sizeof(double)*cell->size);
  if (!cell->d)
    FAIL("Malloc error: out of memory\n");
  cell->d[cell->size - 1] = val;
  return 1;
}

/**
 * fast_thash_set_string:
 * 
 * @th: the hashtable
 * @name: the name of the cell
 * @val: the value to be added
 * @Returns: True if eveything went ok, false if an error occur
 *
 * Set the value of the cell @name in @th with the provided value.
 */ 

int fast_thash_set_string(fast_thash_t *th,const char*name, const char*val) { 
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_string,&cell))
    return 0;

  if (cell->max > 1) {
    cell->size++;
  } else {
    cell->size=1;
  }
  cell->str=realloc(cell->str,sizeof(char*)*cell->size);
  if (!cell->str)
    FAIL("Malloc error: out of memory\n");
  cell->str[cell->size - 1] = strdup(val);
  return 1;
}

/**
 * fast_thash_set_host:
 * 
 * @th: the hash table
 * @name: the name of the cell
 * @host: the host
 * @port: the port number
 * @Return: true if everything went Ok, false if an error occur
 *
 * Set the value of the cell @name in @th with the provided value 
 * on the given @host to the given @port
 */ 

int fast_thash_set_host(fast_thash_t *th,const char*name, 
			const char *host,int port) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_host,&cell))
    return 0;

  if (cell->max > 1) {
    cell->size++;
  } else {
    cell->size=1;
  }
  cell->str=realloc(cell->str,sizeof(char*)*cell->size);
  cell->i=realloc(cell->i,sizeof(int)*cell->size);
  if (!cell->str || !cell->i)
    FAIL("Malloc error: out of memory\n");
  cell->str[cell->size - 1] = strdup(host);
  cell->i  [cell->size - 1] = port;
  return 1;                                                       \
}

/* ---- [ Removing ] ---- */

/**
 * fast_thash_rm_int:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the value to be removed
 * @Returns: true if everything went ok, false if an error occur
 *
 * Remove the provided @val from the cell @name in @th.
 */

int fast_thash_rm_int   (fast_thash_t *th,const char*name, int val) {

  fast_thc_t *cell;
  int i;

  if (!fast_thash_cell_get(th,name,fast_thc_int,&cell)) 
    return 0;
  
  for (i=0;
       i<cell->size && cell->i[i] != val;
       i++);
  if (i>=cell->size)
    FAIL2("Can't remove the value %d of typed hash cell %s: value not found.\n",
	  val,name);

  cell->size--;
  if (cell->size == 0) {
    free(cell->i);
  } else if (i<cell->size -1) {
    memmove((void*)&cell->i[i],(void*)&cell->i[i+1],(cell->size-i-1)*sizeof(int));
    /* don't realloc to increase speed */
  }
  return 1;
}

/**
 * fast_thash_rm_double:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the value to be removed
 * @Returns: True if everything went ok, false if an error occur
 *
 * Remove the provided @val from the cell @name in @th.
 */

int fast_thash_rm_double(fast_thash_t *th,const char*name, double val) {
  fast_thc_t *cell;
  int i;

  if (!fast_thash_cell_get(th,name,fast_thc_double,&cell)) 
    return 0;
  
  for (i=0;
       i<cell->size && cell->d[i] != val;
       i++);
  if (i>=cell->size)
    FAIL2("Can't remove the value %f of typed hash cell %s: value not found.\n",
	  val,name);

  cell->size--;
  if (cell->size == 0) {
    free(cell->d);
  } else if (i<cell->size -1) {
    memmove((void*)&cell->d[i],(void*)&cell->d[i+1],(cell->size-i-1)*sizeof(double));
    /* don't realloc to increase speed */
  }
  return 1;

/**
 * fast_thash_rm_string:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the value of the string which will be removed
 * @Returns: true if everything went ok, false if an error occur
 *
 * Remove the provided @val from the cell @name in @th.
 */

}
int fast_thash_rm_string(fast_thash_t *th,const char*name, const char *val) {
  fast_thc_t *cell;
  int i;

  if (!fast_thash_cell_get(th,name,fast_thc_string,&cell)) 
    return 0;
  
  for (i=0;
       i<cell->size && strcmp(cell->str[i],val);
       i++);
  if (i>=cell->size)
    FAIL2("Can't remove the value %s of typed hash cell %s: value not found.\n",
	  val,name);

  cell->size--;
  if (cell->size == 0) {
    free(cell->str);
  } else if (i<cell->size -1) {
    memmove((void*)&cell->str[i],(void*)&cell->str[i+1],(cell->size-i-1)*sizeof(char*));
    /* don't realloc to increase speed */
  }
  return 1;
}

/**
 * fast_thash_rm_host:
 * 
 * @th: the hash table
 * @name: the name of the cell
 * @host: the hostname
 * @port: the port number
 * @Returns: True if eveything went ok, false if an error occur
 *
 * Remove the provided @host from the cell @name in @th.
 */

int fast_thash_rm_host  (fast_thash_t *th,const char*name, const char *host,int port) {
  fast_thc_t *cell;
  int i;

  if (!fast_thash_cell_get(th,name,fast_thc_host,&cell)) 
    return 0;
  
  for (i=0;
       i<cell->size && (strcmp(cell->str[i],host) || cell->i[i] != port);
       i++);
  if (i>=cell->size)
    FAIL3("Can't remove the value %s:%d of typed hash cell %s: value not found.\n",
	  host,port,name);

  cell->size--;
  if (cell->size == 0) { 
    free(cell->str);
    free(cell->i);
  } else if (i<cell->size -1) {
    memmove((void*)&cell->str[i],(void*)&cell->str[i+1],(cell->size-i-1)*sizeof(char*));
    memmove((void*)&cell->i[i],(void*)&cell->i[i+1],(cell->size-i-1)*sizeof(int));
    /* don't realloc to increase speed */
  }
  return 1;
}

/* rm everything */

/**
 * fast_thash_empty:
 * 
 * @th: the hash table
 * @name: the name of the cell
 * @Returns: True if everything went ok, false if an error occur
 *
 * rm evenything
 */
 
int fast_thash_empty(fast_thash_t *th,const char*name) {
  fast_thc_t *cell;
  int i;

  if (!fast_cache_retrieve((fast_cache_t*)th,name,(void**)&cell))
    FAIL1("Internal error while accessing to the configuration cell %s\n",
	  name);

  if (cell) {
    if (cell->size >0){
      if (cell->type == fast_thc_host ||
	  cell->type == fast_thc_string) {
	for (i=0;i<cell->size;i++) {
	  free(cell->str[i]);
	}
	free(cell->str);
      }
      if (cell->type == fast_thc_host ||
	  cell->type == fast_thc_int) {
	free(cell->i);
      }
      if (cell->type == fast_thc_double) {
	free(cell->d);
      }
    }
  }
  return 1;
}

/*----[ Getting ]---------------------------------------------------------*/

/**
 * fast_thash_get_int:
 * 
 * @th: the hash table
 * @name: the name of the cell
 * @val: the wanted values
 * @size: the count of values
 * @Returns: True if everything went ok, false if an error occur
 *
 * Returns a pointer to the values actually stored in the cache.
 * Do not modify them unless you really know what you're doing.
 */

int fast_thash_get_int   (fast_thash_t *th,const char*name,
			   int **val,int *size) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_int,&cell)) 
    return 0;

  *val=cell->i;
  *size=cell->size;

  return 1;
}

/**
 * fast_thash_get_double:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the wanted values
 * @size: the count of values
 * @Returns: true if eveything went ok, false if not
 *
 * Returns a pointer to the values actually stored in the cache.
 * Do not modify them unless you really know what you're doing.
 */

int fast_thash_get_double(fast_thash_t *th,const char*name,
			   double **val,int *size) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_double,&cell)) 
    return 0;

  *val=cell->d;
  *size=cell->size;

  return 1;
}

/**
 * fast_thash_get_string:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the wanted value
 * @size: the count of values
 * @Returns: true if everything went ok, false if an error occurs
 *
 * Returns a pointer to the values actually stored in the cache.
 * Do not modify them unless you really know what you're doing.
 */

int fast_thash_get_string(fast_thash_t *th,const char*name,
			   char ***val, int *size) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_string,&cell)) {
    *size=0;
    *val=NULL;
    return 0;
  }

  *val=cell->str;
  *size=cell->size;

  return 1;
}

/**
 * fast_thash_get_host:
 *
 * @th: the Hash table
 * @name: the name of the cell
 * @host: the host
 * @port: the port number
 * @size: the count of values
 * @Returns: true if everything went ok, false if not
 *
 * Returns a pointer to the values actually stored in the cache.
 * Do not modify them unless you really know what you're doing.
 */

int fast_thash_get_host  (fast_thash_t *th,const char*name,
		           char ***host,int **port, int *size) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_host,&cell)) 
    return 0;

  *host=cell->str;
  *port=cell->i;
  *size=cell->size;

  return 1;
}

/**
 * fast_thash_getfirst_int:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the value in the first cell
 * @Returns: true if eveything went Ok, false if not
 *
 * Get only the first cell. Usefull if you know there 
 * is only one elem
 */
 

int fast_thash_getfirst_int   (fast_thash_t *th,const char*name,
			      int *val) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_int,&cell)) 
    return 0;

  if (cell->size == 0) {
    if (cell->min) {
      FAIL1("Typed hash cell %s has no set value\n",name);
    } else {
      return 0;
    }
  }
    
  *val=cell->i[0];

  return 1;
}

/**
 * fast_thash_getfirst_double:
 *
 * @th: the hashtable
 * @name: the name of the cell
 * @val: the first value
 * @Returns: True if eveything went ok, false if an error occurs
 *
 * Get only the first cell. Usefull if you know there 
 * is only one elem
 */

int fast_thash_getfirst_double(fast_thash_t *th,const char*name,
			      double *val) {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_double,&cell)) 
    return 0;

  if (cell->size == 0)
    FAIL1("Typed hash cell %s has no set value\n",name);
    
  *val=cell->d[0];

  return 1;
}

/**
 * fast_thash_getfirst_string:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @val: the value of the first cell
 * @Returns: true if eveything went ok, false if not
 *
 * Get only the first cell. Usefull if you know there 
 * is only one elem
 */

int fast_thash_getfirst_string(fast_thash_t *th,const char*name,
			      char **val)  {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_string,&cell)) 
    return 0;

  if (cell->size == 0)
    FAIL1("Typed hash cell %s has no set value\n",name);
    
  *val=cell->str[0];

  return 1;
}

/**
 * fast_thash_getfirst_host:
 *
 * @th: the hash table
 * @name: the name of the cell
 * @host: the host
 * @port: the port number
 * @Returns: true if everything went ok, false if not
 *
 * Get only the first cell. Usefull if you know there 
 * is only one elem
 */

int fast_thash_getfirst_host  (fast_thash_t *th,const char*name,
			      char **host,int *port)  {
  fast_thc_t *cell;

  if (!fast_thash_cell_get(th,name,fast_thc_host,&cell)) 
    return 0;

  if (cell->size == 0)
    FAIL1("Typed hash cell %s has no set value\n",name);
    
  *host=cell->str[0];
  *port=cell->i[0];

  return 1;
}
