/* $Id: typed_hash.h,v 1.4 2003/06/05 12:41:52 mquinson Exp $ */

/* typed_hash - Hash table where all cell is of a determined type.        */

/* This is useful to build named structs, like option or property sets.   */

/* Copyright (C) 2001 the OURAGAN project.                                */
/* Code available under BSD-like licence (see COPYRIGHT file).            */
/* Author(s): Martin Quinson                                              */

#ifndef _TYPED_HASH_H_
#define _TYPED_HASH_H_

/* For now, a typed hash is only a special cache. But don't rely on it, */
/* it may change in the future. */
typedef fast_cache_t fast_thash_t;

/* type of a typed hash cell */
typedef enum {
  fast_thc_int=0, fast_thc_double, fast_thc_string, fast_thc_host,
  fast_thc_type_count
} fast_thc_type_t;

/*----[ Memory management ]-----------------------------------------------*/
int fast_thash_init (fast_thash_t **whereto); /* (whereto == NULL) is ok */
int fast_thash_new_from(fast_thash_t **whereto, fast_thash_t *tocopy);
void fast_thash_free(fast_thash_t **th);
void fast_thash_dump(const char *name,const char*indent,fast_thash_t *th);

/*----[ Registering stuff ]-----------------------------------------------*/
/* Register a possible cell */
int fast_thash_register(fast_thash_t **th,
			const char *name, fast_thc_type_t type,
			int min, int max);
/* Unregister a possible cell */
int fast_thash_unregister(fast_thash_t *th, const char *name);

/* Parse the configuration descriptor and register it */
/* Should be of the form "<name>:<min nb>_to_<max nb>_<type>", */
/*  with type being one of 'string','int', 'host' or 'double'  */
int fast_thash_register_str(fast_thash_t **th, const char *entry);

/* Check that each cell have the right amount of elements */
int fast_thash_check(fast_thash_t *th);

/* Get the type of this option in that repository */
int fast_thash_get_type(fast_thash_t *th, const char *name, 
			/* OUT */ fast_thc_type_t *type);

/*----[ Setting ]---------------------------------------------------------
 * fast_thash_set_* functions.
 *
 * If the registered maximum is equal to 1, those functions remplace the 
 * current value with the provided one. If max>1, the provided value is 
 * appended to the list.
 *
 * string values are strdup'ed before use, so you have to free your copy  */

int fast_thash_set_vargs(fast_thash_t *th, va_list pa);
int fast_thash_set(fast_thash_t *th, ...);

/*
  Add the cells described in a string to a typed hash.
 */
int fast_thash_set_parse(fast_thash_t *th, const char *options);


/*
  Set the value of the cell @name in @th with the provided value.
 */
int fast_thash_set_int   (fast_thash_t *th,const char*name, int val);
int fast_thash_set_double(fast_thash_t *th,const char*name, double val);
int fast_thash_set_string(fast_thash_t *th,const char*name, const char *val);
int fast_thash_set_host  (fast_thash_t *th,const char*name, const char *host,int port);

/*
 Remove the provided value from the cell @name in @th.
 */
int fast_thash_rm_int   (fast_thash_t *th,const char*name, int val);
int fast_thash_rm_double(fast_thash_t *th,const char*name, double val);
int fast_thash_rm_string(fast_thash_t *th,const char*name, const char *val);
int fast_thash_rm_host  (fast_thash_t *th,const char*name, const char *host,int port);

/* rm every values */
int fast_thash_empty(fast_thash_t *th,const char*name);	

/*----[ Getting ]---------------------------------------------------------*/
/* Returns a pointer to the values actually stored in the cache. Do not   */
/* modify them unless you really know what you're doing.                  */
int fast_thash_get_int   (fast_thash_t *th,const char*name,
			  int **val,int *size);
int fast_thash_get_double(fast_thash_t *th,const char*name,
			  double **val,int *size);
int fast_thash_get_string(fast_thash_t *th,const char*name, 
			  char ***val, int *size);
int fast_thash_get_host  (fast_thash_t *th,const char*name, 
			  char ***host,int **port, int *size);

/* Get only the first entry. Usefull if you know there is only one elem */
int fast_thash_getfirst_int   (fast_thash_t *repository,const char*name,
			       int *val);
int fast_thash_getfirst_double(fast_thash_t *repository,const char*name,
			       double *val);
int fast_thash_getfirst_string(fast_thash_t *repository,const char*name,
			       char **val);
int fast_thash_getfirst_host  (fast_thash_t *repository,const char*name,
			       char **host,int *port);

#endif /* _TYPED_HASH_H_ */
